/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "bindio.h"
#include "bindinterface.h"
#include "bindutil.h"

#include "io.h"
#include "jsontype.h"
#include "xmltype.h"
#include "zipfile.h"
#include <stdexcept>
#include <cstring>

namespace bipscript {

namespace io {

namespace binding {

// object references to types in this package
HSQOBJECT FileObject;
HSQOBJECT FolderObject;
HSQOBJECT JsonObject;
HSQOBJECT XmlObject;
HSQOBJECT ZipObject;
HSQOBJECT ZipFileObject;
HSQOBJECT ZipFolderObject;

//
// IO.File class
//
#define BIND_CLASS_NAME "File"
#define BIND_METHOD_NAME "constructor"
SQInteger FileCtor(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    File *obj;
    BIND_NARGS_CHECK_EXACT(1) // check exact parameter count
    // get parameter 1 "path" as string
    BIND_GET_CHECK_STRING(path, 1)
    // call the implementation
    BIND_WRAP_TRY(obj = new File(path))
    // return pointer to new object
    sq_setinstanceup(vm, 1, (SQUserPointer*)obj);
    sq_setreleasehook(vm, 1, BindUtil::Delete<File>);
    return 1;
}
#undef BIND_METHOD_NAME

//
// IO.File property getters
//
#define BIND_METHOD_NAME "getter"
SQInteger FileGetters(HSQUIRRELVM vm)
{
    SQObjectType argtype = sq_gettype(vm, 2);
    if(argtype == OT_STRING) {
    // get property name
    const SQChar* key;
    sq_getstring(vm, 2, &key);
    BIND_GET_THIS_PROPERTY(io::File)
    // dispatch appropriate getter
    if(!strcmp(key, "name")) {
        std::string ret = obj->name();
        BIND_PUSH_STRING(ret)
        return 1;
    }
    if(!strcmp(key, "path")) {
        std::string ret = obj->path();
        BIND_PUSH_STRING(ret)
        return 1;
    }
    }
    sq_pushnull(vm);
    return sq_throwobject(vm);
}
#undef BIND_METHOD_NAME

//
// IO.File property setters
//
#define BIND_METHOD_NAME "setter"
SQInteger FileSetters(HSQUIRRELVM vm)
{
    SQObjectType argtype = sq_gettype(vm, 2);
    if(argtype == OT_STRING) {
    // get property name
    const SQChar* key;
    sq_getstring(vm, 2, &key);
    BIND_GET_THIS_PROPERTY(io::File)
    // dispatch appropriate setter
    if(!strcmp(key, "name")) {
        return sq_throwerror(vm, "property \"name\" is read-only");
    }
    if(!strcmp(key, "path")) {
        return sq_throwerror(vm, "property \"path\" is read-only");
    }
    }
    sq_pushnull(vm);
    return sq_throwobject(vm);
}
#undef BIND_METHOD_NAME

//
// IO.File bytes
//
#define BIND_METHOD_NAME "bytes"
SQInteger Filebytes(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    BIND_GET_THIS_INSTANCE(File)
    BIND_NARGS_CHECK_EXACT(0) // check exact parameter count
    // return value
    bipscript::Buffer* ret;
    // call the implementation
    BIND_WRAP_TRY(ret = obj->bytes())
    // push return value
    BIND_PUSH_OBJECT(ret->addref(), buffer, bipscript, Buffer, Decrement)
    return 1;
}
#undef BIND_METHOD_NAME

//
// IO.File exists
//
#define BIND_METHOD_NAME "exists"
SQInteger Fileexists(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    BIND_GET_THIS_INSTANCE(File)
    BIND_NARGS_CHECK_EXACT(0) // check exact parameter count
    // return value
    SQBool ret;
    // call the implementation
    BIND_WRAP_TRY(ret = obj->exists())
    // push return value
    BIND_PUSH_BOOL(ret)
    return 1;
}
#undef BIND_METHOD_NAME

//
// IO.File folder
//
#define BIND_METHOD_NAME "folder"
SQInteger Filefolder(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    BIND_GET_THIS_INSTANCE(File)
    BIND_NARGS_CHECK_EXACT(0) // check exact parameter count
    // return value
    SQBool ret;
    // call the implementation
    BIND_WRAP_TRY(ret = obj->folder())
    // push return value
    BIND_PUSH_BOOL(ret)
    return 1;
}
#undef BIND_METHOD_NAME

//
// IO.File len
//
#define BIND_METHOD_NAME "len"
SQInteger Filelen(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    BIND_GET_THIS_INSTANCE(File)
    BIND_NARGS_CHECK_EXACT(0) // check exact parameter count
    // return value
    SQInteger ret;
    // call the implementation
    BIND_WRAP_TRY(ret = obj->len())
    // push return value
    BIND_PUSH_INTEGER(ret)
    return 1;
}
#undef BIND_METHOD_NAME

//
// IO.File parent
//
#define BIND_METHOD_NAME "parent"
SQInteger Fileparent(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    BIND_GET_THIS_INSTANCE(File)
    BIND_NARGS_CHECK_EXACT(0) // check exact parameter count
    // return value
    io::Folder* ret;
    // call the implementation
    BIND_WRAP_TRY(ret = obj->parent())
    // push return value
    BIND_PUSH_OBJECT(ret, Folder, io, Folder, Delete)
    return 1;
}
#undef BIND_METHOD_NAME

//
// IO.File readAll
//
#define BIND_METHOD_NAME "readAll"
SQInteger FilereadAll(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    BIND_GET_THIS_INSTANCE(File)
    BIND_NARGS_CHECK_EXACT(0) // check exact parameter count
    // return value
    std::string ret;
    // call the implementation
    BIND_WRAP_TRY(ret = obj->readAll())
    // push return value
    BIND_PUSH_STRING(ret)
    return 1;
}
#undef BIND_METHOD_NAME

//
// IO.File reader
//
#define BIND_METHOD_NAME "reader"
SQInteger Filereader(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    BIND_GET_THIS_INSTANCE(File)
    BIND_NARGS_CHECK_EXACT(0) // check exact parameter count
    // return value
    bipscript::Reader* ret;
    // call the implementation
    BIND_WRAP_TRY(ret = obj->reader())
    // push return value
    BIND_PUSH_OBJECT(ret->addref(), reader, bipscript, Reader, Decrement)
    return 1;
}
#undef BIND_METHOD_NAME

//
// IO.File remove
//
#define BIND_METHOD_NAME "remove"
SQInteger Fileremove(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    BIND_GET_THIS_INSTANCE(File)
    BIND_NARGS_CHECK_EXACT(0) // check exact parameter count
    // return value
    SQBool ret;
    // call the implementation
    BIND_WRAP_TRY(ret = obj->remove())
    // push return value
    BIND_PUSH_BOOL(ret)
    return 1;
}
#undef BIND_METHOD_NAME

//
// IO.File text
//
#define BIND_METHOD_NAME "text"
SQInteger Filetext(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    BIND_GET_THIS_INSTANCE(File)
    BIND_NARGS_CHECK_EXACT(0) // check exact parameter count
    // return value
    const SQChar* ret;
    // call the implementation
    BIND_WRAP_TRY(ret = obj->text())
    // push return value
    BIND_PUSH_CSTRING(ret)
    return 1;
}
#undef BIND_METHOD_NAME

//
// IO.File writeAll
//
#define BIND_METHOD_NAME "writeAll"
SQInteger FilewriteAll(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    BIND_GET_THIS_INSTANCE(File)
    BIND_NARGS_CHECK_EXACT(1) // check exact parameter count
    // get parameter 1 "body" as string
    BIND_GET_CHECK_STRING(body, 1)
    // call the implementation
    BIND_WRAP_TRY(obj->writeAll(body))
    // void method, returns no value
    return 0;
}
#undef BIND_METHOD_NAME

//
// IO.File writer
//
#define BIND_METHOD_NAME "writer"
SQInteger Filewriter(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    BIND_GET_THIS_INSTANCE(File)
    BIND_NARGS_CHECK_EXACT(0) // check exact parameter count
    // return value
    bipscript::Writer* ret;
    // call the implementation
    BIND_WRAP_TRY(ret = obj->writer())
    // push return value
    BIND_PUSH_OBJECT(ret, writer, bipscript, Writer, Delete)
    return 1;
}
#undef BIND_METHOD_NAME

SQInteger FileToString(HSQUIRRELVM vm)
{
    sq_pushstring(vm, "(IO.File)", 9);
    return 1;
}
#undef BIND_CLASS_NAME

//
// IO.Folder class
//
#define BIND_CLASS_NAME "Folder"
#define BIND_METHOD_NAME "constructor"
SQInteger FolderCtor(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    Folder *obj;
    BIND_NARGS_CHECK_EXACT(1) // check exact parameter count
    // get parameter 1 "path" as string
    BIND_GET_CHECK_STRING(path, 1)
    // call the implementation
    BIND_WRAP_TRY(obj = new Folder(path))
    // return pointer to new object
    sq_setinstanceup(vm, 1, (SQUserPointer*)obj);
    sq_setreleasehook(vm, 1, BindUtil::Delete<Folder>);
    return 1;
}
#undef BIND_METHOD_NAME

//
// IO.Folder property getters
//
#define BIND_METHOD_NAME "getter"
SQInteger FolderGetters(HSQUIRRELVM vm)
{
    SQObjectType argtype = sq_gettype(vm, 2);
    if(argtype == OT_STRING) {
    // get property name
    const SQChar* key;
    sq_getstring(vm, 2, &key);
    BIND_GET_THIS_PROPERTY(io::Folder)
    // dispatch appropriate getter
    if(!strcmp(key, "name")) {
        std::string ret = obj->name();
        BIND_PUSH_STRING(ret)
        return 1;
    }
    if(!strcmp(key, "path")) {
        std::string ret = obj->path();
        BIND_PUSH_STRING(ret)
        return 1;
    }
    }
    sq_pushnull(vm);
    return sq_throwobject(vm);
}
#undef BIND_METHOD_NAME

//
// IO.Folder property setters
//
#define BIND_METHOD_NAME "setter"
SQInteger FolderSetters(HSQUIRRELVM vm)
{
    SQObjectType argtype = sq_gettype(vm, 2);
    if(argtype == OT_STRING) {
    // get property name
    const SQChar* key;
    sq_getstring(vm, 2, &key);
    BIND_GET_THIS_PROPERTY(io::Folder)
    // dispatch appropriate setter
    if(!strcmp(key, "name")) {
        return sq_throwerror(vm, "property \"name\" is read-only");
    }
    if(!strcmp(key, "path")) {
        return sq_throwerror(vm, "property \"path\" is read-only");
    }
    }
    sq_pushnull(vm);
    return sq_throwobject(vm);
}
#undef BIND_METHOD_NAME

//
// IO.Folder children
//
#define BIND_METHOD_NAME "children"
SQInteger Folderchildren(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    BIND_GET_THIS_INSTANCE(Folder)
    BIND_NARGS_CHECK_EXACT(0) // check exact parameter count
    // return value
    bipscript::Iterator* ret;
    // call the implementation
    BIND_WRAP_TRY(ret = obj->children())
    // push return value
    BIND_PUSH_OBJECT(ret, iterator, bipscript, Iterator, Delete)
    return 1;
}
#undef BIND_METHOD_NAME

//
// IO.Folder exists
//
#define BIND_METHOD_NAME "exists"
SQInteger Folderexists(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    BIND_GET_THIS_INSTANCE(Folder)
    BIND_NARGS_CHECK_EXACT(0) // check exact parameter count
    // return value
    SQBool ret;
    // call the implementation
    BIND_WRAP_TRY(ret = obj->exists())
    // push return value
    BIND_PUSH_BOOL(ret)
    return 1;
}
#undef BIND_METHOD_NAME

//
// IO.Folder folder
//
#define BIND_METHOD_NAME "folder"
SQInteger Folderfolder(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    BIND_GET_THIS_INSTANCE(Folder)
    BIND_NARGS_CHECK_EXACT(0) // check exact parameter count
    // return value
    SQBool ret;
    // call the implementation
    BIND_WRAP_TRY(ret = obj->folder())
    // push return value
    BIND_PUSH_BOOL(ret)
    return 1;
}
#undef BIND_METHOD_NAME

//
// IO.Folder mkdir
//
#define BIND_METHOD_NAME "mkdir"
SQInteger Foldermkdir(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    BIND_GET_THIS_INSTANCE(Folder)
    BIND_NARGS_CHECK_EXACT(0) // check exact parameter count
    // call the implementation
    BIND_WRAP_TRY(obj->mkdir())
    // void method, returns no value
    return 0;
}
#undef BIND_METHOD_NAME

//
// IO.Folder parent
//
#define BIND_METHOD_NAME "parent"
SQInteger Folderparent(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    BIND_GET_THIS_INSTANCE(Folder)
    BIND_NARGS_CHECK_EXACT(0) // check exact parameter count
    // return value
    io::Folder* ret;
    // call the implementation
    BIND_WRAP_TRY(ret = obj->parent())
    // push return value
    BIND_PUSH_OBJECT(ret, Folder, io, Folder, Delete)
    return 1;
}
#undef BIND_METHOD_NAME

SQInteger FolderToString(HSQUIRRELVM vm)
{
    sq_pushstring(vm, "(IO.Folder)", 11);
    return 1;
}
#undef BIND_CLASS_NAME

//
// IO.Json class
//
#define BIND_CLASS_NAME "Json"
#define BIND_METHOD_NAME "constructor"
SQInteger JsonCtor(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    Json *obj;
    BIND_NARGS_CHECK_EXACT(0) // check exact parameter count
    // call the implementation
    BIND_WRAP_TRY(obj = new Json())
    // return pointer to new object
    sq_setinstanceup(vm, 1, (SQUserPointer*)obj);
    sq_setreleasehook(vm, 1, BindUtil::Delete<Json>);
    return 1;
}
#undef BIND_METHOD_NAME

//
// IO.Json property getters
//
#define BIND_METHOD_NAME "getter"
SQInteger JsonGetters(HSQUIRRELVM vm)
{
    SQObjectType argtype = sq_gettype(vm, 2);
    if(argtype == OT_STRING) {
    // get property name
    const SQChar* key;
    sq_getstring(vm, 2, &key);
    BIND_GET_THIS_PROPERTY(io::Json)
    // dispatch appropriate getter
    if(!strcmp(key, "verbose")) {
        SQBool ret = obj->verbose();
        BIND_PUSH_BOOL(ret)
        return 1;
    }
    }
    sq_pushnull(vm);
    return sq_throwobject(vm);
}
#undef BIND_METHOD_NAME

//
// IO.Json property setters
//
#define BIND_METHOD_NAME "setter"
SQInteger JsonSetters(HSQUIRRELVM vm)
{
    SQObjectType argtype = sq_gettype(vm, 2);
    if(argtype == OT_STRING) {
    // get property name
    const SQChar* key;
    sq_getstring(vm, 2, &key);
    BIND_GET_THIS_PROPERTY(io::Json)
    // dispatch appropriate setter
    if(!strcmp(key, "verbose")) {
        // get parameter 2 "verbose" as bool
        BIND_GET_CHECK_BOOL(verbose, 2)
        BIND_WRAP_TRY(obj->verbose(verbose))
        return 0;
    }
    }
    sq_pushnull(vm);
    return sq_throwobject(vm);
}
#undef BIND_METHOD_NAME

//
// IO.Json read
//
#define BIND_METHOD_NAME "read"
SQInteger Jsonread(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    BIND_GET_THIS_INSTANCE(Json)
    BIND_NARGS_CHECK_MIN(1) // check min parameter count
    BIND_NARGS_CHECK_MAX(2) // check max parameter count
    HSQOBJECT arg1;
    sq_getstackobj(vm, 2, &arg1);
    if(arg1._type == OT_STRING) {
        // get parameter 1 "json" as string
        BIND_GET_STRING(json, 1)
        if(numargs == 2) {
            // call the implementation
            BIND_WRAP_TRY(obj->read(vm, json))
        } else {
            // get parameter 2 "clazz" as class
            BIND_GET_CHECK_CLASS(clazz, 2)
            // call the implementation
            BIND_WRAP_TRY(obj->read(vm, json, clazz))
        }
    }
    else if(arg1._type == OT_INSTANCE) {
    SQUserPointer arg1type = bip_getinstancetypetag(&arg1);
    if(arg1type == &FileObject) {
        // get parameter 1 "file" as IO.File
        BIND_GET_OBJECT(file, 1, IO, File, io, File)
        if(numargs == 2) {
            // call the implementation
            BIND_WRAP_TRY(obj->read(vm, *file))
        } else {
            // get parameter 2 "clazz" as class
            BIND_GET_CHECK_CLASS(clazz, 2)
            // call the implementation
            BIND_WRAP_TRY(obj->read(vm, *file, clazz))
        }
    }
    else {
        return sq_throwerror(vm, "argument 1 is not of type {string, IO.File}");
    }
    }
    else {
        return sq_throwerror(vm, "argument 1 is not of type {string, IO.File}");
    }
    return 1;
}
#undef BIND_METHOD_NAME

//
// IO.Json write
//
#define BIND_METHOD_NAME "write"
SQInteger Jsonwrite(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    BIND_GET_THIS_INSTANCE(Json)
    BIND_NARGS_CHECK_MIN(1) // check min parameter count
    // get parameter 1 "object" as any
    BIND_GET_CHECK_ANY(object, 1)
    if(numargs == 2) {
        // call the implementation
        BIND_WRAP_TRY(obj->write(object))
    } else {
        HSQOBJECT arg2;
        sq_getstackobj(vm, 3, &arg2);
        if(arg2._type == OT_BOOL) {
            BIND_NARGS_CHECK_MAX(2) // check max parameter count
            // get parameter 2 "pretty" as bool
            BIND_GET_BOOL(pretty, 2)
            // call the implementation
            BIND_WRAP_TRY(obj->write(object, pretty))
        }
        else if(arg2._type == OT_INSTANCE) {
        SQUserPointer arg2type = bip_getinstancetypetag(&arg2);
        if(arg2type == &FileObject) {
            BIND_NARGS_CHECK_MIN(2) // check min parameter count
            BIND_NARGS_CHECK_MAX(3) // check max parameter count
            // get parameter 2 "file" as IO.File
            BIND_GET_OBJECT(file, 2, IO, File, io, File)
            if(numargs == 3) {
                // call the implementation
                BIND_WRAP_TRY(obj->write(object, *file))
            } else {
                // get parameter 3 "pretty" as bool
                BIND_GET_CHECK_BOOL(pretty, 3)
                // call the implementation
                BIND_WRAP_TRY(obj->write(object, *file, pretty))
            }
        }
        else {
            return sq_throwerror(vm, "argument 2 is not of type {bool, IO.File}");
        }
        }
        else {
            return sq_throwerror(vm, "argument 2 is not of type {bool, IO.File}");
        }
    }
    // void method, returns no value
    return 0;
}
#undef BIND_METHOD_NAME

SQInteger JsonToString(HSQUIRRELVM vm)
{
    sq_pushstring(vm, "(IO.Json)", 9);
    return 1;
}
#undef BIND_CLASS_NAME

//
// IO.Xml class
//
#define BIND_CLASS_NAME "Xml"
#define BIND_METHOD_NAME "constructor"
SQInteger XmlCtor(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    Xml *obj;
    BIND_NARGS_CHECK_EXACT(0) // check exact parameter count
    // call the implementation
    BIND_WRAP_TRY(obj = new Xml())
    // return pointer to new object
    sq_setinstanceup(vm, 1, (SQUserPointer*)obj);
    sq_setreleasehook(vm, 1, BindUtil::Delete<Xml>);
    return 1;
}
#undef BIND_METHOD_NAME

//
// IO.Xml property getters
//
#define BIND_METHOD_NAME "getter"
SQInteger XmlGetters(HSQUIRRELVM vm)
{
    SQObjectType argtype = sq_gettype(vm, 2);
    if(argtype == OT_STRING) {
    // get property name
    const SQChar* key;
    sq_getstring(vm, 2, &key);
    BIND_GET_THIS_PROPERTY(io::Xml)
    // dispatch appropriate getter
    if(!strcmp(key, "textfield")) {
        std::string ret = obj->textfield();
        BIND_PUSH_STRING(ret)
        return 1;
    }
    if(!strcmp(key, "verbose")) {
        SQBool ret = obj->verbose();
        BIND_PUSH_BOOL(ret)
        return 1;
    }
    if(!strcmp(key, "writedefault")) {
        SQBool ret = obj->writedefault();
        BIND_PUSH_BOOL(ret)
        return 1;
    }
    }
    sq_pushnull(vm);
    return sq_throwobject(vm);
}
#undef BIND_METHOD_NAME

//
// IO.Xml property setters
//
#define BIND_METHOD_NAME "setter"
SQInteger XmlSetters(HSQUIRRELVM vm)
{
    SQObjectType argtype = sq_gettype(vm, 2);
    if(argtype == OT_STRING) {
    // get property name
    const SQChar* key;
    sq_getstring(vm, 2, &key);
    BIND_GET_THIS_PROPERTY(io::Xml)
    // dispatch appropriate setter
    if(!strcmp(key, "textfield")) {
        // get parameter 2 "textfield" as string
        BIND_GET_CHECK_STRING(textfield, 2)
        BIND_WRAP_TRY(obj->textfield(textfield))
        return 0;
    }
    if(!strcmp(key, "verbose")) {
        // get parameter 2 "verbose" as bool
        BIND_GET_CHECK_BOOL(verbose, 2)
        BIND_WRAP_TRY(obj->verbose(verbose))
        return 0;
    }
    if(!strcmp(key, "writedefault")) {
        // get parameter 2 "writedefault" as bool
        BIND_GET_CHECK_BOOL(writedefault, 2)
        BIND_WRAP_TRY(obj->writedefault(writedefault))
        return 0;
    }
    }
    sq_pushnull(vm);
    return sq_throwobject(vm);
}
#undef BIND_METHOD_NAME

//
// IO.Xml read
//
#define BIND_METHOD_NAME "read"
SQInteger Xmlread(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    BIND_GET_THIS_INSTANCE(Xml)
    BIND_NARGS_CHECK_MIN(1) // check min parameter count
    BIND_NARGS_CHECK_MAX(2) // check max parameter count
    HSQOBJECT arg1;
    sq_getstackobj(vm, 2, &arg1);
    if(arg1._type == OT_STRING) {
        // get parameter 1 "json" as string
        BIND_GET_STRING(json, 1)
        if(numargs == 2) {
            // call the implementation
            BIND_WRAP_TRY(obj->read(vm, json))
        } else {
            // get parameter 2 "clazz" as class
            BIND_GET_CHECK_CLASS(clazz, 2)
            // call the implementation
            BIND_WRAP_TRY(obj->read(vm, json, clazz))
        }
    }
    else if(arg1._type == OT_INSTANCE) {
    SQUserPointer arg1type = bip_getinstancetypetag(&arg1);
    if(arg1type == &FileObject) {
        // get parameter 1 "file" as IO.File
        BIND_GET_OBJECT(file, 1, IO, File, io, File)
        if(numargs == 2) {
            // call the implementation
            BIND_WRAP_TRY(obj->read(vm, *file))
        } else {
            // get parameter 2 "clazz" as class
            BIND_GET_CHECK_CLASS(clazz, 2)
            // call the implementation
            BIND_WRAP_TRY(obj->read(vm, *file, clazz))
        }
    }
    else {
        return sq_throwerror(vm, "argument 1 is not of type {string, IO.File}");
    }
    }
    else {
        return sq_throwerror(vm, "argument 1 is not of type {string, IO.File}");
    }
    return 1;
}
#undef BIND_METHOD_NAME

//
// IO.Xml write
//
#define BIND_METHOD_NAME "write"
SQInteger Xmlwrite(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    BIND_GET_THIS_INSTANCE(Xml)
    BIND_NARGS_CHECK_MIN(2) // check min parameter count
    // get parameter 1 "object" as any
    BIND_GET_CHECK_ANY(object, 1)
    // get parameter 2 "root" as string
    BIND_GET_CHECK_STRING(root, 2)
    // return value
    std::string ret;
    if(numargs == 3) {
        // call the implementation
        BIND_WRAP_TRY(ret = obj->write(object, root))
    } else {
        BIND_NARGS_CHECK_MIN(3) // check min parameter count
        BIND_NARGS_CHECK_MAX(3) // check max parameter count
        // get parameter 3 "file" as IO.File
        BIND_GET_CHECK_OBJECT(file, 3, IO, File, io, File)
        // call the implementation
        BIND_WRAP_TRY(obj->write(object, root, *file))
        // void method, returns no value
        return 0;

    }
    // push return value
    BIND_PUSH_STRING(ret)
    return 1;
}
#undef BIND_METHOD_NAME

SQInteger XmlToString(HSQUIRRELVM vm)
{
    sq_pushstring(vm, "(IO.Xml)", 8);
    return 1;
}
#undef BIND_CLASS_NAME

//
// IO.Zip class
//
#define BIND_CLASS_NAME "Zip"
#define BIND_METHOD_NAME "constructor"
SQInteger ZipCtor(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    Zip *obj;
    BIND_NARGS_CHECK_EXACT(1) // check exact parameter count
    // get parameter 1 "path" as string
    BIND_GET_CHECK_STRING(path, 1)
    // call the implementation
    BIND_WRAP_TRY(obj = new Zip(path))
    // return pointer to new object
    sq_setinstanceup(vm, 1, (SQUserPointer*)obj->addref());
    sq_setreleasehook(vm, 1, BindUtil::Decrement<Zip>);
    return 1;
}
#undef BIND_METHOD_NAME

//
// IO.Zip property getters
//
#define BIND_METHOD_NAME "getter"
SQInteger ZipGetters(HSQUIRRELVM vm)
{
    SQObjectType argtype = sq_gettype(vm, 2);
    if(argtype == OT_INTEGER) {
        SQInteger index;
        sq_getinteger(vm, 2, &index);
        BIND_GET_THIS_PROPERTY(io::Zip)
        io::Path* ret = (*obj)[index];
        if(ret) {
            io::binding::pushPath(vm, ret);
            return 1;
        }
    }
    sq_pushnull(vm);
    return sq_throwobject(vm);
}
#undef BIND_METHOD_NAME
//
// IO.Zip _nexti
//
SQInteger ZipNexti(HSQUIRRELVM vm)
{
    SQObjectType argtype = sq_gettype(vm, 2);
	BIND_GET_THIS_PROPERTY(Zip)
    if (argtype == OT_NULL) {
        if(obj->len()) {
            sq_pushinteger(vm, 0);
        }
		return 1;
    } else if (argtype == OT_INTEGER) {
        SQInteger index;
        sq_getinteger(vm, 2, &index);
        if ((size_t)index + 1 < obj->len()) {
            sq_pushinteger(vm, index + 1);
            return 1;
        }
    }
    sq_pushnull(vm);
	return 1;
}
//
// IO.Zip entries
//
#define BIND_METHOD_NAME "entries"
SQInteger Zipentries(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    BIND_GET_THIS_INSTANCE(Zip)
    BIND_NARGS_CHECK_EXACT(0) // check exact parameter count
    // return value
    SQInteger ret;
    // call the implementation
    BIND_WRAP_TRY(ret = obj->entries())
    // push return value
    BIND_PUSH_INTEGER(ret)
    return 1;
}
#undef BIND_METHOD_NAME

//
// IO.Zip entry
//
#define BIND_METHOD_NAME "entry"
SQInteger Zipentry(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    BIND_GET_THIS_INSTANCE(Zip)
    BIND_NARGS_CHECK_EXACT(1) // check exact parameter count
    // get parameter 1 "path" as string
    BIND_GET_CHECK_STRING(path, 1)
    // return value
    io::Path* ret;
    // call the implementation
    BIND_WRAP_TRY(ret = obj->entry(path))
    // push return value
    io::binding::pushPath(vm, ret);
    return 1;
}
#undef BIND_METHOD_NAME

SQInteger ZipToString(HSQUIRRELVM vm)
{
    sq_pushstring(vm, "(IO.Zip)", 8);
    return 1;
}
#undef BIND_CLASS_NAME

//
// IO.ZipFile class
//
#define BIND_CLASS_NAME "ZipFile"
SQInteger ZipFileCtor(HSQUIRRELVM vm)
{
    return sq_throwerror(vm, "cannot directly instantiate IO.ZipFile");
}

SQInteger ZipFileNoclone(HSQUIRRELVM vm)
{
    return sq_throwerror(vm, "cannot clone objects of type IO.ZipFile");
}
//
// IO.ZipFile property getters
//
#define BIND_METHOD_NAME "getter"
SQInteger ZipFileGetters(HSQUIRRELVM vm)
{
    SQObjectType argtype = sq_gettype(vm, 2);
    if(argtype == OT_STRING) {
    // get property name
    const SQChar* key;
    sq_getstring(vm, 2, &key);
    BIND_GET_THIS_PROPERTY(io::ZipFile)
    // dispatch appropriate getter
    if(!strcmp(key, "name")) {
        std::string ret = obj->name();
        BIND_PUSH_STRING(ret)
        return 1;
    }
    if(!strcmp(key, "path")) {
        std::string ret = obj->path();
        BIND_PUSH_STRING(ret)
        return 1;
    }
    }
    sq_pushnull(vm);
    return sq_throwobject(vm);
}
#undef BIND_METHOD_NAME

//
// IO.ZipFile property setters
//
#define BIND_METHOD_NAME "setter"
SQInteger ZipFileSetters(HSQUIRRELVM vm)
{
    SQObjectType argtype = sq_gettype(vm, 2);
    if(argtype == OT_STRING) {
    // get property name
    const SQChar* key;
    sq_getstring(vm, 2, &key);
    BIND_GET_THIS_PROPERTY(io::ZipFile)
    // dispatch appropriate setter
    if(!strcmp(key, "name")) {
        return sq_throwerror(vm, "property \"name\" is read-only");
    }
    if(!strcmp(key, "path")) {
        return sq_throwerror(vm, "property \"path\" is read-only");
    }
    }
    sq_pushnull(vm);
    return sq_throwobject(vm);
}
#undef BIND_METHOD_NAME

//
// IO.ZipFile bytes
//
#define BIND_METHOD_NAME "bytes"
SQInteger ZipFilebytes(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    BIND_GET_THIS_INSTANCE(ZipFile)
    BIND_NARGS_CHECK_EXACT(0) // check exact parameter count
    // return value
    bipscript::Buffer* ret;
    // call the implementation
    BIND_WRAP_TRY(ret = obj->bytes())
    // push return value
    BIND_PUSH_OBJECT(ret->addref(), buffer, bipscript, Buffer, Decrement)
    return 1;
}
#undef BIND_METHOD_NAME

//
// IO.ZipFile exists
//
#define BIND_METHOD_NAME "exists"
SQInteger ZipFileexists(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    BIND_GET_THIS_INSTANCE(ZipFile)
    BIND_NARGS_CHECK_EXACT(0) // check exact parameter count
    // return value
    SQBool ret;
    // call the implementation
    BIND_WRAP_TRY(ret = obj->exists())
    // push return value
    BIND_PUSH_BOOL(ret)
    return 1;
}
#undef BIND_METHOD_NAME

//
// IO.ZipFile folder
//
#define BIND_METHOD_NAME "folder"
SQInteger ZipFilefolder(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    BIND_GET_THIS_INSTANCE(ZipFile)
    BIND_NARGS_CHECK_EXACT(0) // check exact parameter count
    // return value
    SQBool ret;
    // call the implementation
    BIND_WRAP_TRY(ret = obj->folder())
    // push return value
    BIND_PUSH_BOOL(ret)
    return 1;
}
#undef BIND_METHOD_NAME

//
// IO.ZipFile parent
//
#define BIND_METHOD_NAME "parent"
SQInteger ZipFileparent(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    BIND_GET_THIS_INSTANCE(ZipFile)
    BIND_NARGS_CHECK_EXACT(0) // check exact parameter count
    // return value
    io::Folder* ret;
    // call the implementation
    BIND_WRAP_TRY(ret = obj->parent())
    // push return value
    BIND_PUSH_OBJECT(ret, Folder, io, Folder, Delete)
    return 1;
}
#undef BIND_METHOD_NAME

//
// IO.ZipFile size
//
#define BIND_METHOD_NAME "size"
SQInteger ZipFilesize(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    BIND_GET_THIS_INSTANCE(ZipFile)
    BIND_NARGS_CHECK_EXACT(0) // check exact parameter count
    // return value
    SQInteger ret;
    // call the implementation
    BIND_WRAP_TRY(ret = obj->size())
    // push return value
    BIND_PUSH_INTEGER(ret)
    return 1;
}
#undef BIND_METHOD_NAME

//
// IO.ZipFile text
//
#define BIND_METHOD_NAME "text"
SQInteger ZipFiletext(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    BIND_GET_THIS_INSTANCE(ZipFile)
    BIND_NARGS_CHECK_EXACT(0) // check exact parameter count
    // return value
    const SQChar* ret;
    // call the implementation
    BIND_WRAP_TRY(ret = obj->text())
    // push return value
    BIND_PUSH_CSTRING(ret)
    return 1;
}
#undef BIND_METHOD_NAME

SQInteger ZipFileToString(HSQUIRRELVM vm)
{
    sq_pushstring(vm, "(IO.ZipFile)", 12);
    return 1;
}
#undef BIND_CLASS_NAME

//
// IO.ZipFolder class
//
#define BIND_CLASS_NAME "ZipFolder"
SQInteger ZipFolderCtor(HSQUIRRELVM vm)
{
    return sq_throwerror(vm, "cannot directly instantiate IO.ZipFolder");
}

SQInteger ZipFolderNoclone(HSQUIRRELVM vm)
{
    return sq_throwerror(vm, "cannot clone objects of type IO.ZipFolder");
}
//
// IO.ZipFolder property getters
//
#define BIND_METHOD_NAME "getter"
SQInteger ZipFolderGetters(HSQUIRRELVM vm)
{
    SQObjectType argtype = sq_gettype(vm, 2);
    if(argtype == OT_STRING) {
    // get property name
    const SQChar* key;
    sq_getstring(vm, 2, &key);
    BIND_GET_THIS_PROPERTY(io::ZipFolder)
    // dispatch appropriate getter
    if(!strcmp(key, "name")) {
        std::string ret = obj->name();
        BIND_PUSH_STRING(ret)
        return 1;
    }
    if(!strcmp(key, "path")) {
        std::string ret = obj->path();
        BIND_PUSH_STRING(ret)
        return 1;
    }
    }
    sq_pushnull(vm);
    return sq_throwobject(vm);
}
#undef BIND_METHOD_NAME

//
// IO.ZipFolder property setters
//
#define BIND_METHOD_NAME "setter"
SQInteger ZipFolderSetters(HSQUIRRELVM vm)
{
    SQObjectType argtype = sq_gettype(vm, 2);
    if(argtype == OT_STRING) {
    // get property name
    const SQChar* key;
    sq_getstring(vm, 2, &key);
    BIND_GET_THIS_PROPERTY(io::ZipFolder)
    // dispatch appropriate setter
    if(!strcmp(key, "name")) {
        return sq_throwerror(vm, "property \"name\" is read-only");
    }
    if(!strcmp(key, "path")) {
        return sq_throwerror(vm, "property \"path\" is read-only");
    }
    }
    sq_pushnull(vm);
    return sq_throwobject(vm);
}
#undef BIND_METHOD_NAME

//
// IO.ZipFolder exists
//
#define BIND_METHOD_NAME "exists"
SQInteger ZipFolderexists(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    BIND_GET_THIS_INSTANCE(ZipFolder)
    BIND_NARGS_CHECK_EXACT(0) // check exact parameter count
    // return value
    SQBool ret;
    // call the implementation
    BIND_WRAP_TRY(ret = obj->exists())
    // push return value
    BIND_PUSH_BOOL(ret)
    return 1;
}
#undef BIND_METHOD_NAME

//
// IO.ZipFolder folder
//
#define BIND_METHOD_NAME "folder"
SQInteger ZipFolderfolder(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    BIND_GET_THIS_INSTANCE(ZipFolder)
    BIND_NARGS_CHECK_EXACT(0) // check exact parameter count
    // return value
    SQBool ret;
    // call the implementation
    BIND_WRAP_TRY(ret = obj->folder())
    // push return value
    BIND_PUSH_BOOL(ret)
    return 1;
}
#undef BIND_METHOD_NAME

//
// IO.ZipFolder parent
//
#define BIND_METHOD_NAME "parent"
SQInteger ZipFolderparent(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    BIND_GET_THIS_INSTANCE(ZipFolder)
    BIND_NARGS_CHECK_EXACT(0) // check exact parameter count
    // return value
    io::Folder* ret;
    // call the implementation
    BIND_WRAP_TRY(ret = obj->parent())
    // push return value
    BIND_PUSH_OBJECT(ret, Folder, io, Folder, Delete)
    return 1;
}
#undef BIND_METHOD_NAME

SQInteger ZipFolderToString(HSQUIRRELVM vm)
{
    sq_pushstring(vm, "(IO.ZipFolder)", 14);
    return 1;
}
#undef BIND_CLASS_NAME

SQInteger IONewSlot(HSQUIRRELVM vm)
{
	return sq_throwerror(vm, "package table does not support the new slot operator");
}

void bindIO(HSQUIRRELVM vm)
{
    BindUtil binding(vm);
    binding.startTable("IO", 2);

    // create enumeration
    sq_pushstring(vm, "Type", -1);
    sq_newtable(vm);
    sq_pushstring(vm, "Float", -1);
    sq_pushinteger(vm, Type::Float);
    sq_newslot(vm, -3, true);
    sq_pushstring(vm, "Integer", -1);
    sq_pushinteger(vm, Type::Integer);
    sq_newslot(vm, -3, true);
    sq_pushstring(vm, "String", -1);
    sq_pushinteger(vm, Type::String);
    sq_newslot(vm, -3, true);
    sq_newslot(vm, -3, true);
    // create class IO.File
    binding.startClass("File", 4, FileObject);
    binding.addMethod("constructor", 11, &FileCtor); // ctor
    binding.addMethod("_cloned", 7, BindUtil::Clone<File>); // clone
    binding.addMethod("_get", 4, &FileGetters);
    binding.addMethod("_set", 4, &FileSetters);
    binding.addMethod("bytes", 5, &Filebytes);
    binding.addMethod("exists", 6, &Fileexists);
    binding.addMethod("folder", 6, &Filefolder);
    binding.addMethod("len", 3, &Filelen);
    binding.addMethod("parent", 6, &Fileparent);
    binding.addMethod("readAll", 7, &FilereadAll);
    binding.addMethod("reader", 6, &Filereader);
    binding.addMethod("remove", 6, &Fileremove);
    binding.addMethod("text", 4, &Filetext);
    binding.addMethod("writeAll", 8, &FilewriteAll);
    binding.addMethod("writer", 6, &Filewriter);
    binding.addMethod("_tostring", 9, &FileToString);
    binding.endClass();

    // create class IO.Folder
    binding.startClass("Folder", 6, FolderObject);
    binding.addMethod("constructor", 11, &FolderCtor); // ctor
    binding.addMethod("_cloned", 7, BindUtil::Clone<Folder>); // clone
    binding.addMethod("_get", 4, &FolderGetters);
    binding.addMethod("_set", 4, &FolderSetters);
    binding.addMethod("children", 8, &Folderchildren);
    binding.addMethod("exists", 6, &Folderexists);
    binding.addMethod("folder", 6, &Folderfolder);
    binding.addMethod("mkdir", 5, &Foldermkdir);
    binding.addMethod("parent", 6, &Folderparent);
    binding.addMethod("_tostring", 9, &FolderToString);
    binding.endClass();

    // create class IO.Json
    binding.startClass("Json", 4, JsonObject);
    binding.addMethod("constructor", 11, &JsonCtor); // ctor
    binding.addMethod("_cloned", 7, BindUtil::Clone<Json>); // clone
    binding.addMethod("_get", 4, &JsonGetters);
    binding.addMethod("_set", 4, &JsonSetters);
    binding.addMethod("read", 4, &Jsonread);
    binding.addMethod("write", 5, &Jsonwrite);
    binding.addMethod("_tostring", 9, &JsonToString);
    binding.endClass();

    // create class IO.Xml
    binding.startClass("Xml", 3, XmlObject);
    binding.addMethod("constructor", 11, &XmlCtor); // ctor
    binding.addMethod("_cloned", 7, BindUtil::Clone<Xml>); // clone
    binding.addMethod("_get", 4, &XmlGetters);
    binding.addMethod("_set", 4, &XmlSetters);
    binding.addMethod("read", 4, &Xmlread);
    binding.addMethod("write", 5, &Xmlwrite);
    binding.addMethod("_tostring", 9, &XmlToString);
    binding.endClass();

    // create class IO.Zip
    binding.startClass("Zip", 3, ZipObject);
    binding.addMethod("constructor", 11, &ZipCtor); // ctor
    binding.addMethod("_cloned", 7, BindUtil::Clone<Zip>); // clone
    binding.addMethod("_get", 4, &ZipGetters);
    binding.addMethod("_nexti", 6, &ZipNexti);
    binding.addMethod("entries", 7, &Zipentries);
    binding.addMethod("entry", 5, &Zipentry);
    binding.addMethod("_tostring", 9, &ZipToString);
    binding.endClass();

    // create class IO.ZipFile
    binding.startClass("ZipFile", 7, ZipFileObject);
    binding.addMethod("constructor", 11, &ZipFileCtor); // ctor
    binding.addMethod("_cloned", 7, &ZipFileNoclone); // unclonable
    binding.addMethod("_get", 4, &ZipFileGetters);
    binding.addMethod("_set", 4, &ZipFileSetters);
    binding.addMethod("bytes", 5, &ZipFilebytes);
    binding.addMethod("exists", 6, &ZipFileexists);
    binding.addMethod("folder", 6, &ZipFilefolder);
    binding.addMethod("parent", 6, &ZipFileparent);
    binding.addMethod("size", 4, &ZipFilesize);
    binding.addMethod("text", 4, &ZipFiletext);
    binding.addMethod("_tostring", 9, &ZipFileToString);
    binding.endClass();

    // create class IO.ZipFolder
    binding.startClass("ZipFolder", 9, ZipFolderObject);
    binding.addMethod("constructor", 11, &ZipFolderCtor); // ctor
    binding.addMethod("_cloned", 7, &ZipFolderNoclone); // unclonable
    binding.addMethod("_get", 4, &ZipFolderGetters);
    binding.addMethod("_set", 4, &ZipFolderSetters);
    binding.addMethod("exists", 6, &ZipFolderexists);
    binding.addMethod("folder", 6, &ZipFolderfolder);
    binding.addMethod("parent", 6, &ZipFolderparent);
    binding.addMethod("_tostring", 9, &ZipFolderToString);
    binding.endClass();

    sq_newtable(vm); // delegate table for package properties
    binding.addMethod("_newslot", 8, &IONewSlot);
    sq_setdelegate(vm, -2);

    binding.endTable(); // end package table
}
}}}
