/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "bindscript.h"
#include "bindinterface.h"
#include "bindutil.h"

#include "scripthost.h"
#include "script.h"
#include <stdexcept>
#include <cstring>

namespace bipscript {

namespace script {

namespace binding {

// object references to types in this package
HSQOBJECT FunctionEventObject;

#define BIND_CLASS_NAME "Script"
//
// Script clear
//
#define BIND_METHOD_NAME "clear"
SQInteger Scriptclear(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    BIND_NARGS_CHECK_MAX(1) // check max parameter count
    if(numargs == 1) {
        // call the implementation
        BIND_WRAP_TRY(Script::instance().clear())
    } else {
        // get parameter 1 "clock" as Time.Clock
        BIND_GET_CHECK_INTERFACE(clock, 1, Time, Clock, time, Clock)
        // call the implementation
        BIND_WRAP_TRY(Script::instance().clear(*clock))
    }
    // void method, returns no value
    return 0;
}
#undef BIND_METHOD_NAME

//
// Script exit
//
#define BIND_METHOD_NAME "exit"
SQInteger Scriptexit(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    BIND_NARGS_CHECK_EXACT(1) // check exact parameter count
    // get parameter 1 "code" as integer
    BIND_GET_CHECK_INTEGER(code, 1)
    // call the implementation
    BIND_WRAP_TRY(Script::instance().exitScript(code))
    // void method, returns no value
    return 0;
}
#undef BIND_METHOD_NAME

//
// Script importfile
//
#define BIND_METHOD_NAME "importfile"
SQInteger Scriptimportfile(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    BIND_NARGS_CHECK_EXACT(1) // check exact parameter count
    // get parameter 1 "path" as IO.Path
    BIND_GET_CHECK_INTERFACE(path, 1, IO, Path, io, Path)
    // call the implementation
    BIND_WRAP_TRY(Script::instance().import(*path))
    // void method, returns no value
    return 0;
}
#undef BIND_METHOD_NAME

//
// Script reload
//
#define BIND_METHOD_NAME "reload"
SQInteger Scriptreload(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    BIND_NARGS_CHECK_EXACT(0) // check exact parameter count
    // call the implementation
    BIND_WRAP_TRY(ScriptHost::instance().requestRestart())
    // void method, returns no value
    return 0;
}
#undef BIND_METHOD_NAME

//
// Script remove
//
#define BIND_METHOD_NAME "remove"
SQInteger Scriptremove(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    BIND_NARGS_CHECK_EXACT(1) // check exact parameter count
    // get parameter 1 "event" as Script.FunctionEvent
    BIND_GET_CHECK_OBJECT(event, 1, Script, FunctionEvent, script, FunctionEvent)
    // call the implementation
    BIND_WRAP_TRY(Script::instance().remove(*event))
    // void method, returns no value
    return 0;
}
#undef BIND_METHOD_NAME

//
// Script resolve
//
#define BIND_METHOD_NAME "resolve"
SQInteger Scriptresolve(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    BIND_NARGS_CHECK_EXACT(1) // check exact parameter count
    // get parameter 1 "path" as string
    BIND_GET_CHECK_STRING(path, 1)
    // return value
    io::Path* ret;
    // call the implementation
    BIND_WRAP_TRY(ret = ScriptHost::instance().resolve(path))
    // push return value
    io::binding::pushPath(vm, ret);
    return 1;
}
#undef BIND_METHOD_NAME

//
// Script schedule
//
#define BIND_METHOD_NAME "schedule"
SQInteger Scriptschedule(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    BIND_NARGS_CHECK_MIN(2) // check min parameter count
    // get parameter 1 "function" as function
    BIND_GET_CHECK_FUNCTION(function, 1)
    // return value
    script::FunctionEvent* ret;
    HSQOBJECT arg2;
    sq_getstackobj(vm, 3, &arg2);
    if(BIND_CHECK_FLOAT(2)) {
        BIND_NARGS_CHECK_MAX(3) // check max parameter count
        // get parameter 2 "measure" as float
        BIND_GET_FLOAT(measure, 2)
        if(numargs == 3) {
            // call the implementation
            BIND_WRAP_TRY(ret = Script::instance().schedule(function, measure))
        } else {
            // get parameter 3 "clock" as Time.Clock
            BIND_GET_CHECK_INTERFACE(clock, 3, Time, Clock, time, Clock)
            // call the implementation
            BIND_WRAP_TRY(ret = Script::instance().schedule(function, measure, *clock))
        }
    }
    else if(BIND_CHECK_ANY(2)) {
        BIND_NARGS_CHECK_MIN(3) // check min parameter count
        // get parameter 2 "userdata" as any
        BIND_GET_ANY(userdata, 2)
        HSQOBJECT arg3;
        sq_getstackobj(vm, 4, &arg3);
        if(BIND_CHECK_FLOAT(3)) {
            BIND_NARGS_CHECK_MAX(4) // check max parameter count
            // get parameter 3 "measure" as float
            BIND_GET_FLOAT(measure, 3)
            if(numargs == 4) {
                // call the implementation
                BIND_WRAP_TRY(ret = Script::instance().schedule(function, userdata, measure))
            } else {
                // get parameter 4 "clock" as Time.Clock
                BIND_GET_CHECK_INTERFACE(clock, 4, Time, Clock, time, Clock)
                // call the implementation
                BIND_WRAP_TRY(ret = Script::instance().schedule(function, userdata, measure, *clock))
            }
        }
        else if(arg3._type == OT_INSTANCE) {
            BIND_NARGS_CHECK_MIN(4) // check min parameter count
            BIND_NARGS_CHECK_MAX(5) // check max parameter count
            // get parameter 3 "context" as instance
            BIND_GET_INSTANCE(context, 3)
            // get parameter 4 "measure" as float
            BIND_GET_CHECK_FLOAT(measure, 4)
            if(numargs == 5) {
                // call the implementation
                BIND_WRAP_TRY(ret = Script::instance().schedule(function, userdata, context, measure))
            } else {
                // get parameter 5 "clock" as Time.Clock
                BIND_GET_CHECK_INTERFACE(clock, 5, Time, Clock, time, Clock)
                // call the implementation
                BIND_WRAP_TRY(ret = Script::instance().schedule(function, userdata, context, measure, *clock))
            }
        }
        else {
            return sq_throwerror(vm, "argument 3 is not of type {float, instance}");
        }
    }
    else {
        return sq_throwerror(vm, "argument 2 is not of type {float, any}");
    }
    // push return value
    BIND_PUSH_OBJECT(ret->addref(), FunctionEvent, script, FunctionEvent, Decrement)
    return 1;
}
#undef BIND_METHOD_NAME

//
// Script stayAlive
//
#define BIND_METHOD_NAME "stayAlive"
SQInteger ScriptstayAlive(HSQUIRRELVM vm)
{
    BIND_GET_NARGS()
    BIND_NARGS_CHECK_EXACT(0) // check exact parameter count
    // call the implementation
    BIND_WRAP_TRY(Script::stayAlive())
    // void method, returns no value
    return 0;
}
#undef BIND_METHOD_NAME

//
// Script package property getters
//
#define BIND_METHOD_NAME "getter"
SQInteger ScriptGetters(HSQUIRRELVM vm)
{
    // get property name
    const SQChar* key;
    sq_getstring(vm, 2, &key);
    if(!strcmp(key, "argument")) {
        HSQOBJECT ret = script::Script::argument();
        sq_pushobject(vm, ret);
        return 1;
    }
    if(!strcmp(key, "argv")) {
        HSQOBJECT ret = script::Script::argv();
        sq_pushobject(vm, ret);
        return 1;
    }
    sq_pushnull(vm);
    return sq_throwobject(vm);
}
#undef BIND_METHOD_NAME

//
// Script property setters
//
#define BIND_METHOD_NAME "setter"
SQInteger ScriptSetters(HSQUIRRELVM vm)
{
    // get property name
    const SQChar* key;
    sq_getstring(vm, 2, &key);
    if(!strcmp(key, "argument")) {
        return sq_throwerror(vm, "property \"argument\" is read-only");
    }
    if(!strcmp(key, "argv")) {
        return sq_throwerror(vm, "property \"argv\" is read-only");
    }
    sq_pushnull(vm);
    return sq_throwobject(vm);
}
#undef BIND_METHOD_NAME

#undef BIND_CLASS_NAME
//
// Script.FunctionEvent class
//
#define BIND_CLASS_NAME "FunctionEvent"
SQInteger FunctionEventCtor(HSQUIRRELVM vm)
{
    return sq_throwerror(vm, "cannot directly instantiate Script.FunctionEvent");
}

SQInteger FunctionEventNoclone(HSQUIRRELVM vm)
{
    return sq_throwerror(vm, "cannot clone objects of type Script.FunctionEvent");
}
//
// Script.FunctionEvent property getters
//
#define BIND_METHOD_NAME "getter"
SQInteger FunctionEventGetters(HSQUIRRELVM vm)
{
    SQObjectType argtype = sq_gettype(vm, 2);
    if(argtype == OT_STRING) {
    // get property name
    const SQChar* key;
    sq_getstring(vm, 2, &key);
    BIND_GET_THIS_PROPERTY(script::FunctionEvent)
    // dispatch appropriate getter
    if(!strcmp(key, "measure")) {
        SQFloat ret = obj->measure();
        BIND_PUSH_FLOAT(ret)
        return 1;
    }
    }
    sq_pushnull(vm);
    return sq_throwobject(vm);
}
#undef BIND_METHOD_NAME

//
// Script.FunctionEvent property setters
//
#define BIND_METHOD_NAME "setter"
SQInteger FunctionEventSetters(HSQUIRRELVM vm)
{
    SQObjectType argtype = sq_gettype(vm, 2);
    if(argtype == OT_STRING) {
    // get property name
    const SQChar* key;
    sq_getstring(vm, 2, &key);
    BIND_GET_THIS_PROPERTY(script::FunctionEvent)
    // dispatch appropriate setter
    if(!strcmp(key, "measure")) {
        return sq_throwerror(vm, "property \"measure\" is read-only");
    }
    }
    sq_pushnull(vm);
    return sq_throwobject(vm);
}
#undef BIND_METHOD_NAME

SQInteger FunctionEventToString(HSQUIRRELVM vm)
{
    sq_pushstring(vm, "(Script.FunctionEvent)", 22);
    return 1;
}
#undef BIND_CLASS_NAME

SQInteger ScriptNewSlot(HSQUIRRELVM vm)
{
	return sq_throwerror(vm, "package table does not support the new slot operator");
}

void bindScript(HSQUIRRELVM vm)
{
    BindUtil binding(vm);
    binding.startTable("Script", 6);

    // static methods
    binding.addMethod("clear", 5, &Scriptclear);
    binding.addMethod("exit", 4, &Scriptexit);
    binding.addMethod("importfile", 10, &Scriptimportfile);
    binding.addMethod("schedule", 8, &Scriptschedule);
    binding.addMethod("schedule", 8, &Scriptschedule);
    binding.addMethod("schedule", 8, &Scriptschedule);
    binding.addMethod("remove", 6, &Scriptremove);
    binding.addMethod("resolve", 7, &Scriptresolve);
    binding.addMethod("reload", 6, &Scriptreload);
    binding.addMethod("stayAlive", 9, &ScriptstayAlive);

    // create class Script.FunctionEvent
    binding.startClass("FunctionEvent", 13, FunctionEventObject);
    binding.addMethod("constructor", 11, &FunctionEventCtor); // ctor
    binding.addMethod("_cloned", 7, &FunctionEventNoclone); // unclonable
    binding.addMethod("_get", 4, &FunctionEventGetters);
    binding.addMethod("_set", 4, &FunctionEventSetters);
    binding.addMethod("_tostring", 9, &FunctionEventToString);
    binding.endClass();

    sq_newtable(vm); // delegate table for package properties
    binding.addMethod("_get", 4, &ScriptGetters);
    binding.addMethod("_set", 4, &ScriptSetters);
    binding.addMethod("_newslot", 8, &ScriptNewSlot);
    sq_setdelegate(vm, -2);

    binding.endTable(); // end package table
}
}}}
