/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef BIPBUFFER_H
#define BIPBUFFER_H

#include <algorithm>
#include <assert.h>
#include <cstdint>
#include <math.h>
#include <string.h>
#include <string>

class String : public std::string {
  public:
    String(double) {}
    String(const char *) {}
};

class AudioSampleBuffer {
    uint32_t mchannels;
    uint32_t msize;
    const uint32_t maxsize;
    float **mdata;
    bool mallocate;

  public:
    AudioSampleBuffer(uint32_t channels, uint32_t size, bool allocate = true)
        : mchannels(channels), msize(size), maxsize(size), mallocate(allocate) {
        mdata = new float *[channels];
        if (mallocate) {
            for (uint32_t c = 0; c < mchannels; c++) {
                mdata[c] = new float[msize];
            }
        }
    }
    ~AudioSampleBuffer() {
        if(mallocate) {
            for (uint32_t c = 0; c < mchannels; c++) {
                delete[] mdata[c];
            }
        }
        delete[] mdata;
    }
    void operator=(const AudioSampleBuffer &other) {
        assert(maxsize >= other.msize); // don't need to resize buffers
        assert(mchannels >= other.mchannels);
        mchannels = other.mchannels;
        msize = other.msize;
        for (uint32_t c = 0; c < mchannels; c++) {
            memcpy(mdata[c], other.mdata[c], msize * sizeof(float));
        }
    }
    void data(uint32_t channel, float *data) { mdata[channel] = data; }
    void setSize(uint32_t channels, uint32_t size) {
        assert(maxsize >= size); // don't need to resize buffers
        assert(mchannels >= channels);
        mchannels = channels;
        msize = size;
    }
    int getNumSamples() const { return msize; }
    int getNumChannels() const { return mchannels; }
    float getMagnitude(uint32_t start, uint32_t size) const {
        float max = 0.0;
        for (uint32_t c = 0; c < mchannels; c++) {
            for (uint32_t i = 0; i < size; i++) {
                float val = mdata[c][start + i];
                if (abs(val) > max) {
                    max = val;
                }
            }
        }
        return max;
    }
    float *getWritePointer(int channel) { return mdata[channel]; }
};

template <class T> class HeapBlock {
    T *data;
    size_t msize;

  public:
    HeapBlock() : data(nullptr), msize(0) {}
    ~HeapBlock() { delete[] data; }
    void calloc(int size) {
        if(size > msize) {
            delete[] data;
            data = new T[size];
            msize = size;
        }
        std::fill(data, data + size, T());
    }
    T &operator[](int index) { return data[index]; }
    void clear(int size) {
        for (int i = 0; i < size; i++) {
            data[i] = T();
        }
    }
};

#define DEB(text)
#define DBG(text)

#define jmax std::max
#define jassert assert

template <typename T> T min(T x, T y) { return std::min(x, y); }

template <typename T, typename... Args> T min(T x, T y, Args... args) {
    return min(std::min(x, y), args...);
}

#define jmin min<int>

#endif // BIPBUFFER_H
