﻿/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef AUDIOCONNECTION_H
#define AUDIOCONNECTION_H

#include "audioengine.h"
#include "bindaudio.h"
#include "boundarray.h"
#include "buffer.h"
#include "clip.h"
#include "methodqueue.h"
#include "source.h"
#include "statusserver.h"
#include "timeline.h"

namespace bipscript {
namespace audio {

class Source;

class Meter {
    float mmaxlevel;

  public:
    static const samples_t REPORT_PERIOD = 3675; // 44100
    Meter() : mmaxlevel(0.0) {}
    void findmax(float *buffer, samples_t size) {
        for (samples_t i = 0; i < size; i++) {
            float abs = std::abs(buffer[i]);
            if (abs > mmaxlevel) {
                mmaxlevel = abs;
            }
        }
    }
    float maxlevel(float *buffer, samples_t size) {
        findmax(buffer, size);
        float ret = mmaxlevel;
        mmaxlevel = 0.0;
        return ret;
    }
};

/**
 * Represents a mono audio connection
 */
class Connection {
    static float *mdummyBuffer;
    static samples_t mbufferSize;
    Source *msource;
    float *buffer;
    bool allocate;
    Meter meter;

  public:
    // static
    static void bufferSize(samples_t size) {
        if (mdummyBuffer) {
            delete[] mdummyBuffer;
        }
        mdummyBuffer = new float[size];
        memset(mdummyBuffer, 0, size * sizeof(float));
        mbufferSize = size;
    }
    static samples_t bufferSize() { return mbufferSize; }
    static float *dummyBuffer() { return mdummyBuffer; }
    static void deallocate() { delete[] mdummyBuffer; }
    // instance
    Connection(Source *source, bool allocate) : msource(source), allocate(allocate) {
        if (allocate) {
            buffer = new float[mbufferSize];
            std::fill(buffer, buffer + mbufferSize, 0.0);
        }
    }
    ~Connection() {
        if (allocate) {
            delete[] buffer;
        }
    }
    void setBuffer(float *buffer) { this->buffer = buffer; }
    Source *source() { return msource; }
    float *audio() { return buffer; }
    void clear() { std::memset(buffer, 0, sizeof(float) * mbufferSize); }
    void findmax() { meter.findmax(buffer, mbufferSize); }
    float maxlevel() { return meter.maxlevel(buffer, mbufferSize); }
};

class ClipGrab : public Clip, public Listable<ClipGrab> {
    samples_t written;
    ScriptFunction mhandler;

  public:
    ClipGrab(uint32_t channels, uint64_t size, uint32_t framerate, ScriptFunction &handler)
        : Clip(channels, size, framerate), written(0), mhandler(handler) {}
    bool append(BoundArray<Connection> &bus, samples_t offset, samples_t nframes) {
        samples_t towrite = std::min(nframes, uint32_t(size() - written));
        for (uint32_t i = 0; i < channels(); i++) {
            buffer(i)->write(&bus[i].audio()[offset], written, towrite);
        }
        written += towrite;
        return written == size();
    }
    void clear() { written = 0; }
    ScriptFunction &callback() { return mhandler; }
};

class ClipGrabClosure : public ProcessClosure {
    Clip &mclip;

  protected:
    void addParameters() { BindUtil::Push(vm, new Clip(mclip), binding::ClipObject); }

  public:
    ClipGrabClosure(ClipGrab &clip) : ProcessClosure(clip.callback(), 0), mclip(clip) {}
};

typedef Event<ClipGrab> GrabEvent;

/**
 * Base for all classes that produce audio
 */
class Source : virtual public AbstractSource {
    BoundArray<Connection> bus;
    const uint32_t mchannels;
    // for status reporting
    samples_t lastreport;
    TimelineSet<ClipGrab> grabq;
    List<ClipGrab> activeGrabs;

  protected:
    Connection *audioOutput(uint32_t index) { return &bus[index]; }
    void audioOutput(Connection *connection, uint32_t index) {
        bus.set(index, connection, binding::ConnectionObject);
    }
    void grabClips(samples_t nframes) {
        ClipGrab *clip = activeGrabs.first();
        while (clip) {
            if (clip->append(bus, 0, nframes)) {
                MethodQueue::instance().dispatch(new ClipGrabClosure(*clip));
                ClipGrab *done = clip;
                clip = activeGrabs.next(clip);
                activeGrabs.remove(done);
            } else {
                clip = activeGrabs.next(clip);
            }
        }
        GrabEvent *evt = grabq.firstEvent();
        while (evt) {
            ClipGrab &clip = evt->message();
            samples_t offset = evt->start();
            clip.clear();
            clip.append(bus, offset, nframes - offset);
            activeGrabs.add(&clip);
            evt = grabq.nextEvent();
        }
    }
    void reportConnections(void *node) {
        void *array[mchannels];
        for (uint32_t i = 0; i < mchannels; i++) {
            array[i] = &bus[i];
        }
        StatusServer::instance().connections(node, array, mchannels);
    }
    void reportLevels(void *node, samples_t time) {
        StatusServer &statusserver = StatusServer::instance();
        if (statusserver.active()) {
            if (time - lastreport > Meter::REPORT_PERIOD) {
                for (uint32_t i = 0; i < mchannels; i++) {
                    statusserver.audioLevel(node, bus[i].maxlevel(), i, i == mchannels - 1);
                }
                lastreport = time;
            } else {
                for (uint32_t i = 0; i < mchannels; i++) {
                    bus[i].findmax();
                }
            }
        }
    }
    bool activeEvents() { return grabq.active() || activeGrabs.first(); }

  public:
    Source(uint32_t channels, bool allocate = true)
        : bus(channels), mchannels(channels), lastreport(0) {
        for (uint32_t i = 0; i < channels; i++) {
            bus.set(i, new Connection(this, allocate), binding::ConnectionObject);
        }
    }
    ~Source() {
        for (uint32_t i = 0; i < mchannels; i++) {
            delete &bus[i];
        }
    }
    BoundArray<Connection> &outputbus() { return bus; }
    HSQOBJECT output() { return bus.hsqobject(); }
    uint32_t channels() { return mchannels; }
    void grab(float length, ScriptFunction &handler, float measure) {
        uint32_t framerate = AudioEngine::instance().sampleRate();
        samples_t size = length * framerate;
        ClipGrab *grab = new ClipGrab(mchannels, size, framerate, handler);
        grabq.createEvent(grab, measure, Clock::defaultClock());
    }
    void restart() {
        grabq.restart();
        ClipGrab *clip = activeGrabs.first();
        while (clip) {
            ClipGrab *done = clip;
            clip = activeGrabs.next(clip);
            activeGrabs.remove(done);
        }
    }
    void restartScript() { grabq.restartScript(); }
};

} // namespace audio
} // namespace bipscript

#endif // AUDIOCONNECTION_H
