/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <iostream>

#include "audioconnection.h"
#include "audioengine.h"
#include "jackclock.h"
#include "lv2plugin.h"
#include "scripthost.h"
#include "timebase.h"

using namespace std;

namespace bipscript {

bool AudioEngine::loaded = false;
Collector<ProcessClosure> ProcessClosure::collector;

namespace audio {

float *Connection::mdummyBuffer = 0;
jack_nframes_t Connection::mbufferSize;

} // namespace audio

void printPosition(jack_transport_state_t state, jack_position_t *pos,
                   jack_nframes_t runningFrame) {
    cout << "====" << endl << "sync: ";
    cout << "state is ";
    cout << (state == JackTransportStopped ? "STOPPED" : "");
    cout << (state == JackTransportStarting ? "STARTING" : "");
    cout << (state == JackTransportRolling ? "ROLLING" : "");
    cout << endl;
    cout << "frame is " << pos->frame << endl;
    cout << "ticks per beat is " << pos->ticks_per_beat << endl;
    if (pos->valid & JackPositionBBT) {
        std::cout << "BBT position is valid" << std::endl;
    }
    if (pos->valid & JackBBTFrameOffset) {
        std::cout << "frame offset is valid" << std::endl;
    }
    // if (state != JackTransportStopped) {
    cout << "bpm is " << pos->beats_per_minute << endl;
    cout << "beat is " << pos->beat << endl;
    cout << "bar is " << pos->bar << endl;
    //}
    if (pos->frame > runningFrame) {
        cout << "jumped ahead from " << runningFrame << " to " << pos->frame << endl;
    } else if (pos->frame == runningFrame) {
        cout << "no change, frame still " << runningFrame << endl;
    } else {
        cout << "rewinding from " << runningFrame << " back to " << pos->frame << endl;
    }
    cout << "====" << endl;
}

int jack_process(jack_nframes_t nframes, void *arg) {
    return ((AudioEngine *)arg)->process(nframes);
}

int sync_callback(jack_transport_state_t state, jack_position_t *pos, void *arg) {
    AudioEngine *ae = static_cast<AudioEngine *>(arg);
    return ae->sync(state, pos);
}

void latency_callback(jack_latency_callback_mode_t mode, void *arg) {
    static_cast<AudioEngine *>(arg)->latency(mode);
}

int buffersize_callback(jack_nframes_t nframes, void *arg) {
    ((AudioEngine *)arg)->setBufferSize(nframes);
    return 0;
}

int xrun_callback(void *arg) {
    ((AudioEngine *)arg)->handleXrun();
    return 0;
}

void jack_shutdown_callback(void *arg) { ((AudioEngine *)arg)->onshutdown(); }

void AudioEngine::setBufferSize(jack_nframes_t size) {
    // TODO: generic size listener?
    audio::Connection::bufferSize(size);
    lv2::PluginFactory::instance().bufferSize(size);
}

AudioEngine::AudioEngine()
    : client(nullptr), mrestarting(false), timebase(nullptr), externalClock(nullptr),
      activeClocks(16), mactiveProcessors(true), activeProcessors(128), deletedProcessors(16) {
    // connect to Jack
    ScriptHost &host = ScriptHost::instance();
    client = jack_client_open(host.filename(), JackNullOption, NULL);
    if (!client) {
        throw std::logic_error("Failed to create JACK client");
    }

    // get sample rate and buffer size
    msampleRate = jack_get_sample_rate(client);
    setBufferSize(jack_get_buffer_size(client));
    time::Window::sampleRate(msampleRate);

    // Jack callbacks
    jack_set_process_callback(client, jack_process, this);
    jack_set_sync_callback(client, sync_callback, this);
    jack_set_latency_callback(client, latency_callback, this);
    jack_set_buffer_size_callback(client, &buffersize_callback, this);
    jack_set_xrun_callback(client, xrun_callback, this);
    jack_on_shutdown(client, &jack_shutdown_callback, this);

    // activate client
    if (jack_activate(client)) {
        throw std::logic_error("Failed to activate audio engine");
    }
    // sample rate now valid
    lv2::PluginFactory::instance().sampleRate(sampleRate());
    host.addAudioEngine(this);
    loaded = true;
}

void AudioEngine::shutdown() {
    jack_deactivate(client);
    jack_client_close(client);
    audio::Connection::deallocate();
}

HSQOBJECT AudioEngine::getPorts(const char *name, const char *type, unsigned long flags) {
    const char **ports = jack_get_ports(client, name, type, flags);
    // count ports
    size_t count = 0;
    if (ports) {
        for (size_t i = 0; ports[i]; i++) {
            count++;
        }
    }
    // create array
    HSQOBJECT sqarray;
    HSQUIRRELVM vm = VirtualMachine::instance().vm();
    sq_newarray(vm, count);
    sq_getstackobj(vm, -1, &sqarray);
    sq_addref(vm, &sqarray);
    for (size_t i = 0; i < count; i++) {
        sq_pushinteger(vm, i);
        sq_pushstring(vm, ports[i], -1);
        sq_set(vm, -3);
    }
    sq_poptop(vm);
    return sqarray;
}

void timebase_callback(jack_transport_state_t state, jack_nframes_t nframes, jack_position_t *pos,
                       int new_pos, void *arg) {
    static_cast<jack::Timebase *>(arg)->timebase(state, nframes, pos, new_pos);
}

jack::Timebase *AudioEngine::getTimebase(float bpm, time::Signature &signature) {
    timebase = new jack::Timebase(client, bpm, signature);
    int error = jack_set_timebase_callback(client, 0, &timebase_callback, timebase);
    if (error) {
        throw std::logic_error("Cannot create transport master");
    }
    return timebase;
}

jack::Clock *AudioEngine::externalTimeClock() {
    if (!externalClock) {
        externalClock = new jack::Clock(client);
        addClock(externalClock);
    }
    return externalClock;
}

// from the jack API docs: TRUE (non-zero) when ready to roll
int AudioEngine::sync(jack_transport_state_t state, jack_position_t *pos) {
    if (masterSync(state, pos)) {
        syncPeriods = 0;
        return 1;
    }
    syncPeriods++;
    return 0;
}

void AudioEngine::latency(jack_latency_callback_mode_t mode) {
    if (mode == JackPlaybackLatency) {
        // takes max latency of all output audio ports
        const char **ports =
            jack_get_ports(client, NULL, JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput);
        if (ports) {
            for (int i = 0; ports[i] != NULL; i++) {
                jack_latency_range_t latencyRange;
                jack_port_t *port = jack_port_by_name(client, ports[i]);
                jack_port_get_latency_range(port, JackPlaybackLatency, &latencyRange);
                if (latencyRange.max > mlatency) {
                    mlatency.store(latencyRange.max);
                }
            }
            jack_free(ports);
        }
    }
}

void AudioEngine::onshutdown() {
    ScriptHost::instance().removeProcessorThread(this);
    loaded = false;
}

void AudioEngine::connectPort(jack_port_t *port, Regex &regex, const char *type) {
    const char **ports = jack_get_ports(client, NULL, type, JackPortIsInput);
    if (ports) {
        for (int i = 0; ports[i] != NULL; i++) {
            if (regex.matches(ports[i])) {
                jack_connect(client, jack_port_name(port), ports[i]);
            }
        }
        jack_free(ports);
    }
}

void AudioEngine::connectPort(Regex &regex, jack_port_t *port, const char *type) {
    const char **ports = jack_get_ports(client, NULL, type, JackPortIsOutput);
    if (ports) {
        for (int i = 0; ports[i] != NULL; i++) {
            if (regex.matches(ports[i])) {
                jack_connect(client, ports[i], jack_port_name(port));
            }
        }
        jack_free(ports);
    }
}

void AudioEngine::disconnectOutputPorts(jack_port_t *port) {
    const char **ports = jack_port_get_connections(port);
    if (ports) {
        for (int i = 0; ports[i] != NULL; i++) {
            jack_disconnect(client, ports[i], jack_port_name(port));
        }
        jack_free(ports);
    }
}

void AudioEngine::disconnectInputPorts(jack_port_t *port) {
    const char **ports = jack_port_get_connections(port);
    if (ports) {
        for (int i = 0; ports[i] != NULL; i++) {
            jack_disconnect(client, jack_port_name(port), ports[i]);
        }
        jack_free(ports);
    }
}

bool AudioEngine::masterSync(jack_transport_state_t state, jack_position_t *pos) {
    if (timebase) {
        if (!syncPeriods) {
            timebase->sync(state, pos);
        }
        // wait for onReposition to complete
        return timebase->syncComplete();
    } else {
        // no transport master
        return true;
    }
}

// process thread
void AudioEngine::removeDeletedProcessors() {
    AudioProcessor *done;
    while (deletedProcessors.pop(done)) {
        activeProcessors.remove(done);
        // TODO: recycle
    }
}

int AudioEngine::process(jack_nframes_t nframes) {
    // frame time of this period
    jack_nframes_t time = jack_last_frame_time(client);

    // script restart requested
    bool restartRequested = mrestarting.load();
    if (restartRequested) {
        // mark clocks as finished
        Clock *clock = activeClocks.first();
        while (clock) {
            clock->finished();
            clock = activeClocks.next(clock);
        }
        // reset processors
        removeDeletedProcessors();
        AudioProcessor *obj = activeProcessors.first();
        while (obj) {
            obj->restart();
            obj = activeProcessors.next(obj);
        }
        // remove clocks
        clock = activeClocks.first();
        while (clock) {
            Clock *done = clock;
            clock = activeClocks.next(clock);
            activeClocks.remove(done);
            done->recycle();
        }
        externalClock = nullptr;
        mrestarting.store(false);
    }

    // update clocks
    Clock::period(nframes);
    Clock *clock = activeClocks.first();
    while (clock) {
        clock->update(time);
        clock = activeClocks.next(clock);
    }

    // active flag = app stays alive
    bool active = ScriptHost::instance().running();
    active = active || MethodQueue::instance().active();

    // run active processors
    removeDeletedProcessors();
    AudioProcessor *obj = activeProcessors.first();
    while (obj) {
        if (obj->process(time)) {
            active = true;
        }
        obj = activeProcessors.next(obj);
    }

    // free process-allocated objects
    ProcessClosure::free();
    time::RemoteWindow::free();

    // update script host with active processor status
    mactiveProcessors.store(active);
    return 0;
}

void AudioEngine::handleXrun() {
    ScriptFunction *handler = onXrunHandler.load();
    if (handler) {
        MethodQueue::instance().dispatch(new XrunClosure(*handler));
    }
}

} // namespace bipscript
