﻿/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "audioport.h"

#include <algorithm>
#include <cstring>

namespace bipscript {
namespace audio {

const char *SYSTEM_CAPTURE_PREFIX = "system:capture_";
const char *SYSTEM_PLAYBACK_PREFIX = "system:playback_";

AudioInputPort *AudioInputPortFactory::newAudioInputPort(const char *name, const char *connection) {
    // see if port already exists in map
    int key = std::hash<std::string>()(name);
    AudioInputPort *port = findObject(key);
    if (!port) {
        // create new jack port
        jack_port_t *jackPort = AudioEngine::instance().registerAudioInputPort(name);
        if (!jackPort) {
            std::string errormesg("Failed to register port \"");
            errormesg += name;
            errormesg += "\"";
            throw std::logic_error(errormesg);
        }
        // add to map
        port = new AudioInputPort(jackPort, name);
        registerObject(key, port);
        // auto connect output port
    }
    if (connection) {
        port->connect(connection);
    }
    return port;
}

AudioInputPort::AudioInputPort(jack_port_t *jackPort, const char *name)
    : Source(1, false), port(jackPort) {
    StatusServer &statusserver = StatusServer::instance();
    if (statusserver.active()) {
        statusserver.newInput(this, name, 1);
        reportConnections(this);
    }
}

InputPortSet::InputPortSet(const char *name, uint32_t channels) : Source(channels, false) {
    mport = new jack_port_t *[channels];
    for (uint32_t i = 0; i < channels; i++) {
        std::string portname(name);
        if (channels > 1) {
            portname.append("_");
            portname.append(std::to_string(i + 1));
        }
        mport[i] = AudioEngine::instance().registerAudioInputPort(portname.c_str());
        if (!mport[i]) {
            throw std::logic_error(std::string("Failed to register port ") + portname);
        }
    }
    StatusServer &statusserver = StatusServer::instance();
    if (statusserver.active()) {
        statusserver.newInput(this, name, channels);
        reportConnections(this);
    }
}

InputPortSet::~InputPortSet() { delete[] mport; }

void InputPortSet::autoconnect() {
    AudioEngine &engine = AudioEngine::instance();
    for (uint32_t i = 0; i < channels(); i++) {
        std::string name(SYSTEM_CAPTURE_PREFIX);
        name.append(std::to_string(i + 1));
        engine.connectPort(name.c_str(), mport[i]);
    }
}

void InputPortSet::connect(int index, const char *name) {
    AudioEngine &engine = AudioEngine::instance();
    engine.connectPort(name, mport[index]);
}

void InputPortSet::connect(int index, Regex &regex) {
    AudioEngine::instance().connectPort(mport[index], regex, JACK_DEFAULT_AUDIO_TYPE);
}

void InputPortSet::unregister() {
    AudioEngine &engine = AudioEngine::instance();
    for (uint32_t i = 0; i < channels(); i++) {
        engine.unregisterPort(mport[i]);
    }
}

bool InputPortSet::doProcess(samples_t time) {
    for (uint32_t i = 0; i < Source::channels(); i++) {
        audioOutput(i)->setBuffer((float *)jack_port_get_buffer(mport[i], Clock::period()));
    }
    grabClips(Clock::period());
    reportLevels(this, time);
    return Source::activeEvents();
}

void InputPortSetFactory::removeObject(InputPortSet *set) {
    ProcessorCache::removeObject(set);
    set->unregister();
}

InputPortSet *InputPortSetFactory::newInputPortSet(const char *name, uint32_t channels,
                                                   bool autoconnect) {
    int key = std::hash<std::string>()(name) + channels;
    // see if port set already exists in map
    InputPortSet *inputset = findObject(key);
    if (!inputset) {
        inputset = new InputPortSet(name, channels);
        registerObject(key, inputset);
    }
    if (autoconnect) {
        inputset->autoconnect();
    }
    return inputset;
}

OutputPortSet::OutputPortSet(const char *name, uint32_t channels) : Sink(channels), lastreport(0) {
    mport = new jack_port_t *[channels];
    for (uint32_t i = 0; i < channels; i++) {
        std::string portname(name);
        if (channels > 1) {
            portname.append("_");
            portname.append(std::to_string(i + 1));
        }
        mport[i] = AudioEngine::instance().registerAudioOutputPort(portname.c_str());
        if (!mport[i]) {
            throw std::logic_error(std::string("Failed to register port ") + portname);
        }
    }
    StatusServer &statusserver = StatusServer::instance();
    if (statusserver.active()) {
        statusserver.newOutput(this, name, channels);
        reportConnectors(this);
    }
    meter = new Meter[channels];
}

OutputPortSet::~OutputPortSet() {
    delete[] mport;
    delete[] meter;
}

bool OutputPortSet::doProcess(samples_t time) {
    samples_t nframes = Clock::period();
    // buffers for this output
    uint32_t outputs = channels();
    float *buffer[outputs];
    for (uint32_t i = 0; i < outputs; i++) {
        buffer[i] = (float *)jack_port_get_buffer(mport[i], nframes);
        memcpy(buffer[i], inputaudio(i, time, nframes), nframes * sizeof(float)); // audio from sink
    }
    //
    StatusServer &statusserver = StatusServer::instance();
    if (statusserver.active()) {
        if (time - lastreport > Meter::REPORT_PERIOD) {
            for (uint32_t i = 0; i < channels(); i++) {
                float level = meter[i].maxlevel(buffer[i], nframes);
                statusserver.audioLevel(this, level, i, i == channels() - 1);
            }
            lastreport = time;
        } else {
            for (uint32_t i = 0; i < channels(); i++) {
                meter[i].findmax(buffer[i], nframes);
            }
        }
    }
    return Sink::activeEvents();
}

void OutputPortSet::connect(const char **names) {
    AudioEngine &engine = AudioEngine::instance();
    for (uint32_t i = 0; i < channels(); i++) {
        engine.connectPort(mport[i], names[i]);
    }
}

void OutputPortSet::disconnect() {
    AudioEngine &engine = AudioEngine::instance();
    for (uint32_t i = 0; i < channels(); i++) {
        engine.disconnectInputPorts(mport[i]);
    }
}

void OutputPortSet::autoconnect() {
    AudioEngine &engine = AudioEngine::instance();
    for (uint32_t i = 0; i < channels(); i++) {
        std::string name(SYSTEM_PLAYBACK_PREFIX);
        name.append(std::to_string(i + 1));
        engine.connectPort(mport[i], name.c_str());
    }
}

void OutputPortSet::unregister() {
    AudioEngine &engine = AudioEngine::instance();
    for (uint32_t i = 0; i < channels(); i++) {
        engine.unregisterPort(mport[i]);
    }
}

void OutputPortSetFactory::removeObject(OutputPortSet *set) {
    ProcessorCache::removeObject(set);
    set->unregister();
}

OutputPortSet *OutputPortSetFactory::newOutputPortSet(const char *name, uint32_t channels,
                                                      bool autoconnect) {
    AudioEngine::instance(); // make sure audio engine exists before creating a sink
    int key = std::hash<std::string>()(name) + channels;
    // see if port set already exists in map
    OutputPortSet *output = findObject(key);
    if (!output) {
        output = new OutputPortSet(name, channels);
        registerObject(key, output);
    }
    if (autoconnect) {
        output->autoconnect();
    }
    return output;
}

OutputPortSet *OutputPortSetFactory::newAudioOutputPort(const char *name, const char *connect) {
    OutputPortSet *output = newOutputPortSet(name, 1, false);
    const char *connections[1]{connect};
    output->connect(connections);
    return output;
}

} // namespace audio
} // namespace bipscript
