/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef AUDIOSINK_H
#define AUDIOSINK_H

#include "audioconnection.h"
#include "bufferevent.h"

namespace bipscript {

namespace audio {

// for immediate play
class BufferPlayEvent : public Listable<BufferPlayEvent> {
    Buffer *mbuffer;
    uint32_t mstart;

  public:
    BufferPlayEvent(Buffer *buffer) : mbuffer(buffer) {}
    ~BufferPlayEvent() {
        if (mbuffer->delref() == 0) {
            delete mbuffer;
        }
    }
    void start(uint32_t start) { mstart = start; }
    uint32_t start() { return mstart; }
    uint32_t end() { return mstart + mbuffer->size() - 1; }
    float *buffer(uint32_t start) { return mbuffer->data() + start; }
};

class PlayEventCollector {
    Collector<BufferPlayEvent> eventCollector;
    PlayEventCollector() { CollectorThread::instance().addCollector(&eventCollector); }

  public:
    static PlayEventCollector &instance() {
        static PlayEventCollector instance;
        return instance;
    }
    void recycle(BufferPlayEvent *evt) { eventCollector.recycle(evt); }
};

/**
 * Represents an input for a mono audio connection
 */
class Connector {
    std::atomic<Connection *> mconnection;
    AbstractSink *mparent;
    BufferEventTimelineSet timelines;
    ThreadQueue<BufferPlayEvent> playQueue;
    List<BufferPlayEvent> activePlayList;
    float *buffer;

  public:
    Connector(AbstractSink *parent) : mconnection(nullptr), mparent(parent), playQueue(128) {
        buffer = new float[Connection::bufferSize()];
    }
    void connect(Connection &connection) {
        Source *source = connection.source();
        if (source->connectsTo(mparent)) {
            throw std::logic_error("Cannot connect infinite loop");
        }
        mconnection.store(&connection);
        StatusServer &statusserver = StatusServer::instance();
        if (statusserver.active()) {
            statusserver.connect(&connection, this);
        }
    }
    Connection *connection() { return mconnection.load(); }
    float *audio(time_t time, samples_t nframes) {
        // connection audio if exists
        Connection *connection = mconnection.load();
        if (connection) {
            // ensure connection processed
            connection->source()->process(time);
            // if there are buffer events copy connection audio before adding
            if (timelines.active()) {
                memcpy(buffer, connection->audio(), nframes * sizeof(float));
                addBufferEvents(buffer, nframes, time);
                return buffer;
            } else {
                // return connection audio directly without changes
                return connection->audio();
            }
        } else {
            // no connection on this channel, return local buffer with any events
            memset(buffer, 0, nframes * sizeof(float));
            addBufferEvents(buffer, nframes, time);
            return buffer;
        }
    }
    void play(Buffer &buffer) { playQueue.add(new BufferPlayEvent(buffer.addref())); }
    BufferEvent *schedule(Buffer &buffer, float measure, Clock &clock) {
        BufferEvent *evt = new BufferEvent(buffer.addref(), measure);
        timelines.addEvent(evt, clock);
        return evt;
    }
    BufferEvent *schedule(Buffer &buffer, float measure) {
        BufferEvent *evt = new BufferEvent(buffer.addref(), measure);
        timelines.addEvent(evt);
        return evt;
    }
    void remove(BufferEvent &evt) {
        timelines.removeEvent(&evt, Clock::defaultClock()); // TODO: clock from event
    }
    void addBufferEvents(float *buffer, samples_t nframes, samples_t time) {
        timelines.addBufferEvents(buffer);
        // activate and play immediate play clips
        BufferPlayEvent *playEvent = nullptr;
        while (playQueue.pop(playEvent)) {
            playEvent->start(time);
            activePlayList.add(playEvent);
        }
        playEvent = activePlayList.first();
        while (playEvent) {
            // find overlap of this event and period
            if (playEvent->end() > time) {
                uint32_t overlap = std::min(playEvent->end() - time, nframes);
                float *copyBuffer = playEvent->buffer(time - playEvent->start());
                // loop and add
                for (uint32_t i = 0; i < overlap; i++) {
                    buffer[i] += copyBuffer[i];
                }
            }
            // remove finished events
            if (time + nframes > playEvent->end()) {
                BufferPlayEvent *delEvent = playEvent;
                playEvent = activePlayList.next(playEvent);
                activePlayList.remove(delEvent);
                PlayEventCollector::instance().recycle(delEvent);
            } else {
                playEvent = activePlayList.next(playEvent);
            }
        }
    }
    bool activeTimelines() { return activePlayList.first() != nullptr || timelines.active(); }
    void restart() {
        BufferPlayEvent *playEvent = nullptr;
        while (playQueue.pop(playEvent)) {
            PlayEventCollector::instance().recycle(playEvent);
        }
        playEvent = activePlayList.first();
        while (playEvent) {
            BufferPlayEvent *delEvent = playEvent;
            playEvent = activePlayList.next(playEvent);
            activePlayList.remove(delEvent);
            PlayEventCollector::instance().recycle(delEvent);
        }
        timelines.restart();
    }
    void restartScript() {
        timelines.restartScript();
        mconnection.store(nullptr);
    }
}; // namespace audio

class Sink : virtual public AbstractSink {
    BoundArray<Connector> bus;
    const uint32_t mchannels;

  protected:
    float *inputaudio(uint32_t index, time_t time, samples_t nframes) {
        return bus[index].audio(time, nframes);
    }
    Connection *audioInput(uint32_t index) { return bus[index].connection(); }
    bool connectsTo(AbstractSink *sink) {
        for (uint32_t i = 0; i < mchannels; i++) {
            Connection *conn = bus[i].connection();
            if (conn && conn->source()->connectsTo(sink)) {
                return true;
            }
        }
        return false;
    }
    bool activeEvents() {
        for (uint32_t i = 0; i < mchannels; i++) {
            if (bus[i].activeTimelines()) {
                return true;
            }
        }
        return false;
    }
    void reportConnectors(void *node) {
        void *array[mchannels];
        for (uint32_t i = 0; i < mchannels; i++) {
            array[i] = &bus[i];
        }
        StatusServer::instance().connectors(node, array, mchannels);
    }

  public:
    Sink(uint32_t channels) : bus(channels), mchannels(channels) {
        for (uint32_t i = 0; i < channels; i++) {
            bus.set(i, new Connector(this), binding::ConnectorObject);
        }
    }
    ~Sink() {
        for (uint32_t i = 0; i < mchannels; i++) {
            delete &bus[i];
        }
    }
    BoundArray<Connector> &inputbus() { return bus; }
    HSQOBJECT input() { return bus.hsqobject(); }
    uint32_t channels() { return mchannels; }
    void play(Buffer &buffer) {
        checkAudioInputs();
        for (unsigned int i = 0; i < mchannels; i++) {
            bus[i].play(buffer);
        }
    }
    void play(Clip &clip) {
        checkAudioInputs();
        uint32_t connectChannel = 0;
        for (unsigned int i = 0; i < clip.channels(); i++) {
            if (connectChannel < mchannels) {
                bus[connectChannel++].play(*clip.buffer(i));
            }
        }
    }
    void schedule(Buffer &buffer, float measure, Clock &clock) {
        // schedule on all inputs
        checkAudioInputs();
        for (unsigned int i = 0; i < mchannels; i++) {
            bus[i].schedule(buffer, measure, clock);
        }
    }
    void schedule(Buffer &buffer, float measure) {
        // schedule on all inputs
        checkAudioInputs();
        for (unsigned int i = 0; i < mchannels; i++) {
            bus[i].schedule(buffer, measure);
        }
    }
    void schedule(Clip &clip, float measure, Clock &clock) {
        checkAudioInputs();
        uint32_t connectChannel = 0;
        for (unsigned int i = 0; i < clip.channels(); i++) {
            if (connectChannel < mchannels) {
                bus[connectChannel++].schedule(*clip.buffer(i), measure, clock);
            }
        }
    }
    void schedule(Clip &clip, float measure) {
        checkAudioInputs();
        uint32_t connectChannel = 0;
        for (unsigned int i = 0; i < clip.channels(); i++) {
            if (connectChannel < mchannels) {
                bus[connectChannel++].schedule(*clip.buffer(i), measure);
            }
        }
    }
    void processRestart() {
        for (uint32_t i = 0; i < mchannels; i++) {
            bus[i].restart();
        }
    }
    void scriptRestart() {
        for (uint32_t i = 0; i < mchannels; i++) {
            bus[i].restartScript();
        }
    }

  private:
    void addBufferEvents(uint32_t channel, float *buffer, samples_t nframes, samples_t time) {
        bus[channel].addBufferEvents(buffer, nframes, time);
    }
    void checkAudioInputs() {
        if (mchannels == 0) {
            throw std::logic_error("Cannot schedule/play audio, object has no audio inputs");
        }
    }
};

} // namespace audio
} // namespace bipscript

#endif // AUDIOSINK_H
