﻿/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "bufferevent.h"

namespace bipscript {
namespace audio {

BufferEventTimeline::~BufferEventTimeline() {
    CollectorThread::instance().removeCollector(&cmdcollector);
    ScriptCollector::instance().removeCollector(&evtcollector);
    BufferEvent *evt = sortedEvents.first();
    while (evt) {
        BufferEvent *done = evt;
        evt = sortedEvents.pop();
        if (done->delref() == 0) {
            delete done;
        }
    }
}

bool BufferEventTimeline::update() {
    Command *cmd;
    int counter = 0;
    while (eventQueue.pop(cmd)) {
        if (cmd->etype() == Command::CLEAR) {
            BufferEvent *list = sortedEvents.clear();
            // TODO: implement
            // deletedEvents.addAll(list);
        } else if (cmd->etype() == Command::REMOVE) {
            RemoveBufferCommand *rcmd = static_cast<RemoveBufferCommand *>(cmd);
            BufferEvent *evt = rcmd->event();
            sortedEvents.remove(evt);
            evtcollector.recycle(evt);
            cmdcollector.recycle(cmd);
            last = nullptr;
            counter++; // trigger newEvents in update
        } else {
            sortedEvents.insert(static_cast<BufferEvent *>(cmd));
            // TODO: should be factor of existing list size
            if (++counter == BETL_UPDATE_MAX_EVENTS) {
                eventsPending = true;
                return counter;
            }
        }
    }
    eventsPending = false;
    return counter;
}

bool BufferEventTimeline::applyEvents(float *buffer) {
    bool newEvents = update();
    if (clock.timer().running()) {
        time::Window *window = clock.timer().firstpos();
        while (window) {
            // update start on previous events
            if (window->reposition() || (!last && newEvents)) {
                last = sortedEvents.first();
                while (last && last->measure() < window->measure()) {
                    last->start(clock.toFrame(last->measure()));
                    last->active(clock.hasHistory());
                    last = sortedEvents.next(last);
                }
            }
            // update start on events starting this period
            while (last && last->within(window)) {
                // start for buffer events is absolute not relative
                last->start(last->start() - window->start() + window->frame());
                last->active(true);
                last = sortedEvents.next(last);
            }
            // loop over all events started on or before this period
            BufferEvent *event = sortedEvents.first();
            float *output = &buffer[window->start()];
            samples_t frame = window->frame();
            bool activeEvents = false;
            while (event != last) {
                if (event->active() && event->end() >= frame) {
                    // printf("active event %d - %d, window frame %d\n", event->start(),
                    // event->end(), window->frame());
                    samples_t overlapEnd = std::min(event->end() - frame, window->size());
                    if (event->start() < frame) {
                        float *input = event->buffer(frame - event->start());
                        for (samples_t i = 0; i < overlapEnd; i++) {
                            output[i] += input[i];
                        }
                    } else {
                        float *input = event->buffer(0);
                        for (samples_t i = event->start() - frame; i < overlapEnd; i++) {
                            output[i] += input[i];
                        }
                    }
                    //                printf(" into window [%d-%d]\n", window->start(),
                    //                window->start() + overlapEnd);
                    activeEvents = true;
                }
                event = sortedEvents.next(event);
            }
            if (!last && !eventsPending && !activeEvents) {
                mactive = false;
            }
            window = clock.timer().nextpos(window);
        }
    }
    // printf("buffered timeline active %d\n", mactive);
    return mactive;
}

} // namespace audio
} // namespace bipscript
