﻿/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef BUFFEREVENT_H
#define BUFFEREVENT_H

#include <map>

#include "buffer.h"
#include "clip.h"
#include "clock.h"
#include "collectorthread.h"
#include "event.h"
#include "scriptcollector.h"

#define BETL_UPDATE_MAX_EVENTS 32 // TODO: function of period size

namespace bipscript {

using time::Clock;

namespace audio {

class BufferEvent : public Command,
                    public BaseEvent,
                    public Listable<BufferEvent>,
                    public RefCountable<BufferEvent> {
    Buffer *mbuffer;
    bool mactive;

  public:
    BufferEvent(Buffer *buffer, float measure)
        : Command(SCHEDULE), BaseEvent(measure), mbuffer(buffer), mactive(false) {}
    ~BufferEvent() {
        if (mbuffer->delref() == 0) {
            delete mbuffer;
        }
    }
    samples_t end() { return start() + mbuffer->size() - 1; } // index of last frame
    float *buffer(uint32_t start) { return mbuffer->data() + start; }
    void active(bool active) { mactive = active; }
    bool active() { return mactive; }
};

class RemoveBufferCommand : public Command {
    BufferEvent *mevent;

  public:
    RemoveBufferCommand(BufferEvent *evt) : Command(REMOVE), mevent(evt) {}
    BufferEvent *event() { return mevent; }
};

/**
 * Class that holds buffer events in a sorted list with associated clock.
 * Methods allow accessing the events at the scheduled clock time.
 */
class BufferEventTimeline : public Listable<BufferEventTimeline> {
    ThreadQueue<Command> eventQueue; // script thread -> process thread
    List<BufferEvent> sortedEvents;  // local to process thread
    Clock &clock;
    float updateMeasure;
    bool eventsPending;
    bool mactive;
    BufferEvent *last;
    Collector<Command> cmdcollector;
    RefCollector<BufferEvent> evtcollector;

  public:
    BufferEventTimeline(Clock &clock)
        : eventQueue(2048), clock(clock), updateMeasure(0.0), mactive(true), last(nullptr) {
        CollectorThread::instance().addCollector(&cmdcollector);
        // collect events in script thread
        ScriptCollector::instance().addCollector(&evtcollector);
    }
    ~BufferEventTimeline();
    /**
     * Add an event to this timeline, runs in the script thread.
     */
    void add(Command *evt) { eventQueue.add(evt); }
    /**
     * Add events from this timeline to the given buffer. Runs in the process thred.
     */
    bool applyEvents(float *buffer);

  private:
    bool update();
};

/**
 * An iterable set of buffer event timelines that dispatches events to each timeline by clock.
 */
class BufferEventTimelineSet {
    std::map<Clock *, BufferEventTimeline *> timelines;
    QueueList<BufferEventTimeline> activeList;
    bool mactive;
    Collector<BufferEventTimeline> collector;

  public:
    BufferEventTimelineSet() : activeList(16), mactive(false) {
        ScriptCollector::instance().addCollector(&collector);
    }
    /**
     * Add an event to the timeline set and associate with the given clock.
     * Runs in the script thread.
     */
    void addEvent(BufferEvent *evt, Clock &clock) {
        BufferEventTimeline *timeline = timelines[&clock];
        if (timeline == nullptr) {
            timeline = new BufferEventTimeline(clock);
            activeList.add(timeline);
            timelines[&clock] = timeline;
        }
        timeline->add(evt->addref());
    }
    /**
     * Add an event to the timeline set and associate with the default clock.
     * Runs in the script thread.
     */
    void addEvent(BufferEvent *evt) { addEvent(evt, Clock::defaultClock()); }
    void removeEvent(BufferEvent *event, Clock &clock) {
        timelines[&clock]->add(new RemoveBufferCommand(event));
    }
    /**
     * Adds events from this timeline set to the given buffer. Runs
     * in the process thread.
     */
    void addBufferEvents(float *buffer) {
        BufferEventTimeline *timeline = activeList.first();
        mactive = false;
        while (timeline) {
            if (timeline->applyEvents(buffer)) {
                mactive = true;
            }
            timeline = activeList.next(timeline);
        }
    }
    /**
     * Returns false if all events have passed.
     */
    bool active() { return mactive; }
    /**
     * Restart method for the process thread.
     */
    void restart() {
        BufferEventTimeline *timeline = activeList.first();
        while (timeline) {
            BufferEventTimeline *done = timeline;
            timeline = activeList.next(timeline);
            activeList.remove(done);
            collector.recycle(done);
        }
    }
    /**
     * Restart method for the script thread.
     */
    void restartScript() { timelines.clear(); }
};

} // namespace audio
} // namespace bipscript

#endif // BUFFEREVENT_H
