/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <cstring>
#include <stdexcept>

#include "Ebu128LoudnessMeter.h"
#include "buffer.h"
#include "clip.h"

namespace bipscript {
namespace audio {

/**
 * runs in script thread
 */
Clip::Clip(const char *path) {
    SF_INFO sfinfo;
    memset(&sfinfo, 0, sizeof(sfinfo));
    SNDFILE *file = sf_open(path, SFM_READ, &sfinfo);
    if (!file) {
        std::string mesg("Error reading clip from file: ");
        mesg.append(sf_strerror(nullptr));
        throw std::logic_error(mesg);
    }
    if (!load(file, sfinfo)) {
        sf_close(file);
        std::string mesg("Error loading clip at ");
        mesg.append(path);
        throw std::logic_error(mesg);
    }
    sf_close(file);
}

static sf_count_t sfv_filelen(void *handle) {
    BufferReader *reader = static_cast<BufferReader *>(handle);
    return reader->buffer()->size();
}

static sf_count_t sfv_seek(sf_count_t offset, int from, void *handle) {
    BufferReader *reader = static_cast<BufferReader *>(handle);
    sf_count_t pos;
    if (from == SEEK_SET) {
        pos = offset;
    } else if (from == SEEK_CUR) {
        pos = reader->tell() + offset;
    } else {
        pos = reader->buffer()->size() - offset;
    }
    reader->seek(pos);
    return pos;
}

static sf_count_t sfv_read(void *ptr, sf_count_t count, void *handle) {
    BufferReader *reader = static_cast<BufferReader *>(handle);
    unsigned char *data = reader->buffer()->bytes();
    size_t pos = reader->tell();
    memcpy(ptr, &data[pos], count);
    reader->seek(pos + count);
    return count;
}

static sf_count_t sfv_tell(void *handle) { return static_cast<BufferReader *>(handle)->tell(); }

Clip::Clip(bipscript::Buffer &buffer) {
    BufferReader reader(&buffer);

    SF_VIRTUAL_IO sfvirtual;
    memset(&sfvirtual, 0, sizeof(sfvirtual));
    sfvirtual.get_filelen = &sfv_filelen;
    sfvirtual.read = &sfv_read;
    sfvirtual.seek = &sfv_seek;
    sfvirtual.tell = &sfv_tell;

    SF_INFO sfinfo;
    memset(&sfinfo, 0, sizeof(sfinfo));
    SNDFILE *file = sf_open_virtual(&sfvirtual, SFM_READ, &sfinfo, &reader);
    if (!file) {
        std::string mesg("Error reading clip from buffer: ");
        mesg.append(sf_strerror(nullptr));
        throw std::logic_error(mesg);
    }
    if (!load(file, sfinfo)) {
        sf_close(file);
        throw std::logic_error("Error loading clip from buffer");
    }
    sf_close(file);
}

bool Clip::load(SNDFILE *file, SF_INFO &sfinfo) {
    msize = static_cast<uint64_t>(sfinfo.frames);
    mchannels = static_cast<uint32_t>(sfinfo.channels);
    mframerate = static_cast<uint32_t>(sfinfo.samplerate);
    int64_t dataSize = sfinfo.frames * mchannels;
    float *data = new float[dataSize];
    sf_count_t read = sf_read_float(file, data, dataSize);
    if (read != dataSize) {
        return false;
    }
    mbuffer = new Buffer *[mchannels];
    for (uint i = 0; i < mchannels; i++) {
        float *channelData = new float[msize];
        for (uint64_t f = 0; f < msize; f++) {
            channelData[f] = data[f * mchannels + i];
        }
        mbuffer[i] = (new Buffer(channelData, msize))->addref();
    }
    delete[] data;
    return true;
}

Clip::Clip(uint32_t channels, uint64_t size, uint32_t framerate)
    : msize(size), mchannels(channels), mframerate(framerate) {
    mbuffer = new Buffer *[mchannels];
    for (uint i = 0; i < mchannels; i++) {
        float *channelData = new float[msize];
        mbuffer[i] = (new Buffer(channelData, msize))->addref();
    }
}

Clip::Clip(Clip &copy) {
    mframerate = copy.mframerate;
    mchannels = copy.mchannels;
    msize = copy.msize;
    mbuffer = new Buffer *[mchannels];
    for (uint i = 0; i < mchannels; i++) {
        float *channelData = new float[msize];
        memcpy(channelData, copy.mbuffer[i]->data(), msize * sizeof(float));
        mbuffer[i] = (new Buffer(channelData, msize))->addref();
    }
}

Clip::~Clip() {
    for (uint i = 0; i < mchannels; i++) {
        if (mbuffer[i]->delref() == 0) {
            delete mbuffer[i];
        }
    }
    delete[] mbuffer;
}

/*
 * script thread
 */
Buffer *Clip::buffer(uint32_t channel) {
    if (channel >= mchannels) {
        throw std::logic_error("Clip has no such channel");
    }
    return mbuffer[channel];
}

float Clip::lufs() {
    Ebu128LoudnessMeter meter;
    size_t bufsize = 2048;
    meter.prepareToPlay(mframerate, mchannels, bufsize, 10);
    samples_t written = 0;
    while (written < msize) {
        samples_t towrite = std::min(bufsize, msize - written);
        AudioSampleBuffer buffer(mchannels, towrite, false);
        for (uint32_t i = 0; i < mchannels; i++) {
            buffer.data(i, &mbuffer[i]->data()[written]);
        }
        meter.processBlock(buffer);
        written += towrite;
    }
    return meter.getIntegratedLoudness();
}

void Clip::regain(float ratio) {
    for (uint i = 0; i < mchannels; i++) {
        mbuffer[i]->regain(ratio);
    }
}

Clip *ClipFactory::loadClip(const char *file) {
    int key = std::hash<std::string>()(file);
    Clip *clip = findObject(key);
    if (!clip) {
        clip = new Clip(file);
        registerObject(key, clip);
    }
    return clip;
}

Clip *ClipFactory::loadClip(bipscript::Buffer &bytes) {
    int key = bytes.crc();
    Clip *clip = findObject(key);
    if (!clip) {
        clip = new Clip(bytes);
        registerObject(key, clip);
    }
    return clip;
}

} // namespace audio
} // namespace bipscript
