/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "mixer.h"

namespace bipscript {
namespace audio {

/**
 *
 * Mixer constructor.
 *
 * Runs in script thread.
 *
 * Allocates AudioConnector, AudioConnection objects, routing and gain matrices.
 *
 * Sets initial gains if applicable.
 */
Mixer::Mixer(uint32_t inputs, uint32_t outputs, HSQOBJECT *gainsObj)
    : Sink(inputs), Source(outputs) {
    mgain = new atomic<float> *[inputs];
    for (unsigned int i = 0; i < inputs; i++) {
        mgain[i] = new atomic<float>[outputs]();
    }
    if (gainsObj) {
        ScriptArrayIterator gains(*gainsObj);
        if (gains.length() != inputs) {
            throw std::logic_error("gains array has different length than input count");
        }
        for (uint32_t i = 0; gains.hasNext(); i++) {
            ScriptValue &val = gains.next();
            if (val.type == ARRAY) {
                ScriptArrayIterator array(val.arrayObj);
                if (array.length() != outputs) {
                    std::string error("Gain subarray should be length ");
                    error.append(std::to_string(outputs));
                    throw std::logic_error(error);
                }
                for (uint32_t o = 0; array.hasNext(); o++) {
                    ScriptValue &gainValue = array.next();
                    if (gainValue.type == FLOAT) {
                        mgain[i][o] = gainValue.floatValue;
                    } else if (gainValue.type == INTEGER) {
                        mgain[i][o] = gainValue.intValue;
                    } else {
                        throw std::logic_error("Gain value should be a number");
                    }
                }
            } else if (val.type == INTEGER) {
                for (uint32_t o = 0; o < outputs; o++) {
                    mgain[i][o] = val.intValue;
                }
            } else if (val.type == FLOAT) {
                for (uint32_t o = 0; o < outputs; o++) {
                    mgain[i][o] = val.floatValue;
                }
            } else {
                throw std::logic_error("Each gain array entry should be a number or subarray");
            }
        }
    } else {
        uint32_t iindex = 0, ocount = 0;
        while (iindex < inputs) {
            uint32_t oindex = ocount++ % outputs;
            mgain[iindex++][oindex] = 1.0;
        }
    }
    StatusServer &statusserver = StatusServer::instance();
    if (statusserver.active()) {
        statusserver.newMixer(this, inputs, outputs);
        for (uint32_t i = 0; i < inputs; i++) {
            reportGain(i);
        }
        reportConnections(this);
        reportConnectors(this);
    }
}

/**
 * Mixer dtor.
 *
 * Runs in script thread.
 *
 * Deletes AudioConnector, AudioConnection objects, routing and gain matrices.
 *
 */
Mixer::~Mixer() {
    for (unsigned int i = 0; i < audio::Sink::channels(); i++) {
        delete[] mgain[i];
    }
    delete[] mgain;
}

void Mixer::add(Source &source, uint32_t start) {
    uint32_t sourceInputs = source.audio::Source::channels();
    // check enough free inputs
    if (start + sourceInputs > audio::Sink::channels()) {
        throw std::logic_error("Not enough inputs available");
    }
    for (unsigned int i = 0; i < sourceInputs; i++) {
        inputbus()[start + i].connect(source.outputbus()[i]);
    }
}

void Mixer::add(Source &source) { add(source, connectedInputs()); }

void Mixer::add(Source &source, float gain) {
    uint32_t start = connectedInputs();
    add(source, start);
    uint32_t sourceInputs = source.audio::Source::channels();
    // default gains - distribute across outputs
    uint32_t i = 0, o = 0;
    while (i < sourceInputs) {
        uint32_t inputIndex = start + i++;
        uint32_t outputIndex = o++ % audio::Source::channels();
        for (uint32_t j = 0; j < audio::Source::channels(); j++) {
            mgain[inputIndex][j] = 0;
        }
        mgain[inputIndex][outputIndex] = gain;
        reportGain(inputIndex);
    }
}

void Mixer::add(Source &source, HSQOBJECT &gains) {
    uint32_t start = connectedInputs();
    add(source, start);
    uint32_t sourceInputs = source.channels();
    ScriptArrayIterator gainsArray(gains);
    // set gains from array
    if (gainsArray.length() != sourceInputs) {
        throw std::logic_error("Gains array has different length than input count");
    }
    for (uint32_t i = 0; gainsArray.hasNext(); i++) {
        ScriptValue &val = gainsArray.next();
        uint32_t input = start + i;
        if (val.type == ARRAY) {
            ScriptArrayIterator array(val.arrayObj);
            if (array.length() != audio::Source::channels()) {
                std::string error("Gain subarray should be length ");
                error.append(std::to_string(audio::Source::channels()));
                throw std::logic_error(error);
            }
            for (uint32_t o = 0; array.hasNext(); o++) {
                ScriptValue &gainValue = array.next();
                if (gainValue.type == FLOAT) {
                    mgain[input][o] = gainValue.floatValue;
                } else if (gainValue.type == INTEGER) {
                    mgain[input][o] = gainValue.intValue;
                } else {
                    throw std::logic_error("Gain value should be a number");
                }
            }
        } else if (val.type == INTEGER) {
            for (uint32_t o = 0; o < audio::Source::channels(); o++) {
                mgain[input][o] = val.intValue;
            }
        } else if (val.type == FLOAT) {
            for (uint32_t o = 0; o < audio::Source::channels(); o++) {
                mgain[input][o] = val.floatValue;
            }
        } else {
            throw std::logic_error("Each gain array entry should be a number or subarray");
        }
        reportGain(input);
    }
}

void Mixer::add(Connection &connection, uint32_t input) {
    // check enough free inputs
    if (input == audio::Sink::channels()) {
        throw std::logic_error("Not enough inputs available");
    }
    // connect input
    inputbus()[input].connect(connection);
}

void Mixer::add(Connection &connection) { add(connection, connectedInputs()); }

void Mixer::add(Connection &connection, float gain) {
    uint32_t start = connectedInputs();
    add(connection, start);
    // set gain to all outputs
    for (uint32_t o = 0; o < audio::Source::channels(); o++) {
        mgain[start][o].store(gain);
    }
    reportGain(start);
}

void Mixer::add(Connection &connection, HSQOBJECT &gains) {
    uint32_t start = connectedInputs();
    add(connection, start);
    gain(start, gains);
}

void Mixer::gain(uint32_t input, uint32_t output, float gain) {
    if (input >= audio::Sink::channels() || output >= audio::Source::channels()) {
        throw std::logic_error("No such channel on this mixer");
    }
    mgain[input][output].store(gain);
    reportGain(input);
}

void Mixer::gain(uint32_t input, HSQOBJECT &gains) {
    if (input >= audio::Sink::channels()) {
        throw std::logic_error("No such channel on this mixer");
    }
    ScriptArrayIterator array(gains);
    if (array.length() != audio::Source::channels()) {
        throw std::logic_error("Gains array length should match number of outputs");
    }
    for (uint32_t out = 0; array.hasNext(); out++) {
        ScriptValue &value = array.next();
        if (value.type == FLOAT) {
            mgain[input][out].store(value.floatValue);
        } else if (value.type == INTEGER) {
            mgain[input][out].store(value.intValue);
        } else {
            std::logic_error("Gains array item must be a number");
        }
    }
    reportGain(input);
}

/**
 * Schedule a gain change.
 *
 * Runs in script thread.
 *
 * Allocates MixerGainEvent.
 */
void Mixer::schedule(MixerGain &gain, float measure, Clock &clock) {
    if (!gain.applies(audio::Sink::channels(), audio::Source::channels())) {
        throw std::logic_error("No such channel on this mixer");
    }
    gainEventBuffer.createEvent(new MixerGain(gain), measure, clock);
}

/**
 * Restore clean state on a cached mixer before reuse
 *
 * Runs in script thread.
 *
 * No allocations.
 */
void Mixer::restartScript() { audio::Sink::scriptRestart(); }

/**
 * Checks connection path to a source.
 *
 * Runs in script thread.
 */
bool Mixer::connectsTo(AbstractSink *sink) { return sink == this || audio::Sink::connectsTo(sink); }

/**
 * Main process method for a mixer.
 *
 * Runs in process thread.
 *
 * Recycles MidiGainEvent.
 */
bool Mixer::doProcess(samples_t time) {

    // get audio from input connections
    uint32_t inputs = audio::Sink::channels();
    float *audio[inputs];
    for (unsigned int i = 0; i < inputs; i++) {
        Connection *conn = audioInput(i);
        if (conn) {
            conn->source()->process(time);
            audio[i] = conn->audio();
        } else {
            audio[i] = audio::Connection::dummyBuffer();
        }
    }

    // zero out all output channels
    samples_t nframes = Clock::period();
    for (unsigned int i = 0; i < audio::Source::channels(); i++) {
        memset(audioOutput(i)->audio(), 0, nframes * sizeof(float));
    }

    // status reporting
    StatusServer &statusserver = StatusServer::instance();
    bool reporting = statusserver.active();

    // grab first gain event
    Event<MixerGain> *event = gainEventBuffer.firstEvent();

    // loop over frames
    for (jack_nframes_t frame = 0; frame < nframes; frame++) {

        // check for gain changes via scheduled events
        while (event && event->start() == frame) {
            MixerGain &gain = event->message();
            mgain[gain.input()][gain.output()].store(gain.amount());
            if (reporting) {
                statusserver.mixerGain(this, gain.input(), gain.output(), gain.amount());
            }
            event = gainEventBuffer.nextEvent();
        }

        // loop over outputs
        for (unsigned int o = 0; o < audio::Source::channels(); o++) {
            float *buffer = audioOutput(o)->audio();
            // loop over inputs
            for (uint32_t i = 0; i < inputs; i++) {
                buffer[frame] += audio[i][frame] * mgain[i][o];
            }
        }
    }
    grabClips(Clock::period());
    reportLevels(this, time);
    return gainEventBuffer.active();
}

/**
 * Reset method.
 *
 * Runs in the process thread.
 *
 */
void Mixer::restart() {
    // gain event buffer
    gainEventBuffer.restart();
}

void Mixer::reportGain(uint32_t input) {
    StatusServer &statusserver = StatusServer::instance();
    if (statusserver.active()) {
        statusserver.mixerGains(this, input, audio::Source::channels(), mgain[input]);
    }
}

uint32_t Mixer::connectedInputs() {
    int next = 0;
    for (uint32_t i = 0; i < audio::Sink::channels(); i++) {
        if (audioInput(i)) {
            next = i + 1;
        }
    }
    return next;
}

void Mixer::validateInputChannel(uint32_t channel) {
    if (channel == 0) {
        throw std::logic_error("There is no input channel zero");
    }
    if (channel > audio::Sink::channels()) {
        throw std::logic_error("No such input on this mixer");
    }
}

void Mixer::validateOutputChannel(uint32_t channel) {
    if (channel == 0) {
        throw std::logic_error("There is no output channel zero");
    }
    if (channel > audio::Source::channels()) {
        throw std::logic_error("No such output on this mixer");
    }
}

/**
 * Factory method.
 *
 * Runs in the script thread.
 *
 * Allocates Mixer objects.
 */
Mixer *MixerFactory::newMixer(uint16_t inputs, uint16_t outputs, HSQOBJECT *gains) {
    // make sure audio engine started
    AudioEngine::instance();
    // TODO: better hash
    int ioHash = ((int)inputs << 4) | ((int)outputs << 16);
    // count tells how many instances of this mixer io shape
    int count = instanceCount[ioHash];
    instanceCount[ioHash] = count + 1;
    // key for this mixer instance
    int key = ioHash + count;
    Mixer *mixer = findObject(key);
    if (!mixer) {
        mixer = new Mixer(inputs, outputs, gains);
        registerObject(key, mixer);
    } else {
        // TODO: reset gains
    }
    return mixer;
}

} // namespace audio
} // namespace bipscript
