/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "onsetdetector.h"

namespace bipscript {
namespace audio {

OnsetDetector::OnsetDetector()
    : Sink(1), odf(512, 1024, ComplexSpectralDifferenceHWR, HanningWindow), index(0), thold(1.0),
      lastOnsetFrame(0) {
    buffer = new double[ONSET_HOP_SIZE];
    for (unsigned int i = 0; i < HISTORY_SIZE; i++) {
        history[i] = 0.0;
    }
}

OnsetDetector::~OnsetDetector() { delete[] buffer; }

void OnsetDetector::onOnset(ScriptFunction &handler) {
    if (handler.numargs() != 3) {
        throw std::logic_error("onOnset handler should take two arguments");
    }
    onOnsetHandler.store(new ScriptFunction(handler));
}

void OnsetDetector::restore() {
    // TODO: implement type change?
    ScriptFunction *handler = onOnsetHandler.load();
    if (handler) {
        onOnsetHandler.store(nullptr);
        handler->release();
        delete handler;
    }
}

bool OnsetDetector::doProcess(samples_t time) {
    // get incoming audio
    Connection *connection = audioInput(0);
    float *audio;
    if (connection) {
        connection->source()->process(time);
        audio = connection->audio();
    } else {
        audio = Connection::dummyBuffer();
    }

    // add new audio to buffer
    samples_t nframes = time::Clock::period();
    for (jack_nframes_t i = 0; i < nframes; i++) {
        buffer[index++] = audio[i];
        if (index == ONSET_HOP_SIZE) {
            // calculate history median
            double sorted[HISTORY_SIZE];
            for (unsigned int i = 0; i < HISTORY_SIZE; i++) {
                sorted[i] = history[i];
            }
            std::sort(sorted, sorted + HISTORY_SIZE);
            double median = sorted[(HISTORY_SIZE / 2) + 1];
            // process full buffer
            double onset = odf.calculateOnsetDetectionFunctionSample(buffer);
            // calculate time since last onset
            jack_nframes_t sampleRate = AudioEngine::instance().sampleRate();
            float delta = (time - lastOnsetFrame) / (float)sampleRate;
            // fire if ready
            // * onset above local median by threshold amount
            // * long enough period elapsed since last onset
            float fire = thold * 150;
            if (onset - median >= fire && delta > 0.025) {
                ScriptFunction *handler = onOnsetHandler.load();
                if (handler) {
                    // TODO: compute position from interpolation?
                    samples_t offset = 0;
                    MethodQueue::instance().dispatch(
                        new OnOnsetClosure(*handler, onset - median, offset));
                }
                lastOnsetFrame = time;
            }
            // shift history samples one unit down
            for (unsigned int i = 0; i < HISTORY_SIZE - 1; i++) {
                history[i] = history[i + 1];
            }
            history[HISTORY_SIZE - 1] = onset;
            index = 0;
        }
    }
    return false;
}

/**
 * runs in script thread
 */
OnsetDetector *OnsetDetectorFactory::newOnsetDetector() {
    OnsetDetector *detector = findObject(1); // TODO: allow multiple?
    if (!detector) {
        detector = new OnsetDetector();
        registerObject(1, detector);
    } else {
        detector->restore();
    }
    return detector;
}

} // namespace audio
} // namespace bipscript
