/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "bipxtapi.h"

#include "bipsquirrel.h"
#include "extension.h"
#include "xtarray.h"
#include "xtcoreobject.h"
#include "xtfunctioncall.h"
#include "xtprocessor.h"

#include <stdio.h>
#include <string.h>

HSQUIRRELVM vm;

using namespace bipscript;

void bipxt_setvm(HSQUIRRELVM ivm) { vm = ivm; }

// timeline
BXTTimeline bipxt_timeline_create() { return new XtTimelineSet(); }

void bipxt_timeline_schedule(BXTTimeline xtimeline, void *message, float measure) {
    XtTimelineSet *timeline = static_cast<XtTimelineSet *>(xtimeline);
    timeline->createEvent(new XtMessage(message), measure, Clock::defaultClock());
}

void *bipxt_timeline_event_first(BXTTimeline xtimeline) {
    XtTimelineSet *timeline = static_cast<XtTimelineSet *>(xtimeline);
    Event<XtMessage> *evt = timeline->firstEvent();
    return evt ? evt->message().mdata : nullptr;
}

void *bipxt_timeline_event_next(BXTTimeline xtimeline) {
    XtTimelineSet *timeline = static_cast<XtTimelineSet *>(xtimeline);
    Event<XtMessage> *evt = timeline->nextEvent();
    return evt ? evt->message().mdata : nullptr;
}

int bipxt_timeline_active(BXTTimeline xtimeline) {
    return static_cast<XtTimelineSet *>(xtimeline)->active();
}

int bipxt_timeline_restart(BXTTimeline xtimeline) {
    return static_cast<XtTimelineSet *>(xtimeline)->restart();
}

// process thread
BXTProcessThread bipxt_thread_create(BXTBoolean resetproof, BXTProcess process, BXTProcess restart,
                                     void *arg) {
    ProcessorEngine *engine = new XtProcessor(resetproof, process, restart, arg);
    ScriptHost::instance().addProcessorThread(engine);
    return engine;
}

void bipxt_thread_period(BXTProcessThread thread, int period) {
    static_cast<XtProcessor *>(thread)->period(period);
}

void bipxt_thread_finish(BXTProcessThread thread) { static_cast<XtProcessor *>(thread)->join(); }

BXTClass bipxt_bind_class(const char *name, uint32_t namelen, BXTMethod ctor, BXTClass super) {
    sq_pushstring(vm, _SC(name), namelen);
    HSQOBJECT object;
    if (super != nullptr) {
        HSQOBJECT classObj;
        classObj._type = OT_CLASS;
        classObj._unVal.pClass = static_cast<SQClass *>(super);
        sq_pushobject(vm, classObj);
        sq_newclass(vm, true);
    } else {
        sq_newclass(vm, false);
    }
    sq_getstackobj(vm, -1, &object);
    sq_settypetag(vm, -1, object._unVal.pClass);
    bipxt_bind_method("constructor", 11, ctor);
    return object._unVal.pClass;
}

void bipxt_bind_class_end() { sq_newslot(vm, -3, false); }

void bipxt_bind_enumeration(const char *name) {
    sq_pushstring(vm, name, -1);
    sq_newtable(vm);
}

void bipxt_bind_enumerator(const char *name, int32_t value) {
    sq_pushstring(vm, name, -1);
    sq_pushinteger(vm, value);
    sq_newslot(vm, -3, true);
}

void bipxt_bind_enumeration_end() { sq_newslot(vm, -3, true); }

SQRESULT exec_method(HSQUIRRELVM vm) {
    BXTMethod *method;
    sq_getuserdata(vm, -1, (SQUserPointer *)&method, NULL);
    return (*method)(sq_gettop(vm) - 1);
}

void bipxt_bind_method(const char *name, uint32_t namelen, BXTMethod func) {
    sq_pushstring(vm, _SC(name), namelen);
    size_t size = sizeof(func);
    SQUserPointer udata = sq_newuserdata(vm, static_cast<SQUnsignedInteger>(size));
    memcpy(udata, &func, size);
    sq_newclosure(vm, exec_method, 1);
    sq_newslot(vm, -3, false);
}

BXTDataType bipxt_arg_type(BXTIndex index) {
    switch (sq_gettype(vm, index + 1)) {
    case OT_NULL:
        return BXT_TYPE_NULL;
    case OT_INTEGER:
        return BXT_TYPE_INTEGER;
    case OT_FLOAT:
        return BXT_TYPE_FLOAT;
    case OT_STRING:
        return BXT_TYPE_STRING;
    case OT_TABLE:
        return BXT_TYPE_TABLE;
    case OT_ARRAY:
        return BXT_TYPE_ARRAY;
    case OT_USERPOINTER:
        return BXT_TYPE_OBJECT;
    case OT_INSTANCE:
        return BXT_TYPE_INSTANCE;
    default:
        return BXT_TYPE_INTERNAL;
    }
}

BXTValue bipxt_value_create() { return new HSQOBJECT; }

BXTResult bipxt_value_get(BXTIndex index, BXTValue *ptr) {
    if (!SQ_FAILED(sq_getstackobj(vm, index, static_cast<HSQOBJECT *>(*ptr)))) {
        sq_addref(vm, static_cast<HSQOBJECT *>(*ptr));
        return SQ_OK;
    }
    return SQ_ERROR;
}

BXTResult bipxt_get_boolean(BXTIndex index, BXTBoolean *boolp) {
    SQBool sqbool;
    BXTResult result = sq_getbool(vm, index, &sqbool);
    *boolp = sqbool;
    return result;
}

void bipxt_put_boolean(BXTBoolean value) { sq_pushbool(vm, value); }

BXTResult bipxt_get_float(BXTIndex index, BXTFloat *floatp) {
    return sq_getfloat(vm, index, floatp);
}

void bipxt_put_float(float value) { sq_pushfloat(vm, value); }

BXTResult bipxt_get_integer(BXTIndex index, BXTInteger *intp) {
    return sq_getinteger(vm, index, intp);
}

void bipxt_put_integer(BXTInteger value) { sq_pushinteger(vm, value); }

BXTResult bipxt_get_string(BXTIndex index, const char **strp) {
    return sq_getstring(vm, index, strp);
}

void bipxt_put_string(const char *str, uint32_t strlen) { sq_pushstring(vm, str, strlen); }

void bipxt_put_null() { sq_pushnull(vm); }

BXTInstance bipxt_put_this(const void *obj, BXTRelease func) {
    sq_setinstanceup(vm, 1, (SQUserPointer)obj);
    if (func) {
        sq_setreleasehook(vm, 1, func);
    }
    HSQOBJECT instance;
    sq_getstackobj(vm, 1, &instance);
    sq_addref(vm, &instance);
    return instance._unVal.pInstance;
}

BXTResult bipxt_get_this(void **ptr) {
    if (SQ_FAILED(sq_getinstanceup(vm, 1, ptr, 0, false))) {
        return SQ_ERROR;
    }
    return *ptr ? SQ_OK : SQ_ERROR;
}

void bipxt_put_object(void *obj, BXTClass type) {
    HSQOBJECT classObj;
    classObj._type = OT_CLASS;
    classObj._unVal.pClass = static_cast<SQClass *>(type);
    sq_pushobject(vm, classObj);
    sq_createinstance(vm, -1);
    sq_remove(vm, -2);
    sq_setinstanceup(vm, -1, obj);
}

BXTResult bipxt_get_object(BXTIndex index, void **ptr, BXTClass type,
                           BXTBoolean nullable) { // C++ object
    SQUserPointer objPtr;
    if (!SQ_FAILED(sq_getinstanceup(vm, index, (SQUserPointer *)&objPtr, type, false))) {
        *ptr = objPtr;
        return SQ_OK;
    }
    if (nullable && sq_gettype(vm, index) == OT_NULL) {
        *ptr = nullptr;
        return SQ_OK;
    }
    return SQ_ERROR;
}

BXTClass bipxt_object_type(BXTIndex index) {
    HSQOBJECT arg;
    sq_getstackobj(vm, index + 1, &arg);
    return bip_getinstancetypetag(&arg);
}

BXTResult bipxt_core_object_class(const char *package, const char *name, BXTClass *ptr) {
    if (!strncmp("Audio", package, 5)) {
        if (!strncmp("Buffer", name, 6)) {
            *ptr = &audio::binding::BufferObject;
            return SQ_OK;
        }
        if (!strncmp("Clip", name, 4)) {
            *ptr = &audio::binding::ClipObject;
            return SQ_OK;
        }
    } else if (!strncmp("Midi", package, 4)) {
        if (!strncmp("Note", name, 4)) {
            *ptr = &midi::binding::NoteObject;
            return SQ_OK;
        }
    }
    return SQ_ERROR;
}

BXTResult bipxt_core_object_get(BXTIndex index, BXTCoreObject *ptr, BXTClass type) {
    SQUserPointer objPtr;
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &objPtr, type, false))) {
        *ptr = new XtCoreObject(objPtr, type);
        return SQ_OK;
    }
    return SQ_ERROR;
}

BXTResult bipxt_get_instance(BXTIndex index, BXTInstance *ptr) {
    HSQOBJECT obj;
    sq_getstackobj(vm, index, &obj);
    if (obj._type != OT_INSTANCE) {
        return SQ_ERROR;
    }
    *ptr = obj._unVal.pInstance;
    return SQ_OK;
}

BXTResult bipxt_get_function(BXTIndex index, BXTFunction *ptr) {
    HSQOBJECT obj;
    sq_getstackobj(vm, index, &obj);
    if (obj._type != OT_CLOSURE) {
        return SQ_ERROR;
    }
    *ptr = obj._unVal.pClosure;
    return SQ_OK;
}

// ==== arrays

BXTArray bipxt_create_array(BXTIndex size) { return new XtArray(vm, size); }

void bipxt_array_append_object(BXTArray array, void *obj, BXTClass type) {
    HSQOBJECT classObj;
    classObj._type = OT_CLASS;
    classObj._unVal.pClass = static_cast<SQClass *>(type);
    static_cast<XtArray *>(array)->appendObject(vm, obj, classObj);
}

BXTResult bipxt_put_array(BXTArray value) {
    sq_pushobject(vm, static_cast<XtArray *>(value)->hsqobject());
    return SQ_OK;
}

BXTResult bipxt_execute(BXTFunction function, BXTInstance context) {
    HSQOBJECT funcObj;
    funcObj._type = OT_CLOSURE;
    funcObj._unVal.pClosure = static_cast<SQClosure *>(function);
    sq_pushobject(vm, funcObj);
    if (!context) {
        // push closure root as "this" pointer
        sq_getclosureroot(vm, -1);
    } else {
        HSQOBJECT contextObj;
        contextObj._type = OT_INSTANCE;
        contextObj._unVal.pInstance = static_cast<SQInstance *>(context);
        sq_pushobject(vm, contextObj);
    }
    // attempt to call function
    if (SQ_FAILED(sq_call(vm, 1, SQFalse, SQTrue))) {
        return SQ_ERROR;
    }
    // pop closure
    sq_pop(vm, 1);
    return SQ_OK;
}

BXTResult bipxt_error(const char *mesg) { return sq_throwerror(vm, mesg); }

BXTResult bipxt_not_found() {
    sq_pushnull(vm);
    return sq_throwobject(vm);
}

BXTAudioProcessor bipxt_audio_processor(BXTObject instance, const char *name, BXTIndex inputs,
                                        BXTIndex outputs, BXTAudioProcess func) {
    return bipscript::XtAudioProcessorMap::instance().add(instance, name, inputs, outputs, func);
}

BXTFloat *bipxt_audio_input_buffer(BXTAudioProcessor handle, BXTIndex index) {
    XtAudioProcessor *processor = static_cast<XtAudioProcessor *>(handle);
    return processor->audioinputbuffer(index);
}

BXTFloat *bipxt_audio_output_buffer(BXTAudioProcessor handle, BXTIndex index) {
    XtAudioProcessor *processor = static_cast<XtAudioProcessor *>(handle);
    return processor->outputbus()[index].audio();
}

void bipxt_audio_schedule(BXTAudioProcessor handle, BXTCoreObject arg, float measure) {
    XtAudioProcessor *processor = static_cast<XtAudioProcessor *>(handle);
    XtCoreObject *object = static_cast<XtCoreObject *>(arg);
    if (object->type() == &audio::binding::BufferObject) {
        audio::Buffer *buffer = static_cast<audio::Buffer *>(object->obj());
        processor->schedule(*buffer, measure);
    } else if (object->type() == &audio::binding::ClipObject) {
        audio::Clip *clip = static_cast<audio::Clip *>(object->obj());
        processor->schedule(*clip, measure);
    }
}

BXTMidiProcessor bipxt_midi_processor(BXTObject instance, const char *name, BXTIndex inputs,
                                      BXTIndex outputs, BXTAudioProcess func) {
    // TODO: implement
    return nullptr;
}

BXTMidiProcessor bipxt_midi_processor_add(BXTAudioProcessor handle, BXTIndex inputs,
                                          BXTIndex outputs) {
    XtAudioProcessor *audioproc = static_cast<XtAudioProcessor *>(handle);
    return bipscript::XtMidiProcessorMap::instance().add(audioproc, inputs, outputs);
}

void bipxt_midi_input_event(BXTMidiProcessor handle, BXTIndex index, uint32_t *offset,
                            unsigned char *buffer, uint32_t insize, uint32_t *outsize) {
    XtMidiProcessor *processor = static_cast<XtMidiProcessor *>(handle);
    *outsize = processor->nextEvent(index, offset, buffer, insize);
}

void bipxt_midi_schedule(BXTMidiProcessor handle, BXTCoreObject arg, float measure) {
    XtMidiProcessor *processor = static_cast<XtMidiProcessor *>(handle);
    XtCoreObject *object = static_cast<XtCoreObject *>(arg);
    if (object->type() == &midi::binding::NoteObject) {
        midi::Note *note = static_cast<midi::Note *>(object->obj());
        processor->schedule(*note, measure);
    }
}

BXTFunctionCall bipxt_create_call(BXTFunction function) { return new XtFunctionCall(function); }

void bipxt_call_add_integer(BXTFunctionCall call, BXTInteger value) {
    XtFunctionCall *xtcall = static_cast<XtFunctionCall *>(call);
    xtcall->addInt(value);
}

void bipxt_call_add_instance(BXTFunctionCall call, void *obj) {
    XtFunctionCall *xtcall = static_cast<XtFunctionCall *>(call);
    xtcall->addObject(obj);
}

void bipxt_call_add_string(BXTFunctionCall call, const char *value) {
    XtFunctionCall *xtcall = static_cast<XtFunctionCall *>(call);
    xtcall->addString(value);
}

void bipxt_call_add_value(BXTFunctionCall call, BXTValue value) {
    XtFunctionCall *xtcall = static_cast<XtFunctionCall *>(call);
    xtcall->addValue(value);
}

BXTResult bipxt_execute_call(BXTFunctionCall call, BXTInstance context) {
    XtFunctionCall *xtcall = static_cast<XtFunctionCall *>(call);
    HSQOBJECT funcObj;
    funcObj._type = OT_CLOSURE;
    funcObj._unVal.pClosure = static_cast<SQClosure *>(xtcall->function());
    sq_pushobject(vm, funcObj);
    if (!context) {
        // push closure root as "this" pointer
        sq_getclosureroot(vm, -1);
    } else {
        HSQOBJECT contextObj;
        contextObj._type = OT_INSTANCE;
        contextObj._unVal.pInstance = static_cast<SQInstance *>(context);
        sq_pushobject(vm, contextObj);
    }
    // add args
    for (XtCallArg arg : xtcall->args()) {
        if (arg.type == XtCallArgType::INTEGER) {
            sq_pushinteger(vm, arg.ivalue);
        } else if (arg.type == XtCallArgType::STRING) {
            sq_pushstring(vm, arg.svalue, -1);
        } else if (arg.type == XtCallArgType::INSTANCE) {
            HSQOBJECT obj;
            obj._type = OT_INSTANCE;
            obj._unVal.pInstance = static_cast<SQInstance *>(arg.ovalue);
            sq_pushobject(vm, obj);
        } else if (arg.type == XtCallArgType::ANY) {
            sq_pushobject(vm, *static_cast<HSQOBJECT *>(arg.ovalue));
        }
    }
    // attempt to call function
    if (SQ_FAILED(sq_call(vm, 1 + xtcall->nargs(), SQFalse, SQTrue))) {
        return SQ_ERROR;
    }
    // pop closure
    sq_pop(vm, 1);
    return SQ_OK;
}

void bipxt_call_release(BXTFunctionCall call) { delete static_cast<XtFunctionCall *>(call); }

void bipxt_function_reference(BXTFunction function) {
    HSQOBJECT obj;
    obj._type = OT_CLOSURE;
    obj._unVal.pInstance = static_cast<SQInstance *>(function);
    sq_addref(vm, &obj);
}

void bipxt_function_release(BXTFunction function) {
    HSQOBJECT obj;
    obj._type = OT_CLOSURE;
    obj._unVal.pInstance = static_cast<SQInstance *>(function);
    sq_release(vm, &obj);
}

BXTResult bipxt_instance_get_function(BXTInstance instance, const char *key, BXTFunction *ptr) {
    HSQOBJECT instobj;
    instobj._type = OT_INSTANCE;
    instobj._unVal.pInstance = static_cast<SQInstance *>(instance);
    sq_pushobject(vm, instobj);
    sq_pushstring(vm, key, -1);
    if (SQ_FAILED(sq_get(vm, -2))) {
        sq_poptop(vm);
        return SQ_ERROR;
    }
    HSQOBJECT obj;
    sq_getstackobj(vm, -1, &obj);
    if (obj._type != OT_CLOSURE) {
        sq_pop(vm, 2);
        return SQ_ERROR;
    }
    *ptr = obj._unVal.pClosure;
    sq_pop(vm, 2);
    return SQ_OK;
}

BXTResult bipxt_function_args(BXTFunction function, BXTInteger *iptr) {
    HSQOBJECT funcObj;
    funcObj._type = OT_CLOSURE;
    funcObj._unVal.pClosure = static_cast<SQClosure *>(function);
    *iptr = bip_getargcount(&funcObj) - 1;
    return SQ_OK;
}

void bipxt_instance_reference(BXTInstance instance) {
    HSQOBJECT obj;
    obj._type = OT_INSTANCE;
    obj._unVal.pInstance = static_cast<SQInstance *>(instance);
    sq_addref(vm, &obj);
}

void bipxt_instance_release(BXTInstance instance) {
    HSQOBJECT obj;
    obj._type = OT_INSTANCE;
    obj._unVal.pInstance = static_cast<SQInstance *>(instance);
    sq_release(vm, &obj);
}
