/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef BIPXTAPI_H
#define BIPXTAPI_H

#include <stdint.h>

#define BXT_FAILED(result) (result < 0)
#define BXT_SUCCESS 0
#define BXT_ERROR -1

extern "C" {

typedef uint32_t BXTBoolean;
#ifdef _LP64
typedef long long BXTInteger;
#else
typedef int BXTInteger;
#endif
typedef float BXTFloat;

typedef int BXTDataType;
#define BXT_TYPE_ARRAY 0
#define BXT_TYPE_BOOLEAN 1
#define BXT_TYPE_FLOAT 2
#define BXT_TYPE_INTERNAL 3
#define BXT_TYPE_INTEGER 4
#define BXT_TYPE_INSTANCE 5 // script object instance
#define BXT_TYPE_NULL 6
#define BXT_TYPE_OBJECT 7 // C++ Object
#define BXT_TYPE_STRING 8
#define BXT_TYPE_TABLE 9

typedef int32_t BXTResult;
typedef int32_t BXTIndex;
typedef BXTResult (*BXTMethod)(BXTIndex);
typedef BXTInteger (*BXTRelease)(void *, BXTInteger size);

typedef void *BXTArray;
typedef void *BXTAudioProcessor;
typedef void *BXTClass;
typedef void *BXTCoreObject; // built-in script objects, Audio.*, Midi.* etc
typedef void *BXTFunction;
typedef void *BXTFunctionCall;
typedef void *BXTInstance; // script object
typedef void *BXTMidiProcessor;
typedef void *BXTObject; // extension-defined c++ object
typedef void *BXTProcessThread;
typedef void *BXTTimeline;
typedef void *BXTValue; // "any" type

typedef BXTResult (*BXTAudioProcess)(BXTIndex, void *);

// init + shutdowns
BXTResult bipxt_ext_init(uint32_t version);
void bipxt_ext_init_pack(const char *name);
void bipxt_ext_shutdown();

// process thread
typedef int (*BXTProcess)(void *);
BXTProcessThread bipxt_thread_create(BXTBoolean resetproof, BXTProcess process, BXTProcess restart,
                                     void *arg);
void bipxt_thread_period(BXTProcessThread thread, int period);
void bipxt_thread_finish(BXTProcessThread thread);

// scheduling via timeline
BXTTimeline bipxt_timeline_create();
void bipxt_timeline_schedule(BXTTimeline, void *event, float measure); // TODO: non-default clock
void *bipxt_timeline_event_first(BXTTimeline);
void *bipxt_timeline_event_next(BXTTimeline);
int bipxt_timeline_active(BXTTimeline);
int bipxt_timeline_restart(BXTTimeline);

// binding
BXTClass bipxt_bind_class(const char *name, uint32_t namelen, BXTMethod ctor, BXTClass superclass);
void bipxt_bind_method(const char *name, uint32_t namelen, BXTMethod func);
void bipxt_bind_class_end();
void bipxt_bind_enumeration(const char *name);
void bipxt_bind_enumerator(const char *name, int32_t value);
void bipxt_bind_enumeration_end();

// bind arguments
BXTDataType bipxt_arg_type(BXTIndex index);

// primitive types
BXTResult bipxt_get_boolean(BXTIndex index, BXTBoolean *boolp);
void bipxt_put_boolean(BXTBoolean value);
BXTResult bipxt_get_integer(BXTIndex index, BXTInteger *intp);
void bipxt_put_integer(BXTInteger value);
BXTResult bipxt_get_float(BXTIndex index, BXTFloat *floatp);
void bipxt_put_float(BXTFloat value);
BXTResult bipxt_get_string(BXTIndex index, const char **strp);
void bipxt_put_string(const char *str, uint32_t strlen);
void bipxt_put_null();

BXTValue bipxt_value_create();
BXTResult bipxt_value_get(BXTIndex, BXTValue *ptr);

// xt-defined objects
BXTResult bipxt_get_object(BXTIndex index, void **ptr, BXTClass type, BXTBoolean nullable);
void bipxt_put_object(void *obj, BXTClass type);
BXTClass bipxt_object_type(BXTIndex index);
BXTResult bipxt_get_this(void **ptr);
BXTInstance bipxt_put_this(const void *obj, BXTRelease func);

// script class objects
BXTResult bipxt_get_instance(BXTIndex index, BXTInstance *ptr);
BXTResult bipxt_instance_get_function(BXTInstance, const char *, BXTFunction *);
void bipxt_instance_reference(BXTInstance instance);
void bipxt_instance_release(BXTInstance instance);

// core c++ objects
BXTResult bipxt_core_object_class(const char *package, const char *name, BXTClass *ptr);
BXTResult bipxt_core_object_get(BXTIndex index, BXTCoreObject *ptr, BXTClass type);
BXTResult bipxt_core_object_put(BXTCoreObject *ptr);

// arrays
BXTArray bipxt_create_array(BXTIndex size);
void bipxt_array_append_object(BXTArray array, void *obj, BXTClass type);
BXTResult bipxt_put_array(BXTArray value);

// functions
BXTResult bipxt_get_function(BXTIndex index, BXTFunction *ptr);
BXTResult bipxt_function_args(BXTFunction function, BXTInteger *iptr);
BXTResult bipxt_execute(BXTFunction function, BXTInstance context);
void bipxt_function_reference(BXTFunction function);
void bipxt_function_release(BXTFunction function);

// method calls
BXTFunctionCall bipxt_create_call(BXTFunction function);
void bipxt_call_add_integer(BXTFunctionCall call, BXTInteger value);
void bipxt_call_add_instance(BXTFunctionCall call, BXTInstance obj);
void bipxt_call_add_string(BXTFunctionCall call, const char *value);
void bipxt_call_add_value(BXTFunctionCall call, BXTValue value);
BXTResult bipxt_execute_call(BXTFunctionCall call, BXTInstance context);
void bipxt_call_release(BXTFunctionCall call);

// main loop event handler
void bipxt_event_handler(BXTMethod handler);

// audio processor
BXTAudioProcessor bipxt_audio_processor(BXTObject instance, const char *name, BXTIndex inputs,
                                        BXTIndex outputs, BXTAudioProcess func);
BXTFloat *bipxt_audio_input_buffer(BXTAudioProcessor processor, BXTIndex index);
BXTFloat *bipxt_audio_output_buffer(BXTAudioProcessor processor, BXTIndex index);
void bipxt_audio_schedule(BXTAudioProcessor instance, BXTCoreObject arg, float measure);

// midi processor
BXTMidiProcessor bipxt_midi_processor(BXTObject instance, const char *name, BXTIndex inputs,
                                      BXTIndex outputs,
                                      BXTAudioProcess func); // create pure midi processor
BXTMidiProcessor bipxt_midi_processor_add(BXTAudioProcessor, BXTIndex inputs,
                                          BXTIndex outputs); // add midi to existing audio processor
void bipxt_midi_input_event(BXTMidiProcessor processor, BXTIndex index, uint32_t *offset,
                            unsigned char *buffer, uint32_t insize, uint32_t *outsize);
void bipxt_midi_schedule(BXTMidiProcessor processor, BXTCoreObject arg, float measure);

// error handling
BXTResult bipxt_error(const char *mesg);
BXTResult bipxt_not_found();
}

#endif // BIPXTAPI_H
