
#include "connectionop.h"

#include "audiosink.h"
#include "bindinterface.h"
#include "midisink.h"

namespace bipscript {

SQInteger bip_dconnect_operator(HSQUIRRELVM vm) {
    try {
        SQUserPointer objPtr;
        // 1st parameter is Audio.Connection?
        if (!SQ_FAILED(
                sq_getinstanceup(vm, 2, &objPtr, &audio::binding::ConnectionObject, false))) {
            audio::Connection *output = static_cast<audio::Connection *>(objPtr);
            if (!SQ_FAILED(
                    sq_getinstanceup(vm, 3, &objPtr, &audio::binding::ConnectorObject, false))) {
                audio::Connector *input = static_cast<audio::Connector *>(objPtr);
                input->connect(*output);
            } else {
                return sq_throwerror(vm, "connection operator rhs must be an Audio.Connector");
            }
        } else { // not an audio connection, is it a midi connection?
            if (!SQ_FAILED(
                    sq_getinstanceup(vm, 2, &objPtr, &midi::binding::ConnectionObject, false))) {
                midi::Connection *midiout = static_cast<midi::Connection *>(objPtr);
                if (!SQ_FAILED(
                        sq_getinstanceup(vm, 3, &objPtr, &midi::binding::ConnectorObject, false))) {
                    midi::Connector *midiin = static_cast<midi::Connector *>(objPtr);
                    midiin->connect(*midiout);
                } else {
                    return sq_throwerror(vm, "connection operator rhs must be a MIDI.Connector");
                }
            } else {
                return sq_throwerror(vm, "connection operator lhs must be a connection object");
            }
        }
        sq_push(vm, 2); // push 1st arg as return
        return 1;
    } catch (std::exception const &e) {
        return sq_throwerror(vm, e.what());
    }
}

SQInteger bip_mconnect_operator(HSQUIRRELVM vm) {
    try {
        // lhs: how many audio and/or midi outputs?
        uint32_t outputs = 0;
        uint32_t midiout = 0;
        audio::Source *source = nullptr;
        midi::Source *msource = nullptr;
        SQObjectType type = sq_gettype(vm, 2);
        if (type == OT_INSTANCE) {
            source = audio::binding::getSource(vm, 2);
            outputs = source != nullptr ? source->channels() : 0;
            msource = midi::binding::getSource(vm, 2);
            midiout = msource != nullptr ? msource->channels() : 0;
        } else if (type == OT_ARRAY) {
            HSQOBJECT obj;
            sq_getstackobj(vm, 2, &obj); // get and push 1st arg for iterator
            sq_pushobject(vm, obj);
            sq_pushnull(vm); // iterator
            while (SQ_SUCCEEDED(sq_next(vm, -2))) {
                audio::Source *asource = audio::binding::getSource(vm, -1);
                uint32_t channels = asource != nullptr ? asource->channels() : 0;
                if (channels > 0) {
                    outputs += channels;
                } else { // not a source, zero outputs and abort
                    outputs = 0;
                    sq_pop(vm, 2);
                    break;
                }
                sq_pop(vm, 2);
            }
            sq_pop(vm, 1);   // pop iterator
            sq_pushnull(vm); // iterator
            while (SQ_SUCCEEDED(sq_next(vm, -2))) {
                midi::Source *asource = midi::binding::getSource(vm, -1);
                uint32_t channels = asource != nullptr ? asource->channels() : 0;
                if (channels > 0) {
                    midiout += channels;
                } else { // not a source, zero midiout and abort
                    midiout = 0;
                    sq_pop(vm, 2);
                    break;
                }
                sq_pop(vm, 2);
            }
            sq_pop(vm, 1); // pop iterator, note lhs array still on stack
        }

        // rhs: how many audio and/or midi inputs?
        uint32_t inputs = 0;
        uint32_t midiin = 0;
        audio::Sink *sink = nullptr;
        midi::Sink *msink = nullptr;
        type = sq_gettype(vm, 3);
        if (type == OT_INSTANCE) {
            sink = audio::binding::getSink(vm, 3);
            inputs = sink != nullptr ? sink->channels() : 0;
            msink = midi::binding::getSink(vm, 3);
            midiin = msink != nullptr ? msink->channels() : 0;
        } else if (type == OT_ARRAY) {
            HSQOBJECT obj;
            sq_getstackobj(vm, 3, &obj); // get and push 2nd arg for iterator
            sq_pushobject(vm, obj);
            sq_pushnull(vm); // iterator
            while (SQ_SUCCEEDED(sq_next(vm, -2))) {
                audio::Sink *asink = audio::binding::getSink(vm, -1);
                uint32_t channels = asink != nullptr ? asink->channels() : 0;
                if (channels > 0) {
                    inputs += channels;
                } else { // not a sink, zero inputs and abort
                    inputs = 0;
                    sq_pop(vm, 2); // pop iterator
                    break;
                }
                sq_pop(vm, 2);
            }
            sq_pop(vm, 1);   // pop iterator
            sq_pushnull(vm); // iterator
            while (SQ_SUCCEEDED(sq_next(vm, -2))) {
                midi::Sink *asink = midi::binding::getSink(vm, -1);
                uint32_t channels = asink != nullptr ? asink->channels() : 0;
                if (channels > 0) {
                    midiin += channels;
                } else { // not a sink, zero midiin and abort
                    midiin = 0;
                    sq_pop(vm, 2); // pop iterator
                    break;
                }
                sq_pop(vm, 2);
            }
            sq_pop(vm, 1); // pop iterator, note rhs array still on stack
        }

        if (outputs > 0 && inputs > 0) { // it's an audio connection
            if (source != nullptr) {     // lhs is object
                if (sink != nullptr) {   // rhs is object
                    uint32_t icount = 0;
                    uint32_t ocount = 0;
                    while (ocount < outputs && icount < inputs) {
                        sink->inputbus()[icount++].connect(source->outputbus()[ocount++]);
                    }
                } else { // rhs is array
                    uint32_t ocount = 0;
                    sq_pushnull(vm); // rhs iterator
                    while (SQ_SUCCEEDED(sq_next(vm, -2))) {
                        audio::Sink *asink = audio::binding::getSink(vm, -1);
                        uint32_t channels = asink->channels();
                        uint32_t icount = 0;
                        while (icount < channels && ocount < outputs) {
                            asink->inputbus()[icount++].connect(source->outputbus()[ocount++]);
                        }
                        sq_pop(vm, 2);
                    }
                    sq_pop(vm, 2); // pop rhs iterator and array
                }
            } else {                   // lhs is array
                if (sink != nullptr) { // rhs is object
                    uint32_t icount = 0;
                    HSQOBJECT obj;
                    sq_getstackobj(vm, 2, &obj);
                    sq_pushobject(vm, obj);
                    sq_pushnull(vm); // iterator
                    while (SQ_SUCCEEDED(sq_next(vm, -2))) {
                        audio::Source *asource = audio::binding::getSource(vm, -1);
                        uint32_t channels = asource->channels();
                        uint32_t ocount = 0;
                        while (ocount < channels && icount < inputs) {
                            sink->inputbus()[icount++].connect(asource->outputbus()[ocount++]);
                        }
                        sq_pop(vm, 2);
                    }
                    sq_pop(vm, 2);   // pop lhs iterator and array
                } else {             // rhs is array
                    sq_pushnull(vm); // rhs iterator
                    audio::Connector *connectors[inputs];
                    uint32_t i = 0;
                    while (SQ_SUCCEEDED(sq_next(vm, -2))) {
                        audio::Sink *asink = audio::binding::getSink(vm, -1);
                        uint32_t channels = asink->channels();
                        for (uint32_t j = 0; j < channels; j++) {
                            connectors[i++] = &asink->inputbus()[j];
                        }
                        sq_pop(vm, 2);
                    }
                    sq_pop(vm, 2);   // pop rhs iterator and array
                    sq_pushnull(vm); // lhs iterator
                    uint32_t icount = 0;
                    while (SQ_SUCCEEDED(sq_next(vm, -2))) {
                        audio::Source *asource = audio::binding::getSource(vm, -1);
                        uint32_t channels = asource->channels();
                        uint32_t ocount = 0;
                        while (ocount < channels && icount < inputs) {
                            connectors[icount++]->connect(asource->outputbus()[ocount++]);
                        }
                        sq_pop(vm, 2);
                    }
                    sq_pop(vm, 2); // pop lhs iterator and array
                }
            }
        } else if (midiout > 0 && midiin > 0) { // it's a midi connection
            if (msource != nullptr) {           // lhs is object
                if (msink != nullptr) {         // rhs is object
                    uint32_t icount = 0;
                    uint32_t ocount = 0;
                    while (ocount < midiout && icount < midiin) {
                        msink->midiinbus()[icount++].connect(msource->midioutbus()[ocount++]);
                    }
                } else { // rhs is array
                    uint32_t ocount = 0;
                    sq_pushnull(vm); // rhs iterator
                    while (SQ_SUCCEEDED(sq_next(vm, -2))) {
                        midi::Sink *asink = midi::binding::getSink(vm, -1);
                        uint32_t channels = asink->channels();
                        uint32_t icount = 0;
                        while (icount < channels && ocount < midiout) {
                            asink->midiinbus()[icount++].connect(msource->midioutbus()[ocount++]);
                        }
                        sq_pop(vm, 2);
                    }
                    sq_pop(vm, 2); // pop rhs iterator and array
                }
            } else {                    // lhs is array
                if (msink != nullptr) { // rhs is object
                    uint32_t icount = 0;
                    sq_pushnull(vm); // lhs iterator
                    while (SQ_SUCCEEDED(sq_next(vm, -2))) {
                        midi::Source *asource = midi::binding::getSource(vm, -1);
                        uint32_t channels = asource->channels();
                        uint32_t ocount = 0;
                        while (ocount < channels && icount < midiin) {
                            msink->midiinbus()[icount++].connect(asource->midioutbus()[ocount++]);
                        }
                        sq_pop(vm, 2);
                    }
                    sq_pop(vm, 2);   // pop lhs iterator and array
                } else {             // rhs is array
                    sq_pushnull(vm); // rhs iterator
                    midi::Connector *connectors[midiin];
                    uint32_t i = 0;
                    while (SQ_SUCCEEDED(sq_next(vm, -2))) {
                        midi::Sink *asink = midi::binding::getSink(vm, -1);
                        uint32_t channels = asink->channels();
                        for (uint32_t j = 0; j < channels; j++) {
                            connectors[i++] = &asink->midiinbus()[j];
                        }
                        sq_pop(vm, 2);
                    }
                    sq_pop(vm, 2);   // pop rhs iterator and array
                    sq_pushnull(vm); // lhs iterator
                    uint32_t icount = 0;
                    while (SQ_SUCCEEDED(sq_next(vm, -2))) {
                        midi::Source *asource = midi::binding::getSource(vm, -1);
                        uint32_t channels = asource->channels();
                        uint32_t ocount = 0;
                        while (ocount < channels && icount < midiin) {
                            connectors[icount++]->connect(asource->midioutbus()[ocount++]);
                        }
                        sq_pop(vm, 2);
                    }
                    sq_pop(vm, 2); // pop lhs iterator and array
                }
            }
        } else {
            // TODO: more precise error reporting
            return sq_throwerror(vm, "connection operator => cannot find common connections");
        }

        sq_push(vm, 2); // push 1st arg as return
        return 1;
    } catch (std::exception const &e) {
        return sq_throwerror(vm, e.what());
    }
}

} // namespace bipscript
