/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef EVENT_H
#define EVENT_H

#include <stdexcept>

#include "listable.h"
#include "refcountable.h"
#include "time/clock.h"

namespace bipscript {

class Command {
  public:
    enum Type { SCHEDULE, TRANSIENT, REMOVE, CLEAR };
    Command(Type type) : mtype(type) {}
    Type etype() { return mtype; }
    void etype(Type type) { mtype = type; }

  private:
    Type mtype;
};

class BaseEvent {
    const float mmeasure;
    samples_t mstart;

  public:
    BaseEvent(float measure) : mmeasure(measure) {
        if (measure < 1.0) {
            throw std::logic_error("Cannot schedule an event before measure 1.0");
        }
    }
    float measure() { return mmeasure; }
    samples_t start() const { return mstart; }
    void start(samples_t value) { mstart = value; }
    bool operator<(BaseEvent &other) { return mmeasure < other.mmeasure; }
    bool operator<=(BaseEvent &other) { return mmeasure <= other.mmeasure; }
    bool operator>=(BaseEvent &other) { return mmeasure >= other.mmeasure; }

    /**
     * returns true if the given event occurs in the window, also sets the "start" property
     * on the event to the frame offset from the start of this *period*
     */
    bool within(time::Window *window) {
        samples_t frames = window->offset(mmeasure);
        start(window->start() + frames);
        //                printf("%f ready%d @%f %d offset in %d\n", measure(), frames <
        //                window->size(), mmeasure, frames, window->size());
        return frames < window->size();
    }
    /**
     * returns the frame offset from the start of the window, also sets the "start" property
     * on the event to this value
     */
    samples_t offset(time::Window *window) {
        samples_t frames = window->offset(mmeasure);
        start(frames);
        return frames;
    }
};

template <class T>
class Event : public Command,
              public BaseEvent,
              public Listable<Event<T>>,
              public RefCountable<Event<T>> {
    T &mmessage;
    time::Clock *mclock;

  public:
    Event(T *mesg, Command::Type type, const float measure, time::Clock &clock)
        : Command(type), BaseEvent(measure), mmessage(*mesg), mclock(&clock) {}
    Event(T *mesg, Command::Type type, const float measure)
        : Command(type), BaseEvent(measure), mmessage(*mesg), mclock(nullptr) {}
    ~Event() { delete &mmessage; }
    T &message() { return mmessage; }
    time::Clock &clock() { return *mclock; }
};

} // namespace bipscript

#endif // EVENT_H
