/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "extension.h"

#include "io.h"
#include <boost/algorithm/string.hpp>
#include <boost/filesystem.hpp>
#include <dlfcn.h>
#include <iostream>

namespace fs = boost::filesystem;

void bipxt_event_handler(BXTMethod handler) {
    bipscript::ExtensionManager::instance().addHandler(handler);
}

namespace bipscript {

Extension::Extension(const std::string &path) {
    handle = dlopen(path.c_str(), RTLD_LAZY);
    if (!handle) {
        std::string mesg("Failed to load extension ");
        mesg.append(dlerror());
        throw std::logic_error(mesg);
    }
    int (*init)(int) = (int (*)(int))dlsym(handle, "bipxt_ext_init");
    char *error = dlerror();
    if (error != NULL) {
        std::string mesg("Failed to initialize extension ");
        mesg.append(error);
        throw std::logic_error(mesg);
    }
    init(BIPXT_VERSION);
}

Extension::~Extension() { dlclose(handle); }

void Extension::import(HSQUIRRELVM vm, const char *name) {
    void (*fn)(const char *) = (void (*)(const char *))dlsym(handle, "bipxt_ext_init_pack");
    char *error = dlerror();
    if (error != NULL) {
        std::string mesg("Failed to bind package ");
        mesg.append(name);
        mesg.append(" ");
        mesg.append(error);
        throw std::logic_error(mesg);
    }
    sq_pushstring(vm, name, -1);
    sq_newtable(vm);
    bipxt_setvm(vm);
    (*fn)(name);
    // TODO: check stack is the same size?
    sq_newslot(vm, -3, false);
}

void Extension::shutdown() {
    void (*fn)() = (void (*)())dlsym(handle, "bipxt_ext_shutdown");
    char *error;
    if ((error = dlerror()) != NULL) {
        throw std::logic_error(error);
    }
    (*fn)();
}

void ExtensionManager::import(HSQUIRRELVM vm, const char *name) {
    auto test = std::find(mreserved.begin(), mreserved.end(), name);
    // is it a core package?
    if (test == mreserved.end()) {
        test = std::find(mimported.begin(), mimported.end(), name);
        // already imported?
        if (test == mimported.end()) {
            for (std::string dir : mextpath) {
                std::string bxtPath(dir);
                bxtPath.append(name);
                bxtPath.append(".bxt.so");
                // does extension exist in this folder?
                if (boost::filesystem::exists(bxtPath)) {
                    Extension *ext = new Extension(bxtPath);
                    ext->import(vm, name);
                    extensions.push_back(ext);
                    mimported.push_back(name);
                    return;
                }
            }
        }
    }
}

void ExtensionManager::shutdown() {
    for (size_t i = 0; i < extensions.size(); i++) {
        extensions[i]->shutdown();
        delete extensions[i];
    }
}

bool ExtensionManager::eventLoop() {
    bool active = false;
    for (size_t i = 0; i < ehandlers.size(); i++) {
        if ((*ehandlers[i])(0) > 0) {
            active = true;
        }
    }
    return active;
}

} // namespace bipscript
