/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef IO_H
#define IO_H

#include "buffertype.h"
#include "heapbuffer.h"
#include "iterator.h"

#include <boost/filesystem.hpp>
#include <string>

namespace bfs = boost::filesystem;

namespace bipscript {
namespace io {

enum Type { String, Integer, Float, Boolean };

class Folder;

class Path {
  protected:
    const std::string mpath;
    const std::string mname;
    Path(const char *path, const char *name) : mpath(path), mname(name) {}
    Path(const char *path) : mpath(path), mname(bfs::path(path).filename().c_str()) {}
    void testexists();

  public:
    virtual bool exists() = 0;
    virtual bool folder() = 0;
    std::string path() { return mpath; }
    std::string name() { return mname; }
    Folder *parent();
};

class FSPath : public Path {
  public:
    FSPath(const char *path) : Path(path) {}
    bool exists() override { return bfs::exists(mpath); }
};

class File : public FSPath {
    HeapBuffer textbuffer;
    size_t mlength;

  public:
    File(const char *fpath);
    uintmax_t len();
    bool folder() override { return false; }
    const char *modified() { return "yes"; } // TODO: what should be here
    const char *text();
    Reader *reader();
    const char *readAll() {
        fprintf(stderr, "warning: readAll() method is deprecated, use text() instead\n");
        return text();
    }
    Buffer *bytes();
    void writeAll(const char *);
    Writer *writer();
    bool remove();

  private:
    void testfile();
};

class FolderIterator : public Iterator {
    bfs::directory_iterator itr;
    bfs::directory_iterator end;

  public:
    FolderIterator(const std::string &path) : itr(path) {}
    bool hasNext() override { return itr != end; }
    void next(HSQUIRRELVM vm) override;
};

class Folder : public FSPath {
  public:
    Folder(const char *fpath);
    bool folder() override { return true; }
    Iterator *children();
    void mkdir() { bfs::create_directory(mpath); }

  private:
    void testfolder();
};

} // namespace io
} // namespace bipscript

#endif // IO_H
