/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "xmltype.h"
#include "tinyxml/tinyxml2.h"
#include <boost/algorithm/string.hpp>

using namespace tinyxml2;

namespace bipscript {
namespace io {

XmlBinder::XmlBinder(HSQUIRRELVM vm, HSQOBJECT *clazz, bool verbose, std::string textfield)
    : vm(vm), mverbose(verbose), mtextfield(textfield), parent(OBJECT) {
    // load entities
    entities["&amp;"] = '&';
    entities["&apos;"] = '\'';
    entities["&gt;"] = '>';
    entities["&lt;"] = '<';
    entities["&quot;"] = '\"';
    // find class argument in stack (should be there!)
    if (clazz) {
        for (int i = sq_gettop(vm); i > 0; i--) {
            HSQOBJECT obj;
            sq_getstackobj(vm, i, &obj);
            if (obj._type == OT_CLASS && obj._unVal.pClass == (*clazz)._unVal.pClass) {
                tagstack[0].typeloc = i;
                typeTextField(vm, i, &tagstack[0].textfield);
                break;
            }
        }
    } else {
        tagstack[0].typeloc = 0;
        tagstack[0].textfield = nullptr;
    }
}

void XmlBinder::processTokens(const char *buffer, const sxmltok_t tokens[], uint32_t num_tokens) {
    for (uint32_t i = 0; i < num_tokens; i++) {
        const sxmltok_t *token = tokens + i;
        switch (token->type) {
        case SXML_STARTTAG: {
            if (parent == OBJECT) {
                uint32_t type = 0;
                const char *textfield = nullptr;
                if (depth) {
                    const char *str = buffer + token->startpos;
                    uint32_t strlen = token->endpos - token->startpos;
                    // push tag name
                    sq_pushstring(vm, str, strlen);
                    uint32_t nowtype = tagstack[tagtop].typeloc;
                    if (nowtype) {
                        // pop key and push attributes table for this field
                        if (SQ_SUCCEEDED(sq_getattributes(vm, nowtype))) {
                            // push key for attributes table now on stack
                            sq_pushstring(vm, "type", 4);
                            if (SQ_SUCCEEDED(sq_get(vm, -2))) {
                                sq_remove(vm, -2); // remove attributes table
                                if (sq_gettype(vm, -1) == OT_ARRAY) {
                                    sq_pushinteger(vm, 0); // [0] array index
                                    sq_get(vm, -2);        // get first element
                                    sq_remove(vm, -2);     // remove array
                                    parent = ARRAY;
                                }
                                type = sq_gettop(vm); // store location of typedef
                                // does this type mapping have an XmlText attribute?
                                typeTextField(vm, type, &textfield);
                            } else if (mverbose) {
                                std::string field(str, strlen);
                                printf("xml binding: field '%s' has no type mapping, mapping to "
                                       "table\n",
                                       field.c_str());
                            }
                        } else {
                            if (mverbose) {
                                const SQChar *name;
                                sq_getstring(vm, -1, &name);
                                printf("xml binding: field '%s' does not exist in object\n", name);
                            }
                            sq_poptop(vm);
                            sq_pushnull(vm); // no type mapping
                        }
                        sq_pushstring(vm, str, strlen); // for table key
                        // start new object or array of type
                        if (parent == OBJECT) {
                            if (type) {
                                sq_createinstance(vm, type);
                            } else {
                                sq_newtable(vm);
                            }
                            // add attributes from start tag
                            parseAttributes(buffer, token, type);
                        } else {
                            sq_newarray(vm, 0);
                        }
                    } else { // no object mapping, use table
                        sq_poptop(vm);
                        sq_pushnull(vm);                // no type mapping
                        sq_pushstring(vm, str, strlen); // for table key
                        sq_newtable(vm);
                        parseAttributes(buffer, token, type);
                    }
                    tagstack[++tagtop].typeloc = type; // push type to type stack
                    tagstack[tagtop].textfield = textfield;
                } else { // start root
                    type = tagstack[0].typeloc;
                    // start new object or array of type
                    if (type) {
                        sq_createinstance(vm, type);
                    } else {
                        sq_newtable(vm);
                    }
                    // add attributes from start tag
                    parseAttributes(buffer, token, type);
                }
            } else if (parent == ARRAY) {
                uint32_t type = tagstack[tagtop].typeloc;
                if (type) {
                    sq_createinstance(vm, type);
                } else {
                    sq_newtable(vm);
                }
                parseAttributes(buffer, token, type);
                parent = OBJECT;
            }
            depth++;
            break;
        }

        case SXML_ENDTAG: {
            SQObjectType parenttype = sq_gettype(vm, -2);
            // wrap up text field if any
            taginfo &tag = tagstack[tagtop];
            if (parent == OBJECT) {
                if (tag.textfield) {
                    sq_pushstring(vm, tag.textfield, strlen(tag.textfield));
                    sq_pushstring(vm, tag.text.c_str(), tag.text.length());
                    sq_set(vm, -3);
                } else if (!tag.typeloc) { // plain table, not an instance
                    sq_pushstring(vm, mtextfield.c_str(), mtextfield.size());
                    sq_pushstring(vm, tag.text.c_str(), tag.text.length());
                    sq_newslot(vm, -3, false);
                }
                tag.text.clear(); // clear inner text for next element in array
            }
            if (--depth) {
                switch (parenttype) {
                case OT_STRING: { // = key within table
                    parent = OBJECT;
                    if (tagtop && tagstack[tagtop - 1].typeloc) { // in a mapped object
                        if (SQ_FAILED(sq_set(vm, -4))) {
                            if (mverbose) {
                                const SQChar *name;
                                sq_getstring(vm, -2, &name);
                                printf("xml binding: bind to object with key '%s' failed\n", name);
                            }
                            sq_pop(vm, 3); // pop type, key, value
                        } else {
                            sq_pop(vm, 1); // pop type for this key
                        }
                    } else {                                // in a table
                        sq_push(vm, -2);                    // push the current key
                        if (SQ_SUCCEEDED(sq_get(vm, -5))) { // key already exists, need array
                            SQObjectType sqtype = sq_gettype(vm, -1);
                            if (sqtype == OT_TABLE) {
                                // convert to array and add existing objects
                                sq_newarray(vm, 0);
                                sq_push(vm, -2); // existing table is now first in array
                                sq_arrayappend(vm, -2);
                                sq_push(vm, -3); // append table for this end tag
                                sq_arrayappend(vm, -2);
                                sq_remove(vm, -2); // remove child tables
                                sq_remove(vm, -2);
                                sq_set(vm, -4); // set array
                                sq_pop(vm, 1);  // pop null type
                            } else if (sqtype == OT_ARRAY) {
                                // already converted to array
                                sq_push(vm, -2); // append table for this end tag
                                sq_arrayappend(vm, -2);
                                sq_pop(vm, 4); // pop array, table, key, type
                            } else {
                                throw std::logic_error("Unknown child type not supported here");
                            }
                        } else {
                            sq_newslot(vm, -4, false);
                            sq_pop(vm, 1); // pop type for this key
                        }
                    }
                    tagtop--;
                    break;
                }
                case OT_ARRAY:
                    parent = ARRAY;
                    sq_arrayappend(vm, -2);
                    break;
                default:
                    throw std::logic_error("Unknown type not supported here");
                }
            }
            break;
        }
        case SXML_CHARACTER: {
            if (parent == OBJECT) {
                const char *str = buffer + token->startpos;
                uint32_t strlen = token->endpos - token->startpos;
                for (uint32_t i = 0; i <= tagtop; i++) {
                    if (tagstack[i].textfield || !tagstack[i].typeloc) {
                        if (str[0] == '&') {
                            tagstack[i].text += escape(str, strlen);
                        } else {
                            tagstack[i].text.append(str, strlen);
                        }
                    }
                }
            }
            break;
        }
            //            case SXML_INSTRUCTION
            //            case SXML_DOCTYPE:
            //            case SXML_COMMENT:
            //            case SXML_CDATA:

        default:
            break;
        }

        i += token->size; // skip tokens associated with current
    }
}

char XmlBinder::escape(const char *text, uint32_t textlen) {
    std::string entity(text, textlen);
    auto iterator = entities.find(entity);
    if (iterator == entities.end()) {
        std::string unknown(text, textlen);
        throw std::logic_error("Unknown entity: " + unknown);
    }
    return iterator->second;
}

void XmlBinder::parseAttributes(const char *buffer, const sxmltok_t *tokens, uint32_t type) {
    uint index = 1; // skip token[0] the tag name
    while (index < tokens->size) {
        // push key
        const sxmltok_t *keytok = tokens + index;
        std::string key(buffer + keytok->startpos, keytok->endpos - keytok->startpos);
        sq_pushstring(vm, key.c_str(), key.length());
        index++;
        // assemble value
        std::string val;
        while (index < tokens->size + 1) {
            const sxmltok_t *token = tokens + index;
            if (token->type != SXML_CHARACTER) {
                break;
            }
            const char *start = buffer + token->startpos;
            uint32_t len = token->endpos - token->startpos;
            if (*start == '&') {
                val += escape(start, len);
            } else {
                val.append(start, len);
            }
            index++;
        }
        if (type) {
            std::map<std::string, int> typemap = attributetypes[type];
            if (typemap.find(key) == typemap.end()) {
                if (SQ_SUCCEEDED(sq_getattributes(vm, type))) {
                    // push key for attributes table now on stack
                    sq_pushstring(vm, "type", 4);
                    if (SQ_SUCCEEDED(sq_get(vm, -2))) {
                        sq_remove(vm, -2); // remove attributes table
                        if (sq_gettype(vm, -1) == OT_INTEGER) {
                            SQInteger typenum;
                            sq_getinteger(vm, -1, &typenum);
                            typemap[key] = typenum;
                        } else {
                            throw std::logic_error(
                                "Error: XML binding attribute must be primitive type");
                        }
                    } else {
                        typemap[key] = Type::String;
                        if (mverbose) {
                            printf("xml binding: attribute %s has no type, mapping to "
                                   "string\n",
                                   key.c_str());
                        }
                    }
                    sq_poptop(vm);
                } else {
                    typemap[key] = Type::String;
                    sq_poptop(vm);
                }
            }
            sq_pushstring(vm, key.c_str(), key.length());
            // push correct type
            switch (typemap[key]) {
            case Type::String:
                sq_pushstring(vm, val.c_str(), val.length());
                break;
            case Type::Integer: {
                SQInteger ival = std::stoi(val, nullptr, 0);
                sq_pushinteger(vm, ival);
                break;
            }
            case Type::Float: {
                SQFloat fval = std::stof(val);
                sq_pushfloat(vm, fval);
                break;
            }
            case Type::Boolean: {
                // TODO: parse
                sq_pushstring(vm, val.c_str(), val.length());
                break;
            }
            default:
                throw std::logic_error("Error XML binding primitive type: unknown type");
            }

            if (SQ_FAILED(sq_set(vm, -3))) {
                if (mverbose) {
                    const SQChar *name;
                    sq_getstring(vm, -2, &name);
                    printf("xml binding: bind to object with key '%s' failed\n", name);
                }
                sq_pop(vm, 2); // pop failed key and value
            }
        } else {
            sq_pushstring(vm, val.c_str(), val.length());
            if (SQ_FAILED(sq_newslot(vm, -3, false))) { // table, use newslot
                sq_pop(vm, 2);
            }
        }
    }
}

void XmlBinder::typeTextField(HSQUIRRELVM vm, uint32_t typeloc, const SQChar **fieldname) {
    sq_pushnull(vm);
    if (SQ_SUCCEEDED(sq_getattributes(vm, typeloc))) {
        sq_pushstring(vm, "XmlText", 7);
        if (SQ_SUCCEEDED(sq_get(vm, -2))) { // get XmlText value from table
            sq_getstring(vm, -1, fieldname);
            sq_poptop(vm);
        }
        sq_poptop(vm);
    }
}

void Xml::read(HSQUIRRELVM &vm, const char *string, bipscript::io::File *iofile, HSQOBJECT *clazz) {

    XmlBinder binder(vm, clazz, mverbose, mtextfield);

    FILE *file = nullptr;
    size_t stringsize = 0;
    if (iofile) {
        if (!iofile->exists()) {
            std::string mesg("XML file does not exist: ");
            mesg.append(iofile->path());
            throw std::logic_error(mesg);
        }
        file = fopen(iofile->path().c_str(), "rb");
    } else {
        stringsize = strlen(string);
    }

    sxmltok_t tokens[TOKENCOUNT];
    char buffer[BUFFERSIZE];
    uint32_t bufferlen = 0;
    uint32_t strptr = 0; // used for strings only

    sxml_t parser;
    sxml_init(&parser);
    uint32_t linenum = 1;

    for (;;) {
        sxmlerr_t err = sxml_parse(&parser, buffer, bufferlen, tokens, TOKENCOUNT);
        if (err == SXML_SUCCESS)
            break;

        switch (err) {
        case SXML_ERROR_TOKENSFULL: {
            binder.processTokens(buffer, tokens, parser.ntokens);
            parser.ntokens = 0;
            break;
        }

        case SXML_ERROR_BUFFERDRY: {
            binder.processTokens(buffer, tokens, parser.ntokens);
            parser.ntokens = 0;
            linenum += linecount(buffer, parser.bufferpos);

            // reuse existing buffer array
            bufferlen -= parser.bufferpos;
            memmove(buffer, buffer + parser.bufferpos, bufferlen);
            size_t read = 0;
            if (file) {
                read = fread(buffer + bufferlen, 1, BUFFERSIZE - bufferlen, file);
            } else {
                read = std::min(BUFFERSIZE - bufferlen, stringsize - strptr);
                strncpy(buffer + bufferlen, &string[strptr], read);
                strptr += read;
            }
            if (!read) {
                if (file) {
                    fclose(file);
                }
                throw std::logic_error("XML parse error: unclosed tag at EOF");
            }
            bufferlen += read;
            parser.bufferpos = 0;
            break;
        }

        case SXML_ERROR_XMLINVALID: {
            if (file) {
                fclose(file);
            }
            std::string mesg("XML parse error @ line ");
            mesg.append(std::to_string(linenum + linecount(buffer, parser.bufferpos)));
            mesg.append(": ");
            uint32_t snippetSize = std::min<uint32_t>(bufferlen - parser.bufferpos, 100);
            mesg.append(buffer + parser.bufferpos, snippetSize);
            throw std::logic_error(mesg);
        }

        default:
            assert(0);
            break;
        }
    }
    binder.processTokens(buffer, tokens, parser.ntokens);
    if (file) {
        fclose(file);
    }
}

uint32_t Xml::linecount(const char *buffer, uint32_t bufferlen) {
    const char *end = buffer + bufferlen;
    const char *it = buffer;

    for (uint32_t i = 0;; i++) {
        it = (const char *)memchr(it, '\n', end - it);
        if (it == nullptr) {
            return i;
        }
        it++;
    }
}

void Xml::write(HSQUIRRELVM vm, const char *ename, File &file) {
    FILE *f = openfile(file.path());
    tinyxml2::XMLPrinter printer(f);
    printer.PushHeader(false, true);
    // table is -3 in stack, behind element name and file
    SQInteger pos = sq_gettop(vm) - 2;
    instance2xml(vm, pos, ename, printer);
    fclose(f);
}

std::string Xml::write(HSQUIRRELVM vm, const char *ename) {
    tinyxml2::XMLPrinter printer;
    printer.PushHeader(false, true);
    // table is -2 in stack, behind element name
    SQInteger pos = sq_gettop(vm) - 1;
    instance2xml(vm, pos, ename, printer);
    return printer.CStr();
}

void Xml::instance2xml(HSQUIRRELVM vm, SQInteger pos, const char *ename,
                       tinyxml2::XMLPrinter &printer) {
    // start element
    printer.OpenElement(ename);

    // get class for this instance
    sq_getclass(vm, pos);
    std::list<std::string> childnodes;
    std::list<std::string> arraychildnodes;

    // push null iterator and iterate over class fields
    // print attributes directly, add object children to list
    sq_pushnull(vm); // push iterator
    while (SQ_SUCCEEDED(sq_next(vm, -2))) {
        // key now at -2, value at -1
        const char *key;
        if (SQ_FAILED(sq_getstring(vm, -2, &key))) {
            throw std::logic_error("XML attribute name must be a string");
        }
        // meta type = integer (for primitive) or array or class
        auto metatype = OT_NULL; // null = no declared type

        // push key on top
        sq_pushstring(vm, key, -1);
        // field has attributes table?
        if (SQ_SUCCEEDED(sq_getattributes(vm, -5))) {
            // attributes table now at top of stack
            // get the "type" attribute (annotation) on this field if any
            sq_pushstring(vm, "type", 4);
            if (SQ_SUCCEEDED(sq_get(vm, -2))) {
                // update meta type = integer (for primitive) or array or class
                metatype = sq_gettype(vm, -1);
                sq_pop(vm, 1); // pop type value
            }
        }
        sq_poptop(vm); // pop attributes table (or null if not found)

        // push key again and get actual instance value
        sq_pushstring(vm, key, -1);
        if (SQ_SUCCEEDED(sq_get(vm, pos))) {
            // instance value and default value and now -1 and -2 on the stack
            // compare will be zero if they are the same
            SQInteger compare = sq_cmp(vm);

            // if the values or different or property writedefault is true
            if (compare || mwritedefault) {

                if (metatype == OT_INTEGER) {
                    // mapped type is primitive: write primitive value to attribute
                    // ignore exact mapped type for now and switch on actual value type
                    char buffer[1024];
                    switch (sq_gettype(vm, -1)) {
                    case OT_NULL:
                        printer.PushAttribute(key, "null");
                        break;
                    case OT_BOOL:
                        SQBool b;
                        sq_getbool(vm, -1, &b);
                        printer.PushAttribute(key, b ? "true" : "false");
                        break;
                    case OT_INTEGER:
                        SQInteger i;
                        sq_getinteger(vm, -1, &i);
                        sprintf(buffer, "%lld", i);
                        printer.PushAttribute(key, buffer);
                        break;
                    case OT_FLOAT:
                        SQFloat fl;
                        sq_getfloat(vm, -1, &fl);
                        sprintf(buffer, "%f", fl);
                        printer.PushAttribute(key, buffer);
                        break;
                    case OT_STRING:
                        const SQChar *str;
                        sq_getstring(vm, -1, &str);
                        printer.PushAttribute(key, str);
                        break;
                    default:
                        break;
                    }
                } else if (metatype == OT_CLASS) {
                    // add to child list
                    childnodes.push_back(key);
                } else if (metatype == OT_ARRAY) {
                    // add to array list
                    arraychildnodes.push_back(key);
                }
            }
            sq_poptop(vm); // pop instance value
        }
        sq_pop(vm, 2); // end iteration - clear last key/value
    }
    sq_poptop(vm); // pop iterator

    // check for inner text
    const char *textfield = nullptr;
    XmlBinder::typeTextField(vm, sq_gettop(vm), &textfield);
    if (textfield) {
        sq_pushstring(vm, textfield, -1);
        if (SQ_SUCCEEDED(sq_get(vm, pos))) {
            if (sq_gettype(vm, -1) == OT_STRING) {
                const SQChar *xmltext;
                sq_getstring(vm, -1, &xmltext);
                printer.PushText(xmltext);
                sq_poptop(vm);
            }
        }
    }
    sq_poptop(vm); // pop class definition

    // loop over child nodes and recurse
    for (auto it = childnodes.begin(); it != childnodes.end(); it++) {
        const char *ename = it->c_str();
        sq_pushstring(vm, ename, -1);
        if (SQ_SUCCEEDED(sq_get(vm, pos))) {
            auto type = sq_gettype(vm, -1);
            if (type == OT_INSTANCE) {
                instance2xml(vm, sq_gettop(vm), ename, printer);
            } else if (type != OT_NULL) {
                // field is mapped as a class, value should be an instance
                throw std::logic_error(
                    "Error serializing XML: field type is class, value should be instance or null");
            }
            sq_poptop(vm); // pop value
        }
    }
    // loop over array child nodes and recurse
    for (auto it = arraychildnodes.begin(); it != arraychildnodes.end(); it++) {
        const char *ename = it->c_str();
        // get the instance value for this field
        sq_pushstring(vm, ename, -1);
        if (SQ_SUCCEEDED(sq_get(vm, pos))) {
            // make sure value is an array
            auto type = sq_gettype(vm, -1);
            if (type == OT_ARRAY) {
                // open parent tag with name = field name
                printer.OpenElement(ename);
                // iterate over array elements
                sq_pushnull(vm);
                while (SQ_SUCCEEDED(sq_next(vm, -2))) {
                    // array element value at -1
                    // what class is it?
                    std::string classname = findClassName(vm, sq_gettop(vm));
                    instance2xml(vm, sq_gettop(vm), classname.c_str(), printer);
                    sq_pop(vm, 2); // end iteration - clear last key/value
                }
                // pop array iterator and close parent tag
                sq_poptop(vm);
                printer.CloseElement();
            } else if (type != OT_NULL) {
                // field is mapped as a class, value should be an instance
                throw std::logic_error(
                    "Error serializing XML: field type is class, value should be instance or null");
            }
            sq_poptop(vm); // pop value
        }
    }
    printer.CloseElement();
}

std::string Xml::findClassName(HSQUIRRELVM vm, SQInteger pos) {
    std::string ret;
    // iterate over root elements
    sq_pushroottable(vm);
    sq_pushnull(vm);
    while (SQ_SUCCEEDED(sq_next(vm, -2))) {
        // instance at pos, put class on stack
        sq_getclass(vm, pos);
        // now compare iteration value with class
        SQInteger different = sq_cmp(vm);
        if (!different) {
            const SQChar *key;
            sq_getstring(vm, -3, &key);
            ret = key;
        }
        sq_pop(vm, 3); // pop iteration key/value + class
    }
    // pop iterator and root table
    sq_pop(vm, 2);
    boost::algorithm::to_lower(ret);
    return ret;
}

} // namespace io
} // namespace bipscript
