/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <jack/jack.h>

#include "audioengine.h"
#include "methodqueue.h"
#include "timebase.h"

namespace bipscript {
namespace jack {

void SyncClosure::finished() {
    clock->handlerCompleted();
    ProcessClosure::finished();
}

void Timebase::timebase(jack_transport_state_t state, jack_nframes_t nframes, jack_position_t *pos,
                        int new_pos) {

    time::ClockChange *command;
    float changed = false;
    while (commands.pop(command)) {
        if (command->type() == time::ClockChange::Type::TEMPO) {
            tempoChange = static_cast<time::TempoChange *>(command)->tempo();
        }
        time::ClockCollector::instance().recycle(command);
        changed = true;
    }
    if (changed) {
        statusUpdate(mtempo);
    }

    if (new_pos) {
        relocate(pos->frame);
        pos->valid = JackPositionBBT;
        statusUpdate(mtempo);
    } else {
        tick();
    }

    // output time sig and tempo
    pos->beats_per_bar = current.signature().numerator();
    pos->beat_type = current.signature().denominator();
    pos->beats_per_minute = static_cast<double>(current.bpm());
    pos->ticks_per_beat = ticksPerBeat;
    // output position in BBT
    pos->bar = static_cast<int>(current.measure());
    float remainder = current.measure() - pos->bar;
    float ticks = remainder * ticksPerBeat * current.signature().beatsPerMeasure();
    pos->beat = 1;
    while (ticks >= ticksPerBeat) {
        pos->beat++;
        ticks -= ticksPerBeat;
    }
    pos->tick = static_cast<int>(ticks + 0.5f);
    //    printf("timebase in: state %d nframes %d frame %d", state, nframes, pos->frame);
    //    printf(" -> %d/%d/%d [%d]\n", pos->bar, pos->beat, pos->tick, pos->frame_rate);
}

void Timebase::update(samples_t) {
    jack_position_t pos;
    jack_transport_state_t state = transport.query(&pos);
    bool jackrolling = state == JackTransportRolling;
    if (jackrolling != mtimer.running()) {
        mtimer.running(jackrolling);
        statusUpdate(mtempo);
    }
    updateRemotes();
}

void Timebase::reposition(float location) {
    if (location < 1.0) {
        throw std::logic_error("cannot reposition before 1.0");
    }
    transport.relocate(toFrame(location));
    statusUpdate(mtempo);
}

void Timebase::sync(jack_transport_state_t state, jack_position_t *pos) {
    SyncHandler *repoHandler = onSyncHandler.load();
    if (repoHandler) {
        mrepositionComplete.store(!repoHandler->wait());
        double measure = current.measure();
        // convert jack position to float
        MethodQueue::instance().dispatch(new SyncClosure(*repoHandler, state, *pos, measure, this));
    }
}

void Timebase::onSync(ScriptFunction &handler, bool wait) {
    if (handler.numargs() != 4) {
        throw std::logic_error("onSync handler should take three arguments");
    }
    onSyncHandler.store(new SyncHandler(handler, wait));
}

/**
 * runs in script thread
 */
Timebase *TimebaseCache::newTimebase(float bpm) {
    Timebase *master = cachedMaster.load();
    if (master) {
        master->tempo(bpm);
    } else {
        time::Signature defsig;
        master = AudioEngine::instance().getTimebase(bpm, defsig);
        StatusServer &statusserver = StatusServer::instance();
        if (statusserver.active()) {
            statusserver.newClock(master, "jack", bpm);
        }
        cachedMaster.store(master);
    }
    AudioEngine::instance().addClock(master);
    active = true;
    return master;
}

Timebase *TimebaseCache::newTimebase(float bpm, time::Signature signature) {
    Timebase *master = cachedMaster.load();
    if (master) {
        // TODO: how do we handle this
        // cached->setBpm(bpm);
        // cached->setTimeSignature(beatsPerBar, beatUnit);
    } else {
        master = AudioEngine::instance().getTimebase(bpm, signature);
        StatusServer &statusserver = StatusServer::instance();
        if (statusserver.active()) {
            statusserver.newClock(master, "jack", bpm, signature.numerator(),
                                  signature.denominator());
        }
        cachedMaster.store(master);
    }
    AudioEngine::instance().addClock(master);
    active = true;
    return master;
}

/**
 * runs in script thread
 */
void TimebaseCache::scriptComplete() {
    Timebase *cached = cachedMaster.load();
    if (cached && !active) {
        cachedMaster.store(nullptr);
        // TODO: implement
        // AudioEngine::instance().releaseTimebaseMaster();
        delete cached;
    }
    active = false;
}

TimebaseCache::~TimebaseCache() {
    Timebase *cached = cachedMaster.load();
    delete cached;
}
} // namespace jack
} // namespace bipscript
