/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "lv2/buf-size/buf-size.h"
#include "lv2/presets/presets.h"

#include "lv2plugin.h"
#include "scripthost.h"

namespace fs = boost::filesystem;

namespace bipscript {
namespace lv2 {

uint32_t MidiEvent::midiEventTypeId;

static LV2_URID uridMap(LV2_URID_Map_Handle handle, const char *uri) {
    return ((UridMapper *)handle)->uriToId(uri);
}

static const char *uridUnmap(LV2_URID_Unmap_Handle handle, LV2_URID urid) {
    return ((UridMapper *)handle)->idToUri(urid);
}

static char *mapAbsolutePath(LV2_State_Map_Path_Handle handle, const char *abstractPath) {
    return ((PathMapper *)handle)->mapAbsolutePath(abstractPath);
}

static char *mapAbstractPath(LV2_State_Map_Path_Handle handle, const char *absolutePath) {
    return ((PathMapper *)handle)->mapAbstractPath(absolutePath);
}

static void setPortValue(const char *port, void *data, const void *value, uint32_t size,
                         uint32_t type) {
    ((Plugin *)data)->setPortValue(port, value, type);
}

static const void *getPortValue(const char *port, void *data, uint32_t *size, uint32_t *type) {
    return ((Plugin *)data)->portValue(port, type, size);
}

static const void *stateRetrieve(LV2_State_Handle handle, uint32_t key, size_t *size,
                                 uint32_t *type, uint32_t *flags) {
    return ((State *)handle)->retrieveState(key, size, type, flags);
}

static LV2_Worker_Status scheduleWork(LV2_Worker_Schedule_Handle handle, uint32_t size,
                                      const void *data) {
    return ((Worker *)handle)->scheduleWork(data, size);
}

static LV2_Worker_Status workerRespond(LV2_Worker_Respond_Handle handle, uint32_t size,
                                       const void *data) {
    return ((Worker *)handle)->queueResponse(data, size);
}

LV2_URID UridMapper::uriToId(const char *uri) {
    const std::string uriString(uri);
    const Lv2UridMap::const_iterator iter = uridMap.find(uriString);
    if (iter == uridMap.end()) {
        const uint32_t newId = uridMap.size() + 10;
        uridMap.insert(std::make_pair(uriString, newId));
        // std::cout << "uridMap storing for " << uri << ": " << newId << std::endl;
        return newId;
    }
    // std::cout << "uridMap returning for " << uri << ": " << iter->second << std::endl;
    return iter->second;
}

const char *UridMapper::idToUri(LV2_URID urid) {
    Lv2UridMap::iterator iter;
    for (iter = uridMap.begin(); iter != uridMap.end(); iter++) {
        if (iter->second == urid) {
            // std::cout << "uridUnMap returning for " << urid << ": " << iter->first << std::endl;
            return iter->first.c_str();
        }
    }
    std::cerr << "!!! uridUnMap failed to find uri for " << urid << std::endl;
    return NULL;
}

char *PathMapper::mapAbsolutePath(const char *abstractPath) {
    char *ret = (char *)malloc(strlen(abstractPath) + 1);
    strcpy(ret, abstractPath);
    return ret;
}

char *PathMapper::mapAbstractPath(const char *absolutePath) {
    char *ret = (char *)malloc(strlen(absolutePath) + 1);
    strcpy(ret, absolutePath);
    return ret;
}

Constants::Constants(LilvWorld *world) {
    lv2AtomPort = lilv_new_uri(world, LV2_ATOM__AtomPort);
    lv2AudioPort = lilv_new_uri(world, LV2_CORE__AudioPort);
    lv2InputPort = lilv_new_uri(world, LV2_CORE__InputPort);
    lv2OutputPort = lilv_new_uri(world, LV2_CORE__OutputPort);
    lv2ControlPort = lilv_new_uri(world, LV2_CORE__ControlPort);
    lv2MidiEvent = lilv_new_uri(world, LILV_URI_MIDI_EVENT);
    lv2AtomSequence = lilv_new_uri(world, LV2_ATOM__Sequence);
    lv2AtomSupports = lilv_new_uri(world, LV2_ATOM__supports);
    lv2AtomBufferType = lilv_new_uri(world, LV2_ATOM__bufferType);
    lv2Presets = lilv_new_uri(world, LV2_PRESETS__Preset);
    lv2WorkerSchedule = lilv_new_uri(world, LV2_WORKER__schedule);
    lv2WorkerInterface = lilv_new_uri(world, LV2_WORKER__interface);
    lv2RdfsLabel = lilv_new_uri(world, LILV_NS_RDFS "label");
}

Constants::~Constants() {
    lilv_free(lv2AtomPort);
    lilv_free(lv2AudioPort);
    lilv_free(lv2InputPort);
    lilv_free(lv2OutputPort);
    lilv_free(lv2ControlPort);
    // midi
    lilv_free(lv2MidiEvent);
    // atom
    lilv_free(lv2AtomSequence);
    lilv_free(lv2AtomSupports);
    lilv_free(lv2AtomBufferType);
    // presets
    lilv_free(lv2Presets);
    // worker
    lilv_free(lv2WorkerInterface);
    lilv_free(lv2WorkerSchedule);
    // rdf
    lilv_free(lv2RdfsLabel);
}

// ----------------------------- Lv2MidiOutput

Event<midi::RawMessage> *MidiOutput::firstEvent() {
    iterator = lv2_atom_sequence_begin(&(atomSequence)->body);
    return nextEvent();
}

Event<midi::RawMessage> *MidiOutput::nextEvent() {
    while (
        !lv2_atom_sequence_is_end(&(atomSequence)->body, (atomSequence)->atom.size, (iterator))) {
        if (iterator->body.type == MidiEvent::midiEventTypeId) {
            if (iterator->body.size <= MAX_EVENT_SIZE) {
                event.message().importBytes((uint8_t *)(iterator + 1), iterator->body.size);
                event.start(iterator->time.frames);
                iterator = lv2_atom_sequence_next(iterator);
                return &event;
            } else {
                iterator = lv2_atom_sequence_next(iterator);
            }
        }
    }
    return 0;
}

// ----------------------------- Lv2Plugin

Plugin::Plugin(const PluginType *type, LilvInstance *instance, LV2_Feature **features,
               Worker *worker)
    : audio::Source(type->audioOutputCount()), audio::Sink(type->audioInputCount()),
      midi::Source(type->midiOutputCount()), midi::Sink(type->midiInputCount()), type(type),
      instance(instance), lv2Features(features), controlChanges(256), worker(worker) {

    for (uint32_t i = 0; i < type->controlInputCount(); i++) {
        const ControlPortDef *def = &type->controlInput(i);
        ControlPort *newPort = new ControlPort(def);
        lilv_instance_connect_port(instance, def->index(), &(newPort->value));
        controlMap[def->name()] = newPort;
    }
    controlOutput = new float[type->controlOutputCount()];
    for (uint32_t i = 0; i < type->controlOutputCount(); i++) {
        lilv_instance_connect_port(instance, type->controlOutput(i).index(), &controlOutput[i]);
    }
    matomOutput = new LV2_Atom_Sequence *[type->atomOutputCount()];
    mmidiOutput = new MidiOutput *[type->midiOutputCount()];
    uint32_t midiOutputCount = 0;
    for (uint32_t i = 0; i < type->atomOutputCount(); i++) {
        const AtomPortDef &def = type->atomOutput(i);
        matomOutput[i] =
            static_cast<LV2_Atom_Sequence *>(malloc(sizeof(LV2_Atom_Sequence) + ATOM_CAPACITY));
        lilv_instance_connect_port(instance, def.index(), matomOutput[i]);
        if (def.midi()) {
            MidiOutput *output = new MidiOutput(this, matomOutput[i], ATOM_CAPACITY);
            midiOutput(output, midiOutputCount);
            mmidiOutput[midiOutputCount++] = output;
        }
    }
    matomInput = new LV2_Atom_Sequence *[type->atomInputCount()];
    midiInput = new LV2_Atom_Sequence *[type->midiInputCount()];
    uint32_t midiInputCount = 0;
    for (uint32_t i = 0; i < type->atomInputCount(); i++) {
        const AtomPortDef &def = type->atomInput(i);
        matomInput[i] =
            static_cast<LV2_Atom_Sequence *>(malloc(sizeof(LV2_Atom_Sequence) + ATOM_CAPACITY));
        lilv_instance_connect_port(instance, def.index(), matomInput[i]);
        if (def.midi()) {
            midiInput[midiInputCount++] = matomInput[i];
        }
    }
    // register collectors
    CollectorThread::instance().addCollector(&controlCollector);
    // init mutex
    if (pthread_mutex_init(&mmutex, NULL) != 0) {
        throw std::logic_error("LV2 Plugin failed to initialize mutex");
    }
    StatusServer &statusserver = StatusServer::instance();
    if (statusserver.active()) {
        statusserver.newLv2Plugin(this, type->name().c_str(), type->audioInputCount(),
                                  type->audioOutputCount(), type->midiInputCount(),
                                  type->midiOutputCount());
        audio::Source::reportConnections(this);
        audio::Sink::reportConnectors(this);
        midi::Source::reportConnections(this);
        midi::Sink::reportConnectors(this);
    }
    // connect worker with plugin instance
    if (worker) {
        worker->setInstance(instance);
    }
}

Plugin::~Plugin() {
    // worker
    if (worker) {
        delete static_cast<LV2_Worker_Schedule *>(lv2Features[6]->data);
        delete lv2Features[6];
        delete worker;
    }
    delete[] lv2Features;
    // midi inputs
    for (uint32_t i = 0; i < type->midiInputCount(); i++) {
        lilv_free(matomInput[i]);
    }
    delete[] matomInput;
    delete[] midiInput;
    // midi outputs
    for (uint32_t i = 0; i < type->atomOutputCount(); i++) {
        lilv_free(matomOutput[i]);
    }
    delete[] matomOutput;
    for (uint32_t i = 0; i < type->midiOutputCount(); i++) {
        delete mmidiOutput[i];
    }
    delete[] mmidiOutput;
    // control outputs
    delete[] controlOutput;
    // control ports
    for (auto iterator = controlMap.begin(); iterator != controlMap.end(); iterator++) {
        ControlPort *obj = iterator->second;
        delete obj;
    }
    // plugin instance
    lilv_instance_deactivate(instance);
    lilv_instance_free(instance);
}

State::State(const char *filename) {
    throw std::logic_error("LV2 State from file not yet implemented"); // TODO implement
}

State::State(ScriptHashIterator &stateHash) {
    UridMapper &uridMapper = PluginFactory::instance().uridMapper;
    while (stateHash.hasNext()) {
        ScriptHashPair &pair = stateHash.next();
        if (pair.key.type != STRING || pair.value.type != STRING) {
            throw std::logic_error("setState: key and value must be of type string");
        }
        uint32_t key = uridMapper.uriToId(pair.key.stringValue);
        stateMap[key] = pair.value.stringValue;
    }
}

const void *State::retrieveState(uint32_t key, size_t *size, uint32_t *type, uint32_t *flags) {
    *type = Plugin::atomTypes.stringType;
    if (stateMap.find(key) != stateMap.end()) {
        std::string &value = stateMap[key];
        *size = value.length();
        return value.c_str();
    } else {
        *size = 0;
        return 0;
    }
}

int State::getHash() {
    int hash = 0;
    for (auto const &entry : stateMap) {
        hash = hash ^ (entry.first << 1);
        int valueHash = std::hash<std::string>()(entry.second);
        hash = hash ^ (valueHash << 1);
    }
    return hash;
}

void *run_worker(void *arg) {
    ((Worker *)arg)->run();
    return 0;
}

Worker::Worker() : active(true), workResponseHandler(nullptr) {
    requestBuffer = jack_ringbuffer_create(16384);
    responseBuffer = jack_ringbuffer_create(16384);
    sem_init(&semaphore, 0, 0);
    int result = pthread_create(&thread, NULL, run_worker, this);
    if (result) {
        throw std::runtime_error("could not create LV2 worker thread");
    }
}

Worker::~Worker() {
    active.store(false);
    sem_post(&semaphore);
    pthread_join(thread, nullptr);
    jack_ringbuffer_free(requestBuffer);
    jack_ringbuffer_free(responseBuffer);
}

void Worker::run() {
    void *buffer = 0;
    while (true) {
        // wait for work
        sem_wait(&semaphore);
        if (!active) {
            break;
        }
        // read data size
        uint32_t size = 0;
        jack_ringbuffer_read(requestBuffer, (char *)&size, sizeof(size));
        // resize buffer to hold incoming data
        if (!(buffer = realloc(buffer, size))) {
            throw std::runtime_error("failed to allocate buffer for worker data");
        }
        // read data
        jack_ringbuffer_read(requestBuffer, (char *)buffer, size);
        // call back to plugin on the worker interface
        interface->work(handle, workerRespond, this, size, buffer);
    }
    free(buffer);
    pthread_exit(0);
}

LV2_Worker_Status Worker::scheduleWork(const void *data, uint32_t size) {
    if (jack_ringbuffer_write_space(requestBuffer) < size + sizeof(size)) {
        return LV2_WORKER_ERR_NO_SPACE;
    }
    jack_ringbuffer_write(requestBuffer, (const char *)&size, sizeof(size));
    jack_ringbuffer_write(requestBuffer, (const char *)data, size);
    sem_post(&semaphore); // wake up worker
    return LV2_WORKER_SUCCESS;
}

LV2_Worker_Status Worker::queueResponse(const void *data, uint32_t size) {
    jack_ringbuffer_write(responseBuffer, (const char *)&size, sizeof(size));
    jack_ringbuffer_write(responseBuffer, (const char *)data, size);
    return LV2_WORKER_SUCCESS;
}

void Worker::respond() {
    uint32_t read_space = jack_ringbuffer_read_space(responseBuffer);
    while (read_space) {
        uint32_t size = 0;
        jack_ringbuffer_read(responseBuffer, (char *)&size, sizeof(size));
        jack_ringbuffer_read(responseBuffer, response, size);
        interface->work_response(handle, size, response);
        if (workResponseHandler) {
            MethodQueue::instance().dispatch(new WorkResponseClosure(*workResponseHandler));
        }
        read_space -= sizeof(size) + size;
    }
}

void Worker::setInstance(LilvInstance *instance) {
    handle = instance->lv2_handle;
    interface =
        (LV2_Worker_Interface *)lilv_instance_get_extension_data(instance, LV2_WORKER__interface);
}

void Worker::onWorkResponse(ScriptFunction &handler) {
    // TODO: add validation
    //    if (handler.numargs() != 3) {
    //        throw std::logic_error("onControl handler should take two arguments");
    //    }
    workResponseHandler.store(new ScriptFunction(handler));
}

void Plugin::setPortValue(const char *portname, const void *value, uint32_t type) {
    ControlPort *lv2Port = controlMap[portname];
    if (lv2Port) {
        if (type == atomTypes.floatType) {
            lv2Port->value = *(const float *)value;
        } else if (type == atomTypes.doubleType) {
            lv2Port->value = *(const double *)value;
        } else if (type == atomTypes.intType) {
            lv2Port->value = *(const int32_t *)value;
        } else if (type == atomTypes.longType) {
            lv2Port->value = *(const int64_t *)value;
        } else {
            std::cerr << "warning: attempt to set control port with unknown type: " << type
                      << std::endl;
        }
    } else {
        std::cerr << "warning: attempt to set unknown control port '" << portname << "'"
                  << std::endl;
    }
}

void *Plugin::portValue(const char *portname, uint32_t *type, uint32_t *size) {
    ControlPort *lv2Port = controlMap[portname];
    if (lv2Port) {
        *size = sizeof(float);
        *type = Plugin::atomTypes.floatType;
        return &lv2Port->value;
    } else {
        std::cerr << "warning: attempt to get unknown control port '" << portname << "'"
                  << std::endl;
    }
    return 0;
}

/**
 * Looks up an Lv2ControlPort pointer that corresponds to a string symbol
 *
 * Runs in the script thread.
 */
ControlPort *Plugin::getPort(const char *symbol) {
    ControlPort *port = controlMap[symbol];
    if (!port) {
        throw std::logic_error(std::string("Plugin has no such port: ") + symbol);
    }
    return port;
}

/**
 * Update the Lv2ControlPort value.
 *
 * Runs in the script thread.
 */
void Plugin::control(const char *symbol, float value) {
    ControlPort *port = getPort(symbol);
    // TODO: check return value from push
    controlChanges.push(new ControlPortChange(port, value));
}

/**
 * Schedule a control value change.
 *
 * Runs in the script thread.
 *
 * Allocates ControlEvent objects.
 */
void Plugin::schedule(ControlChange &change, float measure, Clock &clock) {
    ControlPort *port = getPort(change.port());
    ControlPortChange *control = new ControlPortChange(port, change.value());
    controlBuffer.createEvent(control, measure, clock);
}

/**
 * Load an LV2 preset.
 *
 * Runs in the script thread.
 */
void Plugin::preset(const char *name) {
    PluginFactory &cache = PluginFactory::instance();
    LilvWorld *world = cache.getWorld();
    LV2_URID_Map *map = cache.getMap();
    const Constants &lv2Constants = cache.getConstants();
    LilvNodes *presets = lilv_plugin_get_related(type->plugin(), lv2Constants.lv2Presets);
    LilvNode *myPreset = 0;
    LILV_FOREACH(nodes, i, presets) {
        const LilvNode *presetNode = lilv_nodes_get(presets, i);
        lilv_world_load_resource(world, presetNode);
        LilvNodes *labels =
            lilv_world_find_nodes(world, presetNode, lv2Constants.lv2RdfsLabel, NULL);
        if (labels) {
            const LilvNode *label = lilv_nodes_get_first(labels);
            const char *labelString = lilv_node_as_string(label); // TODO: free?
            if (!strcmp(labelString, name)) {
                myPreset = lilv_node_duplicate(presetNode);
            }
            lilv_nodes_free(labels);
        }
    }
    lilv_nodes_free(presets);
    if (myPreset) {
        LilvState *presetState = lilv_state_new_from_world(world, map, myPreset);
        lilv_node_free(myPreset);
        if (pthread_mutex_lock(&mmutex) != 0) {
            throw std::logic_error("LV2 Plugin mutex failed to lock for preset");
        }
        lilv_state_restore(presetState, instance, bipscript::lv2::setPortValue, this, 0, NULL);
        if (pthread_mutex_unlock(&mmutex) != 0) {
            throw std::logic_error("LV2 Plugin mutex failed to unlock from preset");
        }
        lilv_state_free(presetState);
    } else {
        throw std::logic_error(std::string("Plugin ") + type->uri() +
                               " has no such preset: " + name);
    }
}

void Plugin::saveState(const char *dir) {
    PluginFactory &cache = PluginFactory::instance();

    LV2_URID_Map *map = cache.getMap();
    LilvState *const state = lilv_state_new_from_instance(
        type->plugin(), instance, map, dir, dir, dir, dir, getPortValue, this,
        LV2_STATE_IS_POD | LV2_STATE_IS_PORTABLE, NULL);

    LV2_URID_Unmap *unmap = cache.getUnmap();
    LilvWorld *world = cache.getWorld();
    if (lilv_state_save(world, map, unmap, state, NULL, dir, "state.ttl") != LV2_STATE_SUCCESS) {
        throw std::logic_error("could not save LV2 plugin state");
    }

    lilv_state_free(state);
}

bool Plugin::connectsTo(AbstractSink *sink) {
    return sink == this || audio::Sink::connectsTo(sink) || midi::Sink::connectsTo(sink);
}

bool Plugin::doProcess(samples_t time) {
    bool active = false;

    // process dependencies
    midi::Sink::process(time);

    // clear all atom inputs and outputs
    for (uint32_t i = 0; i < type->atomInputCount(); i++) {
        lv2_atom_sequence_clear(matomInput[i]);
    }
    for (uint32_t i = 0; i < type->atomOutputCount(); i++) {
        lv2_atom_sequence_clear(matomOutput[i]);
    }

    // process MIDI inputs
    for (uint32_t i = 0; i < type->midiInputCount(); i++) {
        midi::Connector &connector = midi::Sink::connector(i);
        Event<midi::RawMessage> *event = connector.firstEvent();
        while (event) {
            // lv2 event
            MidiEvent lv2Event;
            lv2Event.event.time.frames = event->start();
            // copy event data
            event->message().exportBytes(lv2Event.buffer);
            // append this event to sequence
            lv2_atom_sequence_append_event(midiInput[i], ATOM_CAPACITY, &lv2Event.event);
            event = connector.nextEvent();
        }
        if (connector.active()) {
            active = true;
        }
    }

    // process and connect audio inputs
    samples_t nframes = Clock::period();
    for (uint32_t i = 0; i < type->audioInputCount(); i++) {
        lilv_instance_connect_port(instance, type->audioInput(i).index(),
                                   inputaudio(i, time, nframes));
    }
    // immediate control changes
    ControlPortChange *change;
    while (controlChanges.pop(change)) {
        // update gain
        change->port()->value = change->value();
        controlCollector.recycle(change);
    }

    // update control port values from scheduled control events
    Event<ControlPortChange> *evt = controlBuffer.firstEvent();
    while (evt) {
        evt->message().port()->value = evt->message().value();
        // ObjectCollector::scriptCollector().recycle(evt);
        evt = controlBuffer.nextEvent();
    }

    // clear event output buffers
    for (uint32_t i = 0; i < type->midiOutputCount(); i++) {
        mmidiOutput[i]->clear();
    }

    // set up audio output buffers
    for (uint32_t i = 0; i < type->audioOutputCount(); i++) {
        lilv_instance_connect_port(instance, type->audioOutput(i).index(), audioOutput(i)->audio());
    }

    // non-blocking attempt to lock mutex, will fail if preset is loading
    if (pthread_mutex_trylock(&mmutex) == 0) {
        // run the plugin
        lilv_instance_run(instance, Clock::period());
        if (pthread_mutex_unlock(&mmutex) != 0) {
            throw std::logic_error("LV2 Plugin mutex failed to unlock from process");
        }
        // fire MIDI events
        fireMidiEvents();
    }
    // emit worker responses
    if (worker) {
        worker->respond();
    }
    grabClips(nframes);
    midi::Source::reportLevels(this, time);
    audio::Source::reportLevels(this, time);
    return active || audio::Sink::activeEvents() || audio::Source::activeEvents();
}

void Plugin::restart() {
    midi::Sink::processRestart();
    audio::Sink::processRestart();
    audio::Source::restart();
    // recycle control buffer
    controlBuffer.restart();
}

/**
 * Restore clean state on a cached plugin before reuse
 *
 * Runs in script thread.
 */
void Plugin::restartScript() {
    midi::Sink::scriptRestart();
    midi::Source::restartScript();
    // TODO: reset control values
}

void Plugin::print() {

    std::cout << "----------- plugin " << this << std::endl;
    // std::cout << atomMidiInputs.size() << " event inputs and zero outputs" << std::endl;
    //    for(uint i = 0; i < atomMidiInputs.size(); i++) {
    //        std::cout << "event input port " << i << " has buffer " << &atomMidiInputs[i].buffer
    //        << std::endl;
    //    }

    //    std::cout << numAudioInputs << " audio inputs and " << numAudioOutputs << " outputs" <<
    //    std::endl; for(uint i = 0; i < numAudioInputs; i++) {
    //        std::cout << "audio input port " << i << " is connected to " << audioInputs[i].output
    //        << std::endl;
    //        // std::cout << "\thas input " << audioInputs[i].input << std::endl;
    //    }

    //    for(uint i = 0; i < numAudioOutputs; i++) {
    //        std::cout << "audio output port " << i << " is " << &audioOutputs[i] << std::endl;
    //    }
}

PluginType::PluginType(std::string &uri, const LilvPlugin *plugin, const Constants &uris)
    : muri(uri), mplugin(plugin), //
      maudioInputCount(
          lilv_plugin_get_num_ports_of_class(plugin, uris.lv2AudioPort, uris.lv2InputPort, 0)), //
      maudioOutputCount(
          lilv_plugin_get_num_ports_of_class(plugin, uris.lv2AudioPort, uris.lv2OutputPort, 0)),
      mcontrolInputCount(
          lilv_plugin_get_num_ports_of_class(plugin, uris.lv2ControlPort, uris.lv2InputPort, 0)),
      mcontrolOutputCount(
          lilv_plugin_get_num_ports_of_class(plugin, uris.lv2ControlPort, uris.lv2OutputPort, 0)),
      matomInputCount(
          lilv_plugin_get_num_ports_of_class(plugin, uris.lv2AtomPort, uris.lv2InputPort, 0)),
      matomOutputCount(
          lilv_plugin_get_num_ports_of_class(plugin, uris.lv2AtomPort, uris.lv2OutputPort, 0)),
      mmidiInputCount(0), mmidiOutputCount(0) {
    // get plugin name
    LilvNode *namenode = lilv_plugin_get_name(plugin);
    mname = lilv_node_as_string(namenode);
    lilv_free(namenode);
    // initialize port structures
    maudioInput = new PortDef[maudioInputCount];
    maudioOutput = new PortDef[maudioOutputCount];
    mcontrolInput = new ControlPortDef[mcontrolInputCount];
    mcontrolOutput = new ControlPortDef[mcontrolOutputCount];
    matomInput = new AtomPortDef[matomInputCount];
    matomOutput = new AtomPortDef[matomOutputCount];

    uint32_t numPorts = lilv_plugin_get_num_ports(plugin);
    uint32_t audioInputCounter = 0;
    uint32_t audioOutputCounter = 0;
    uint32_t controlInputCounter = 0;
    uint32_t controlOutputCounter = 0;
    uint32_t eventInputCounter = 0;
    uint32_t eventOutputCounter = 0;
    for (uint32_t i = 0; i < numPorts; i++) {
        const LilvPort *port = lilv_plugin_get_port_by_index(plugin, i);
        if (lilv_port_is_a(plugin, port, uris.lv2AudioPort)) {
            if (lilv_port_is_a(plugin, port, uris.lv2InputPort)) {
                maudioInput[audioInputCounter++].index(i);
            } else if (lilv_port_is_a(plugin, port, uris.lv2OutputPort)) {
                maudioOutput[audioOutputCounter++].index(i);
            }
        } else if (lilv_port_is_a(plugin, port, uris.lv2ControlPort)) {
            if (lilv_port_is_a(plugin, port, uris.lv2InputPort)) {
                ControlPortDef &newPort = mcontrolInput[controlInputCounter++];
                newPort.index(i);
                // get control name
                const LilvNode *symbol = lilv_port_get_symbol(plugin, port);
                newPort.name(lilv_node_as_string(symbol));
                // get default, max/min values
                LilvNode *dfault, *minimum, *maximum;
                lilv_port_get_range(plugin, port, &dfault, &minimum, &maximum);
                newPort.dfault(dfault ? lilv_node_as_float(dfault) : 0);
                newPort.minimum(lilv_node_as_float(minimum));
                newPort.maximum(lilv_node_as_float(maximum));
                lilv_free(dfault);
                lilv_free(minimum);
                lilv_free(maximum);
            } else if (lilv_port_is_a(plugin, port, uris.lv2OutputPort)) {
                mcontrolOutput[controlOutputCounter++].index(i);
            }
        } else if (lilv_port_is_a(plugin, port, uris.lv2AtomPort)) {
            LilvNodes *atomBufferType = lilv_port_get_value(plugin, port, uris.lv2AtomBufferType);
            LilvNodes *atomSupports = lilv_port_get_value(plugin, port, uris.lv2AtomSupports);
            if (lilv_port_is_a(plugin, port, uris.lv2InputPort)) {
                AtomPortDef &port = matomInput[eventInputCounter++];
                port.index(i);
                if (lilv_nodes_contains(atomBufferType, uris.lv2AtomSequence) &&
                    lilv_nodes_contains(atomSupports, uris.lv2MidiEvent)) {
                    port.midi(true);
                    mmidiInputCount++;
                }
            } else if (lilv_port_is_a(plugin, port, uris.lv2OutputPort)) {
                AtomPortDef &port = matomOutput[eventOutputCounter++];
                port.index(i);
                if (lilv_nodes_contains(atomBufferType, uris.lv2AtomSequence) &&
                    lilv_nodes_contains(atomSupports, uris.lv2MidiEvent)) {
                    port.midi(true);
                    mmidiOutputCount++;
                }
            }
            lilv_nodes_free(atomBufferType);
            lilv_nodes_free(atomSupports);
        }
    }
}

PluginType::~PluginType() {
    delete[] maudioInput;
    delete[] maudioOutput;
    delete[] mcontrolInput;
    delete[] mcontrolOutput;
    delete[] matomInput;
    delete[] matomOutput;
}

AtomTypes Plugin::atomTypes;

PluginFactory::PluginFactory() : world(lilv_world_new()), mplugins(nullptr), lv2Constants(world) {

    // supported features
    supported[LV2_URID__map] = true;
    supported[LV2_URID__unmap] = true;
    supported[LV2_OPTIONS__options] = true;
    supported[LV2_BUF_SIZE__boundedBlockLength] = true;
    supported[LV2_STATE__mapPath] = true;
    supported[LV2_STATE__loadDefaultState] = true;
    supported[LV2_WORKER__schedule] = true;

    // URID feature (map/unmap)
    map.map = &uridMap;
    map.handle = &uridMapper;
    static LV2_Feature uridMapFeature = {LV2_URID__map, &map};
    mlv2Features[0] = &uridMapFeature;

    unmap.unmap = &uridUnmap;
    unmap.handle = &uridMapper;
    static LV2_Feature uridUnmapFeature = {LV2_URID__unmap, &unmap};
    mlv2Features[1] = &uridUnmapFeature;

    // Options feature
    options[0] = {LV2_OPTIONS_INSTANCE,
                  0,
                  uridMapper.uriToId(LV2_BUF_SIZE__minBlockLength),
                  sizeof(float),
                  uridMapper.uriToId(LV2_ATOM__Int),
                  &blockLength};
    options[1] = {LV2_OPTIONS_INSTANCE,
                  0,
                  uridMapper.uriToId(LV2_BUF_SIZE__maxBlockLength),
                  sizeof(float),
                  uridMapper.uriToId(LV2_ATOM__Int),
                  &blockLength};
    options[2] = {LV2_OPTIONS_INSTANCE, 0, 0, 0, 0, NULL};

    static LV2_Feature optionsFeature = {LV2_OPTIONS__options, &options};
    mlv2Features[2] = &optionsFeature;

    static LV2_Feature boundedBlockFeature = {LV2_BUF_SIZE__boundedBlockLength, NULL};
    mlv2Features[3] = &boundedBlockFeature;

    // state map path feature
    mapPath.handle = &pathMapper;
    mapPath.absolute_path = &mapAbsolutePath;
    mapPath.abstract_path = &mapAbstractPath;
    static LV2_Feature stateMapPathFeature = {LV2_STATE__mapPath, &mapPath};
    mlv2Features[4] = &stateMapPathFeature;

    static LV2_Feature defaultStateFeature = {LV2_STATE__loadDefaultState, NULL};
    mlv2Features[5] = &defaultStateFeature;

    // URIDs
    Plugin::atomTypes.intType = uridMapper.uriToId(LV2_ATOM__Int);
    Plugin::atomTypes.longType = uridMapper.uriToId(LV2_ATOM__Long);
    Plugin::atomTypes.floatType = uridMapper.uriToId(LV2_ATOM__Float);
    Plugin::atomTypes.doubleType = uridMapper.uriToId(LV2_ATOM__Double);
    Plugin::atomTypes.stringType = uridMapper.uriToId(LV2_ATOM__String);
    MidiEvent::midiEventTypeId = uridMapper.uriToId(LILV_URI_MIDI_EVENT);
}

const PluginType *PluginFactory::getPluginType(std::string &uriString) {
    // look in cache for plugin type by uri
    const PluginType *type = typeMap[uriString];
    if (!type) {
        // find lv2 plugin in lilv
        LilvNode *lilvUri = lilv_new_uri(world, uriString.c_str());
        const LilvPlugin *lilvPlugin = lilv_plugins_get_by_uri(plugins(), lilvUri);
        lilv_node_free(lilvUri);

        if (!lilvPlugin) {
            throw std::logic_error(std::string("Plugin is not installed on this system: ") +
                                   uriString);
        }

        // check that required features are supported
        LilvNodes *features = lilv_plugin_get_required_features(lilvPlugin);
        for (LilvIter *f = lilv_nodes_begin(features); !lilv_nodes_is_end(features, f);
             f = lilv_nodes_next(features, f)) {
            const char *featureUri = lilv_node_as_uri(lilv_nodes_get(features, f));
            if (!supported[featureUri]) {
                throw std::logic_error(std::string("Plugin ") + uriString +
                                       " requires unsupported feature: " + featureUri);
            }
        }
        lilv_nodes_free(features);
        type = new PluginType(uriString, lilvPlugin, lv2Constants);
        typeMap[uriString] = type;
    }
    return type;
}

LV2_Feature *PluginFactory::workFeature(Worker *worker) {
    // worker schedule feature
    LV2_Worker_Schedule *schedule = new LV2_Worker_Schedule;
    schedule->handle = worker;
    schedule->schedule_work = &scheduleWork;
    LV2_Feature *ret = new LV2_Feature;
    ret->URI = LV2_WORKER__schedule;
    ret->data = schedule;
    return ret;
}

const LilvPlugins *PluginFactory::plugins() {
    if (!mplugins) {
        // plugins
        lilv_world_load_all(world);
        mplugins = lilv_world_get_all_plugins(world);
    }
    return mplugins;
}

Plugin *PluginFactory::newPlugin(const char *uri, const char *preset, State *state) {
    // make sure audio engine started
    AudioEngine::instance();

    // create a unique key for uri/preset/state
    std::string keyString(uri);
    if (preset) {
        keyString.append(":");
        keyString.append(preset);
    }
    std::size_t hash = std::hash<std::string>()(keyString);
    if (state) {
        hash = hash ^ (state->getHash() << 1);
    }

    // count tells how many instances of this type of plugin
    int count = ++instanceCount[hash];

    // return cached plugin instance if available
    int key = hash + count;
    Plugin *cachedPlugin = findObject(key);
    if (cachedPlugin) {
        return cachedPlugin;
    }

    std::string uriString(uri);
    const PluginType *type = getPluginType(uriString);

    // create features array
    LV2_Feature **lv2Features = new LV2_Feature *[8];
    for (int i = 0; i < 6; i++) {
        lv2Features[i] = mlv2Features[i];
    }
    // create worker if required
    Worker *worker = nullptr;
    if (lilv_plugin_has_feature(type->plugin(), lv2Constants.lv2WorkerSchedule) &&
        lilv_plugin_has_extension_data(type->plugin(), lv2Constants.lv2WorkerInterface)) {
        worker = new Worker();
        lv2Features[6] = workFeature(worker);
        lv2Features[7] = nullptr;
    } else {
        lv2Features[6] = nullptr;
    }

    // instantiate
    LilvInstance *instance = lilv_plugin_instantiate(type->plugin(), msampleRate, lv2Features);

    // create plugin object
    Plugin *plugin = new Plugin(type, instance, lv2Features, worker);

    // restore baseline default state
    LilvState *defaultState =
        lilv_state_new_from_world(world, &map, lilv_plugin_get_uri(type->plugin()));
    lilv_state_restore(defaultState, instance, setPortValue, plugin, 0, lv2Features);
    lilv_state_free(defaultState);

    // find and restore preset
    if (preset) {
        plugin->preset(preset);
    }
    // restore state
    else if (state) { // TODO: what if state is requested on a plugin that doesn't support?
        LV2_State_Interface *iState =
            (LV2_State_Interface *)lilv_instance_get_extension_data(instance, LV2_STATE__interface);
        if (iState) {
            LV2_State_Status status = iState->restore(instance->lv2_handle, &stateRetrieve,
                                                      (LV2_State_Handle)state, 0, lv2Features);
            if (status != LV2_STATE_SUCCESS) {
                throw std::logic_error(std::string("Plugin ") + uriString +
                                       " setting state failed with code " + std::to_string(status));
            }
        }
    }

    // activate plugin
    lilv_instance_activate(instance);

    // add to cache
    registerObject(key, plugin);

    return plugin;
}

Plugin *PluginFactory::newPlugin(io::File &file) {
    // make sure audio engine started
    AudioEngine::instance();
    // create a unique key for file + time
    std::string keyString(file.path());
    keyString.append(":");
    keyString.append(file.modified());
    std::size_t hash = std::hash<std::string>()(keyString);

    // return cached plugin instance if available
    int count = ++instanceCount[hash];
    int key = hash + count;
    Plugin *cachedPlugin = findObject(key);
    if (cachedPlugin) {
        return cachedPlugin;
    }

    // create new plugin from preset file
    LilvState *state = lilv_state_new_from_file(world, &map, NULL, file.path().c_str());
    if (!state) {
        string mesg("LV2 plugin failed to load state from file: ");
        mesg.append(file.path());
        throw std::logic_error(mesg);
    }
    const LilvNode *uri_node = lilv_state_get_plugin_uri(state);
    const char *uri = lilv_node_as_string(uri_node);

    std::string uriString(uri);
    const PluginType *type = getPluginType(uriString);

    // create features array
    LV2_Feature **lv2Features = new LV2_Feature *[8];
    for (int i = 0; i <= 5; i++) {
        lv2Features[i] = mlv2Features[i];
    }
    // create worker if required
    Worker *worker = nullptr;
    if (lilv_plugin_has_feature(type->plugin(), lv2Constants.lv2WorkerSchedule) &&
        lilv_plugin_has_extension_data(type->plugin(), lv2Constants.lv2WorkerInterface)) {
        worker = new Worker();
        lv2Features[6] = workFeature(worker);
        lv2Features[7] = nullptr;
    } else {
        lv2Features[6] = nullptr;
    }

    // instantiate
    LilvInstance *instance = lilv_plugin_instantiate(type->plugin(), msampleRate, mlv2Features);

    // create plugin object
    Plugin *plugin = new Plugin(type, instance, lv2Features, worker);

    // state from file
    lilv_state_restore(state, instance, setPortValue, plugin, 0, mlv2Features);

    // activate plugin
    lilv_instance_activate(instance);

    // add to cache
    registerObject(key, plugin);

    return plugin;
}

PluginFactory::~PluginFactory() {
    deleteInstances();
    // delete types
    for (auto iter = typeMap.begin(); iter != typeMap.end(); iter++) {
        delete iter->second;
    }
    lilv_world_free(world);
}

} // namespace lv2
} // namespace bipscript
