/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef MIDICONNECTION_H
#define MIDICONNECTION_H

#include <atomic>

#include "bindmidi.h"
#include "bindutil.h"
#include "boundarray.h"
#include "methodqueue.h"
#include "midievent.h"
#include "processclosure.h"
#include "source.h"
#include "statusserver.h"

namespace bipscript {
namespace midi {

template <class T> class MessageEventClosure : public ProcessClosure {
    T mmessage;
    const float mmeasure;

  protected:
    void addParameters() {
        MessageEvent *event = new MessageEvent(*(new T(mmessage)), mmeasure);
        BindUtil::PushRef(vm, event, binding::MessageEventObject);
    }

  public:
    MessageEventClosure(ScriptFunction function, T &mesg, RawEvent *evt)
        : ProcessClosure(function, evt->start()), mmessage(mesg), mmeasure(evt->measure()) {}
};

class Meter {
    bool mon;
    samples_t lasttime;
    static const samples_t REPORT_PERIOD = 11025; // 44100
  public:
    Meter() : mon(false) {}
    void update(void *node, uint16_t index, bool events, samples_t time) {
        StatusServer &statusserver = StatusServer::instance();
        if (events) {
            if (!mon) {
                statusserver.midiLevel(node, index, true);
                mon = true;
            }
            lasttime = time;
        } else if (mon && time - lasttime > REPORT_PERIOD) {
            statusserver.midiLevel(node, index, false);
            mon = false;
        }
    }
};

class Source;

class MessageCallback : public Listable<MessageCallback> {
    ScriptFunction *mhandler;
    uint8_t mtype;

  public:
    MessageCallback(ScriptFunction &handler, uint8_t type)
        : mhandler(new ScriptFunction(handler)), mtype(type) {}
    uint8_t type() { return mtype; }
    ScriptFunction *handler() { return mhandler; }
};

class Connection {
    Source *msource;
    Meter mmeter;
    ThreadQueue<MessageCallback> mhandlerq; // script thread -> process thread
    List<MessageCallback> mhandlerlist;

  public:
    Connection(Source *source) : msource(source), mhandlerq(64) {}
    ~Connection() {}
    Source *source() { return msource; }
    // methods to access events in current period
    virtual RawEvent *firstEvent() = 0;
    virtual RawEvent *nextEvent() = 0;

    void onMessage(ScriptFunction &handler) {
        checkHandler("onMessage", handler);
        onMessage(handler, 0);
    }
    void onControl(ScriptFunction &handler) {
        checkHandler("onControl", handler);
        onMessage(handler, Control::COMMAND);
    }
    void onNoteOn(ScriptFunction &handler) {
        checkHandler("onNoteOn", handler);
        onMessage(handler, NoteOn::COMMAND);
    }
    void onNoteOff(ScriptFunction &handler) {
        checkHandler("onNoteOff", handler);
        onMessage(handler, NoteOff::COMMAND);
    }
    void onProgramChange(ScriptFunction &handler) {
        checkHandler("onProgramChange", handler);
        onMessage(handler, ProgramChange::COMMAND);
    }
    void onSysex(ScriptFunction &handler) {
        checkHandler("onSysex", handler);
        onMessage(handler, Sysex::COMMAND);
    }
    void fireEvents() {
        MessageCallback *handler;
        while (mhandlerq.pop(handler)) {
            mhandlerlist.add(handler);
        }
        MethodQueue &mq = MethodQueue::instance();
        RawEvent *event = firstEvent();
        while (event) {
            RawMessage &mesg = event->message();
            switch (mesg.command()) {
            // channel messages
            case Control::COMMAND:
                callChannelHandlers<Control>(mesg, mq, event);
                break;
            case NoteOn::COMMAND:
                callChannelHandlers<NoteOn>(mesg, mq, event);
                break;
            case NoteOff::COMMAND:
                callChannelHandlers<NoteOff>(mesg, mq, event);
                break;
            case PitchBend::COMMAND:
                callChannelHandlers<PitchBend>(mesg, mq, event);
                break;
            case ProgramChange::COMMAND:
                callChannelHandlers<ProgramChange>(mesg, mq, event);
                break;
            // realtime messages
            case ClockStart::COMMAND:
                callRealtimeHandlers<ClockStart>(mq, event);
                break;
            case ClockContinue::COMMAND:
                callRealtimeHandlers<ClockContinue>(mq, event);
                break;
            case ClockStop::COMMAND:
                callRealtimeHandlers<ClockStop>(mq, event);
                break;
            // sysex
            case Sysex::COMMAND: {
                MessageCallback *callback = mhandlerlist.first();
                while (callback) {
                    if (!callback->type() || callback->type() == Sysex::COMMAND) {
                        Sysex sysex(mesg.bytes(), mesg.size());
                        mq.dispatch(
                            new MessageEventClosure<Sysex>(*callback->handler(), sysex, event));
                    }
                    callback = mhandlerlist.next(callback);
                }
                break;
            }
            default:
                break;
            }
            event = nextEvent();
        }
    }

    void restartScript() {
        ScriptFunctionCollector &collector = ScriptFunctionCollector::instance();
        MessageCallback *evt = mhandlerlist.first();
        while (evt) {
            MessageCallback *done = evt;
            evt = mhandlerlist.pop();
            collector.recycle(done->handler()); // TODO: callback object too
        }
    }

    void reportLevels(void *node, uint16_t index, samples_t time) {
        mmeter.update(node, index, firstEvent() != nullptr, time);
    }

  private:
    // check incoming message against handlers
    template <class T>
    void callChannelHandlers(RawMessage &mesg, MethodQueue &mq, RawEvent *event) {
        MessageCallback *callback = mhandlerlist.first();
        while (callback) {
            if (!callback->type() || callback->type() == T::COMMAND) {
                T arg(mesg.bytes());
                mq.dispatch(new MessageEventClosure<T>(*callback->handler(), arg, event));
            }
            callback = mhandlerlist.next(callback);
        }
    }
    template <class T> void callRealtimeHandlers(MethodQueue &mq, RawEvent *event) {
        MessageCallback *callback = mhandlerlist.first();
        while (callback) {
            if (!callback->type() || callback->type() == T::COMMAND) {
                T arg;
                mq.dispatch(new MessageEventClosure<T>(*callback->handler(), arg, event));
            }
            callback = mhandlerlist.next(callback);
        }
    }
    // check user-provided message handler
    void checkHandler(const char *name, ScriptFunction &handler) {
        if (handler.numargs() != 2) {
            std::string mesg(name);
            mesg += " handler should take a single Midi.MessageEvent argument";
            throw std::logic_error(mesg);
        }
    }
    // set the handler for a given message type
    void onMessage(ScriptFunction &handler, uint8_t type) {
        mhandlerlist.add(new MessageCallback(handler, type));
    }
};

class Source : virtual public AbstractSource {
    BoundArray<Connection> bus;
    uint32_t mchannels;

  protected:
    Connection *midiOutput(uint32_t index) { return &bus[index]; }
    void midiOutput(Connection *connection, uint32_t index) {
        bus.set(index, connection, binding::ConnectionObject);
    }

  public:
    Source(uint32_t channels) : bus(channels), mchannels(channels) {}
    BoundArray<Connection> &midioutbus() { return bus; }
    HSQOBJECT midiout() { return bus.hsqobject(); }
    uint32_t channels() { return mchannels; }

    // event handlers
    void onMessage(ScriptFunction &handler) {
        for (uint32_t i = 0; i < mchannels; i++) {
            bus[i].onMessage(handler);
        }
    }
    void onControl(ScriptFunction &handler) {
        for (uint32_t i = 0; i < mchannels; i++) {
            bus[i].onControl(handler);
        }
    }
    void onNoteOn(ScriptFunction &handler) {
        for (uint32_t i = 0; i < mchannels; i++) {
            bus[i].onNoteOn(handler);
        }
    }
    void onNoteOff(ScriptFunction &handler) {
        for (uint32_t i = 0; i < mchannels; i++) {
            bus[i].onNoteOff(handler);
        }
    }
    void onProgramChange(ScriptFunction &handler) {
        for (uint32_t i = 0; i < mchannels; i++) {
            bus[i].onProgramChange(handler);
        }
    }
    void onSysex(ScriptFunction &handler) {
        for (uint32_t i = 0; i < mchannels; i++) {
            bus[i].onSysex(handler);
        }
    }

  protected:
    void fireMidiEvents() {
        for (uint32_t i = 0; i < mchannels; i++) {
            bus[i].fireEvents();
        }
    }
    void restartScript() {
        for (uint32_t i = 0; i < mchannels; i++) {
            bus[i].restartScript();
        }
    }
    void reportConnections(void *node) {
        void *array[mchannels];
        for (uint32_t i = 0; i < mchannels; i++) {
            array[i] = &bus[i];
        }
        StatusServer::instance().midiconnections(node, array, mchannels);
    }
    void reportLevels(void *node, samples_t time) {
        StatusServer &statusserver = StatusServer::instance();
        if (statusserver.active()) {
            for (uint32_t i = 0; i < mchannels; i++) {
                bus[i].reportLevels(node, i, time);
            }
        }
    }
};

} // namespace midi
} // namespace bipscript

#endif // MIDICONNECTION_H
