/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef MIDIEVENT_H
#define MIDIEVENT_H

#include <io.h>

#include "event.h"
#include "midimessage.h"
#include "module_midi.h"
#include "refcountable.h"

namespace bipscript {
namespace midi {

class RawMessage {
    uint8_t *mbytes;
    uint16_t msize;

  public:
    RawMessage(uint16_t size) : msize(size) { mbytes = new uint8_t[size]; }
    ~RawMessage() { delete[] mbytes; }
    friend std::ostream &operator<<(std::ostream &out, RawMessage &evt) {
        switch (evt.command()) {
        case 0x80:
            out << "NOTE OFF ";
            break;
        case 0x90:
            out << "NOTE ON ";
            break;
        case 0xB0:
            out << "CONTROL ";
            break;
        default:
            out << "UNK[" << (int)evt.command() << "] ";
            break;
        }
        return out;
    }
    uint16_t size() { return msize; }
    void size(uint16_t size) { msize = size; }
    uint8_t command() { return mbytes[0] & 0xF0; }
    uint8_t *bytes() { return mbytes; }
    void exportBytes(uint8_t *buffer) { memcpy(buffer, mbytes, msize); }
    void importBytes(const uint8_t *buffer, uint16_t size) {
        memcpy(mbytes, buffer, size);
        msize = size;
    }
};

typedef Event<RawMessage> RawEvent;

class MessageEvent : public RefCountable<MessageEvent> {
    Message *mmessage;
    const float mmeasure;
    Event<RawMessage> *mptr;

  public:
    MessageEvent(Message &message, float measure)
        : mmessage(message.addref()), mmeasure(measure), mptr(nullptr) {}
    ~MessageEvent();
    Message *message() const { return mmessage; }
    float measure() const { return mmeasure; }
    bool operator<(MessageEvent &other);
    static bool cmp(MessageEvent *e1, MessageEvent *e2) { return *e1 < *e2; }
    void ptr(Event<RawMessage> *ptr) { mptr = ptr; }
    Event<RawMessage> *ptr() { return mptr; }
};

class NoteEvent : public RefCountable<NoteEvent> {
    Note *const mnote;
    const float mmeasure;
    Event<RawMessage> *monptr;
    Event<RawMessage> *moffptr;

  public:
    NoteEvent(Note &note, const float measure) : mnote(note.addref()), mmeasure(measure) {}
    NoteEvent(const NoteEvent &copy) : mnote(copy.mnote->addref()), mmeasure(copy.mmeasure) {}
    ~NoteEvent();
    float measure() const { return mmeasure; }
    // for internal usage
    Note *note() const { return mnote; }
    void ptr(Event<RawMessage> *ptr1, Event<RawMessage> *ptr2) {
        monptr = ptr1;
        moffptr = ptr2;
    }
    Event<RawMessage> *onptr() { return monptr; }
    Event<RawMessage> *offptr() { return moffptr; }
    bool operator<(NoteEvent &other) { return mmeasure < other.mmeasure; }
    static bool cmp(NoteEvent *e1, NoteEvent *e2) { return *e1 < *e2; }
};

} // namespace midi
} // namespace bipscript

#endif // MIDIEVENT_H
