/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <cmath>
#include <cstring>
#include <fstream>
#include <iostream>

#include "bindmidi.h"
#include "midifile.h"

extern "C" {
#include "midilib/midifile.h"
}

namespace bipscript {
namespace midi {

File::File() : mppq(480), mtype(1) {
    mtrack = new ScriptArray<Sequence>(0, binding::SequenceObject);
}

File::File(std::vector<Sequence *> seqs) : mppq(480), mtype(1) {
    mtrack = new ScriptArray<Sequence>(seqs.size(), binding::SequenceObject);
    for (uint32_t i = 0; i < seqs.size(); i++) {
        mtrack->set(i, seqs[i]);
    }
}

File::File(std::string path) {
    MIDI_FILE *midifile = midiFileOpen(path.c_str());
    if (!midifile) {
        std::string mesg("Could not read MIDI file: ");
        mesg.append(path);
        throw std::logic_error(mesg);
    }
    uint32_t tracks = midiReadGetNumTracks(midifile);
    // create track objects
    mtrack = new ScriptArray<Sequence>(tracks, binding::SequenceObject);
    mppq = midiFileGetPPQN(midifile);
    mtype = midiFileGetVersion(midifile);
    std::vector<TpmMapEntry> tpmMap;
    MIDI_MSG msg;
    midiReadInitMessage(&msg);
    for (uint32_t t = 0; t < tracks; t++) {
        Sequence *track = new Sequence();
        mtrack->set(t, track);
        double tpm = mppq * 4;
        double measure = 1.0;
        uint32_t lasttick = 0;
        uint32_t mapindex = 0;
        while (midiReadGetNextMessage(midifile, t, &msg)) {
            // MIDI type 1: apply time sig changes from track 0 to all tracks
            if (mtype == 1 && t > 0) {
                while (mapindex < tpmMap.size() && msg.dwAbsPos >= tpmMap[mapindex].tick) {
                    TpmMapEntry entry = tpmMap[mapindex];
                    measure += (entry.tick - lasttick) / tpm;
                    lasttick = entry.tick;
                    tpm = entry.tpm;
                    mapindex++;
                }
                measure += (msg.dwAbsPos - lasttick) / tpm;
                lasttick = msg.dwAbsPos;
            } else {
                measure += msg.dt / tpm;
            }
            int type;
            if (msg.bImpliedMsg) {
                type = msg.iImpliedMsg;
            } else {
                type = msg.iType;
            }
            if (type == msgNoteOn) {
                NoteOn *note = new NoteOn(msg.MsgData.NoteOn.iNote, msg.MsgData.NoteOn.iVolume,
                                          msg.MsgData.NoteOn.iChannel - 1);
                track->add(*note, measure);
            } else if (type == msgNoteOff) {
                NoteOff *noteoff =
                    new NoteOff(msg.MsgData.NoteOff.iNote, msg.MsgData.NoteOff.iVolume,
                                msg.MsgData.NoteOff.iChannel - 1);
                track->add(*noteoff, measure);
            } else if (type == msgControlChange) {
                Control *cc = new Control(msg.MsgData.NoteParameter.iControl,
                                          msg.MsgData.NoteParameter.iParam,
                                          msg.MsgData.NoteParameter.iChannel - 1);
                track->add(*cc, measure);
            } else if (type == msgPatchChange) {
                ProgramChange *pc = new ProgramChange(msg.MsgData.ChangeProgram.iProgram,
                                                      msg.MsgData.ChangeProgram.iChannel - 1);
                track->add(*pc, measure);
            } else if (type == msgSetPitchWheel) {
                PitchBend *pb = new PitchBend(msg.MsgData.PitchWheel.iPitch,
                                              msg.MsgData.PitchWheel.iChannel - 1);
                track->add(*pb, measure);
            } else if (type == msgMetaEvent) {
                tMIDI_META metatype = msg.MsgData.MetaEvent.iType;
                if (metatype == tMIDI_META::metaSetTempo) {
                    TempoChange *tempo = new TempoChange(msg.MsgData.MetaEvent.Data.Tempo.fBPM);
                    track->add(*tempo, measure);
                } else if (metatype == tMIDI_META::metaKeySig) {
                    auto key = msg.MsgData.MetaEvent.Data.KeySig;
                    KeySignature *sig = new KeySignature(key.iKey, key.iMinor);
                    track->add(*sig, measure);
                } else if (metatype == tMIDI_META::metaTimeSig) {
                    auto sig = msg.MsgData.MetaEvent.Data.TimeSig;
                    TimeSignature *tempo = new TimeSignature(
                        sig.iNom, sig.iDenom, sig.iClockInMetroTick, sig.iNotated32nds);
                    track->add(*tempo, measure);
                    // update ticks per measure
                    tpm = mppq * sig.iNom * 4 / pow(2, sig.iDenom);
                    if (mtype == 1 && t == 0) { // add to tpm map
                        TpmMapEntry entry;
                        entry.tpm = tpm;
                        entry.tick = msg.dwAbsPos;
                        tpmMap.push_back(entry);
                    }
                } else if (metatype == tMIDI_META::metaTextEvent) {
                    char *body = (char *)msg.MsgData.MetaEvent.Data.Text.pData;
                    Text *text = new Text(body, msg.iMsgSize - TextMetaEvent::HSIZE);
                    track->add(*text, measure);
                } else if (metatype == tMIDI_META::metaCopyright) {
                    char *body = (char *)msg.MsgData.MetaEvent.Data.Text.pData;
                    Copyright *text = new Copyright(body, msg.iMsgSize - TextMetaEvent::HSIZE);
                    track->add(*text, measure);
                } else if (metatype == tMIDI_META::metaTrackName) {
                    char *body = (char *)msg.MsgData.MetaEvent.Data.Text.pData;
                    TrackName *text = new TrackName(body, msg.iMsgSize - TextMetaEvent::HSIZE);
                    track->add(*text, measure);
                } else if (metatype == tMIDI_META::metaInstrument) {
                    char *body = (char *)msg.MsgData.MetaEvent.Data.Text.pData;
                    Instrument *text = new Instrument(body, msg.iMsgSize - TextMetaEvent::HSIZE);
                    track->add(*text, measure);
                } else if (metatype == tMIDI_META::metaLyric) {
                    char *body = (char *)msg.MsgData.MetaEvent.Data.Text.pData;
                    Lyric *text = new Lyric(body, msg.iMsgSize - TextMetaEvent::HSIZE);
                    track->add(*text, measure);
                } else if (metatype == tMIDI_META::metaMarker) {
                    char *body = (char *)msg.MsgData.MetaEvent.Data.Text.pData;
                    Marker *text = new Marker(body, msg.iMsgSize - TextMetaEvent::HSIZE);
                    track->add(*text, measure);
                } else if (metatype == tMIDI_META::metaCuePoint) {
                    char *body = (char *)msg.MsgData.MetaEvent.Data.Text.pData;
                    CuePoint *text = new CuePoint(body, msg.iMsgSize - TextMetaEvent::HSIZE);
                    track->add(*text, measure);
                }
            }
        }
    }
    midiReadFreeMessage(&msg);
    midiFileClose(midifile);
}

File::~File() { delete mtrack; }

void File::save(std::string path) {
    float tpm = mppq * 4;
    MIDI_FILE *mf = midiFileCreate(path.c_str(), true); // TODO: overwrite as parameter?
    midiFileSetVersion(mf, mtype);
    if (mtype == 0 && tracks() > 1) {
        throw std::logic_error("MIDI file with multiple tracks cannot save as type 0");
    }
    midiFileSetPPQN(mf, mppq);
    std::vector<TpmMapEntry> tpmMap;
    for (uint32_t t = 0; t < tracks(); t++) {
        float measure = 1.0;
        uint32_t mapindex = 0;
        Sequence &seq = (*mtrack)[t];
        for (uint32_t e = 0; e < seq.len(); e++) {
            Message *mesg = seq[e]->message();
            int delta = 0;
            // MIDI type 1: fix delta
            if (mtype == 1 && t > 0) {
                while (mapindex < tpmMap.size() && seq[e]->measure() >= tpmMap[mapindex].measure) {
                    TpmMapEntry entry = tpmMap[mapindex];
                    delta += static_cast<int>((entry.measure - measure) * tpm + 0.5);
                    measure = entry.measure;
                    tpm = entry.tpm;
                    mapindex++;
                }
            }
            delta += static_cast<int>((seq[e]->measure() - measure) * tpm + 0.5);

            uint8_t buf[8];
            if (mesg->command() == NoteOn::COMMAND) {
                NoteOn *note = static_cast<NoteOn *>(mesg);
                note->exportBytes(buf);
                midiTrackAddRaw(mf, t, 3, buf, true, delta);
            } else if (mesg->command() == NoteOff::COMMAND) {
                NoteOff *note = static_cast<NoteOff *>(mesg);
                note->exportBytes(buf);
                midiTrackAddRaw(mf, t, 3, buf, true, delta);
            } else if (mesg->command() == Control::COMMAND) {
                Control *cc = static_cast<Control *>(mesg);
                cc->exportBytes(buf);
                midiTrackAddRaw(mf, t, 3, buf, true, delta);
            } else if (mesg->command() == PitchBend::COMMAND) {
                PitchBend *bend = static_cast<PitchBend *>(mesg);
                bend->exportBytes(buf);
                midiTrackAddRaw(mf, t, 3, buf, true, delta);
            } else if (mesg->command() == ProgramChange::COMMAND) {
                ProgramChange *pc = static_cast<ProgramChange *>(mesg);
                pc->exportBytes(buf);
                midiTrackAddRaw(mf, t, 2, buf, true, delta);
            } else if (mesg->command() == 0xFF) {
                MetaEvent *meta = static_cast<MetaEvent *>(mesg);
                uint8_t metabuf[meta->size()];
                if (meta->metatype() == TempoChange::METATYPE) {
                    TempoChange *te = static_cast<TempoChange *>(meta);
                    te->exportBytes(metabuf);
                    if (mtype == 1 && t > 0) {
                        throw std::logic_error(
                            "Type 1 MIDI file cannot save tempo change meta after track zero");
                    }
                } else if (meta->metatype() == TimeSignature::METATYPE) {
                    TimeSignature *sig = static_cast<TimeSignature *>(meta);
                    sig->exportBytes(metabuf);
                    // update tpm
                    tpm = mppq * sig->numerator() * 4 / pow(2, sig->denompower());
                    if (mtype == 1) {
                        if (t == 0) { // add to tpm map
                            TpmMapEntry entry;
                            entry.tpm = tpm;
                            entry.measure = seq[e]->measure();
                            tpmMap.push_back(entry);
                        } else {
                            throw std::logic_error(
                                "Type 1 MIDI file cannot save time signature "
                                "meta after track zero"); // TODO: this is not true
                        }
                    }
                } else {
                    meta->exportBytes(metabuf);
                }
                midiTrackAddRaw(mf, t, meta->size(), metabuf, true, delta);
            }
            measure = seq[e]->measure();
        }
    }
    midiFileClose(mf);
    free(mf);
}

void File::track(HSQOBJECT track) {
    delete mtrack;
    mtrack = new ScriptArray<Sequence>(track, binding::SequenceObject);
}

} // namespace midi
} // namespace bipscript
