/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "midipattern.h"
#include "bindmidi.h"
#include "bindutil.h"
#include "midisequence.h"

#include <sstream>

namespace bipscript {
namespace midi {

NoteEvent::~NoteEvent() {
    if (mnote->delref() == 0) {
        delete mnote;
    }
}

PatternFilter::PatternFilter(ScriptFunction &function) : ScriptFunction(function) {
    if (function.numargs() < 2 || function.numargs() > 3) {
        throw std::logic_error("Pattern filter function should take one or two arguments");
    }
}

bool PatternFilter::execute(size_t index, NoteEvent *event) {
    // push function pointer as closure
    sq_pushobject(vm, function());
    // push associated closure root as "this" pointer
    sq_getclosureroot(vm, -1);
    // add other parameters
    if (numargs() == 2) { // single parameter: note event
        BindUtil::PushRef(vm, event, binding::NoteEventObject);
    } else { // two parameters: index + note event
        sq_pushinteger(vm, index);
        BindUtil::PushRef(vm, event, binding::NoteEventObject);
    }
    // attempt to call function
    if (SQ_FAILED(sq_call(vm, numargs(), SQTrue, SQTrue))) {
        throw std::logic_error("Pattern filter method failed\n");
    }
    SQBool retval;
    if (SQ_FAILED(sq_getbool(vm, -1, &retval))) {
        throw std::logic_error("Pattern filter method did not evaluate to boolean");
    }
    // pop retval + closure
    sq_pop(vm, 2);
    return retval;
}

Pattern::Pattern(Pattern &copy) {
    for (uint32_t i = 0; i < copy.noteList.size(); i++) {
        noteList.push_back(copy.noteList[i]->addref());
    }
}

Pattern::~Pattern() {
    for (uint32_t i = 0; i < noteList.size(); i++) {
        if (noteList[i]->delref() == 0) {
            delete noteList[i];
        }
    }
}

void Pattern::add(Note &note, float measure) {
    if (measure < 1.0) {
        throw std::logic_error("Cannot add a pattern event before measure 1.0");
    }
    NoteEvent *evt = (new NoteEvent(note, measure))->addref();
    noteList.insert(std::upper_bound(noteList.begin(), noteList.end(), evt, NoteEvent::cmp), evt);
}

void Pattern::add(Pattern &pattern, float measure) {
    if (measure < 1.0) {
        throw std::logic_error("Cannot add a pattern event before measure 1.0");
    }
    for (uint32_t i = 0; i < pattern.len(); i++) {
        const NoteEvent &pn = *pattern[i];
        float nMeasure = measure + pn.measure() - 1;
        Note &note = *pn.note();
        NoteEvent *evt = (new NoteEvent(note, nMeasure))->addref();
        noteList.push_back(evt);
    }
    std::sort(noteList.begin(), noteList.end(), NoteEvent::cmp);
}

std::string Pattern::print() {
    std::stringstream sstream;
    sstream << noteList.size() << " events" << std::endl;
    for (auto it = noteList.begin(); it != noteList.end(); it++) {
        sstream << "N" << (*it)->note()->pitch();
        sstream << " duration " << (*it)->note()->duration();
        sstream << " @" << (*it)->measure();
        sstream << std::endl;
    }
    return sstream.str();
}

Pattern *Pattern::filter(ScriptFunction func) {
    Pattern *ret = new Pattern();
    PatternFilter filter(func);
    for (uint32_t i = 0; i < noteList.size(); i++) {
        NoteEvent *evt = noteList[i];
        if (filter.execute(i, evt)) {
            ret->add(*evt->note(), evt->measure());
        }
    }
    return ret;
}

Pattern *Pattern::transpose(int amount) {
    Pattern *ret = new Pattern();
    for (auto it = noteList.begin(); it != noteList.end(); it++) {
        Note *note = (*it)->note();
        int newval = note->pitch() + amount;
        if (newval < 0) {
            throw std::logic_error("Cannot transpose, note too low");
        } else if (newval > 127) {
            throw std::logic_error("Cannot transpose, note too high");
        }
        Note *transnote = new Note(newval, note->velocity(), note->duration(), note->channel());
        ret->add(transnote, (*it)->measure());
    }
    return ret;
}

Pattern *Pattern::slice(float start, float end) {
    Pattern *ret = new Pattern();
    for (auto it = noteList.begin(); it != noteList.end(); it++) {
        if ((*it)->measure() >= start && (*it)->measure() < end) {
            ret->add((*it)->note(), (*it)->measure() - start);
        }
    }
    return ret;
}

Sequence *Pattern::sequence() {
    Sequence *ret = new Sequence();
    for (unsigned int i = 0; i < noteList.size(); i++) {
        const NoteEvent *pn = noteList[i];
        const Note &note = *pn->note();

        NoteOn *noteOn = new NoteOn(note.pitch(), note.velocity(), note.channel());
        ret->add(*noteOn, pn->measure());

        NoteOff *noteOff = new NoteOff(note.pitch(), 64, note.channel());
        ret->add(*noteOff, pn->measure() + note.duration());
    }
    return ret;
}

} // namespace midi
} // namespace bipscript
