/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "midiport.h"
#include "audioengine.h"

namespace bipscript {
namespace midi {

RawEvent *MidiInputConnection::firstEvent() {
    eventCount = jack_midi_get_event_count(buffer);
    eventIndex = 0;
    return nextEvent();
}

RawEvent *MidiInputConnection::nextEvent() {
    while (eventIndex < eventCount) {
        jack_midi_event_t in_event;
        jack_midi_event_get(&in_event, buffer, eventIndex++);
        if (in_event.size <= MAX_EVENT_SIZE) {
            message->importBytes(in_event.buffer, in_event.size);
            event.start(in_event.time);
            return &event;
        }
    }
    return nullptr;
}

MidiInputPort *MidiInputPortFactory::newMidiInputPort(const char *name, const char *connectTo) {
    MidiInputPort *port = createInputPort(name);
    port->connect(connectTo);
    return port;
}

MidiInputPort *MidiInputPortFactory::newMidiInputPort(const char *name, Regex &regex) {
    MidiInputPort *port = createInputPort(name);
    port->connect(regex);
    return port;
}

MidiInputPort *MidiInputPortFactory::createInputPort(const char *name) {
    // see if port already exists in map
    int key = std::hash<std::string>()(name);
    MidiInputPort *port = findObject(key);
    if (!port) {
        // create new jack port
        jack_port_t *jackPort = AudioEngine::instance().registerMidiInputPort(name);
        if (!jackPort) {
            std::string message = "Failed to register midi input port: ";
            throw std::logic_error(message + name);
        }
        // add to map
        port = new MidiInputPort(name, jackPort);
        registerObject(key, port);
    }
    return port;
}

bool MidiOutputPort::doProcess(samples_t time) {
    // process dependencies
    Sink::process(time);

    // grab and clear the buffer for this port
    void *port_buf = jack_port_get_buffer(mport, Clock::period());
    jack_midi_clear_buffer(port_buf);

    // get top events from buffer + connection
    Connector &connector = Sink::connector(0);
    RawEvent *event = connector.firstEvent();
    if (StatusServer::instance().active()) {
        mmeter.update(this, 0, event != nullptr, time);
    }

    // loop while events on either buffer or connection
    while (event) {
        long frame = event->start();
        jack_midi_data_t *jackEvent =
            jack_midi_event_reserve(port_buf, frame > 0 ? frame : 0, event->message().size());
        if (jackEvent) {
            event->message().exportBytes(jackEvent);
        }
        event = connector.nextEvent();
    }
    return connector.active();
}

MidiOutputPort::MidiOutputPort(const char *name, jack_port_t *jackPort) : Sink(1), mport(jackPort) {
    StatusServer &statusserver = StatusServer::instance();
    if (statusserver.active()) {
        statusserver.newMidiOutput(this, name, 1);
        midi::Sink::reportConnectors(this);
    }
}

void MidiOutputPortFactory::removeObject(MidiOutputPort *port) {
    ProcessorCache::removeObject(port);
    port->unregister();
}

MidiOutputPort *MidiOutputPortFactory::newMidiOutputPort(const char *portName,
                                                         const char *connection) {
    int key = std::hash<std::string>()(portName);
    // see if port already exists in map
    MidiOutputPort *port = findObject(key);
    if (!port) {
        // create system port
        jack_port_t *jackPort = AudioEngine::instance().registerMidiOutputPort(portName);
        if (!jackPort) {
            throw std::logic_error(std::string("Failed to register port ") + portName);
        }
        port = new MidiOutputPort(portName, jackPort);
        registerObject(key, port);
    }
    if (connection) {
        port->connect(connection);
    }
    return port;
}

MidiInputPort::MidiInputPort(const char *name, jack_port_t *jackPort)
    : Source(1), connection(this, jackPort) {
    midiOutput(&connection, 0);
    StatusServer &statusserver = StatusServer::instance();
    if (statusserver.active()) {
        statusserver.newMidiInput(this, name, 1);
        midi::Source::reportConnections(this);
    }
}

} // namespace midi
} // namespace bipscript
