/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef MIDIPORT_H
#define MIDIPORT_H

#include "midisink.h"
#include "processorcache.h"

#include <jack/jack.h>
#include <jack/midiport.h>

#include <map>
#include <set>

namespace bipscript {
namespace midi {

class MidiInputConnection : public Connection {
    jack_port_t *mport;
    void *buffer;
    RawMessage *message;
    RawEvent event;
    uint32_t eventCount;
    uint32_t eventIndex; // iterator
    static const size_t MAX_EVENT_SIZE = 32768;

  public:
    MidiInputConnection(Source *source, jack_port_t *jackPort)
        : Connection(source), mport(jackPort), message(new RawMessage(MAX_EVENT_SIZE)),
          event(message, Command::Type::TRANSIENT, 1.0) {}
    void process() { buffer = jack_port_get_buffer(mport, Clock::period()); }
    void connect(const char *name) { AudioEngine::instance().connectPort(name, mport); }
    void connect(Regex &regex) {
        AudioEngine::instance().connectPort(regex, mport, JACK_DEFAULT_MIDI_TYPE);
    }
    void disconnect() { AudioEngine::instance().disconnectOutputPorts(mport); }
    void disconnect(const char *name) { AudioEngine::instance().disconnectPort(name, mport); }
    RawEvent *firstEvent() override;
    RawEvent *nextEvent() override;
};

class MidiInputPort : public Source {
    MidiInputConnection connection;

  public:
    MidiInputPort(const char *name, jack_port_t *jackPort);
    void connect(const char *name) { connection.connect(name); }
    void connect(Regex &regex) { connection.connect(regex); }
    void disconnect() { connection.disconnect(); }
    void disconnect(const char *name) { connection.disconnect(name); }
    // Source interface
    bool connectsTo(AbstractSink *) override { return false; }
    bool doProcess(samples_t time) override {
        connection.process();
        fireMidiEvents();
        reportLevels(this, time);
        return false;
    }
    void restart() override {}
    void restartScript() override { midi::Source::restartScript(); }
};

class MidiInputPortFactory : public ProcessorCache<MidiInputPort> {
  public:
    static MidiInputPortFactory &instance() {
        static MidiInputPortFactory instance;
        return instance;
    }
    MidiInputPort *newMidiInputPort(const char *name, const char *connectTo);
    MidiInputPort *newMidiInputPort(const char *name, Regex &regex);
    MidiInputPort *newMidiInputPort(const char *name) { return createInputPort(name); }

  private:
    MidiInputPort *createInputPort(const char *name);
};

class MidiOutputPort : public AudioProcessor, public Sink {
    jack_port_t *mport;
    Meter mmeter;

  public:
    MidiOutputPort(const char *name, jack_port_t *jackPort);
    void connect(const char *name) { AudioEngine::instance().connectPort(mport, name); }
    void connect(Regex &regex) {
        AudioEngine::instance().connectPort(mport, regex, JACK_DEFAULT_MIDI_TYPE);
    }
    void disconnect() { AudioEngine::instance().disconnectInputPorts(mport); }
    void disconnect(const char *name) { AudioEngine::instance().disconnectPort(mport, name); }
    void unregister() { AudioEngine::instance().unregisterPort(mport); }
    // Processor interface
    bool doProcess(samples_t time);
    void restart() { midi::Sink::processRestart(); }
    void restartScript() { midi::Sink::scriptRestart(); }
};

class MidiOutputPortFactory : public ProcessorCache<MidiOutputPort> {
  protected:
    void removeObject(MidiOutputPort *);

  public:
    static MidiOutputPortFactory &instance() {
        static MidiOutputPortFactory instance;
        return instance;
    }
    MidiOutputPort *newMidiOutputPort(const char *portName, const char *connectTo);
    MidiOutputPort *newMidiOutputPort(const char *portName) {
        return newMidiOutputPort(portName, 0);
    }
};

} // namespace midi
} // namespace bipscript

#endif // MIDIPORT_H
