/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "midisink.h"

namespace bipscript {
namespace midi {

void Connector::connect(Connection &connection) {
    Source *source = connection.source();
    if (source->connectsTo(mparent)) {
        throw std::logic_error("Cannot connect infinite loop");
    }
    mconnection.store(&connection);
    StatusServer &statusserver = StatusServer::instance();
    if (statusserver.active()) {
        statusserver.connect(&connection, this);
    }
}

MessageEvent *Connector::schedule(Message &mesg, float measure, Clock &clock) {
    RawMessage *raw = new RawMessage(mesg.size());
    mesg.exportBytes(raw->bytes());
    RawEvent *evt = timelines.createEvent(raw, measure, clock);
    reportEvent(clock, evt, mesg);
    // create return event
    MessageEvent *ret = new MessageEvent(mesg, measure);
    ret->ptr(evt);
    return ret;
}

NoteEvent *Connector::schedule(Note &note, float measure, Clock &clock) {
    // note on event
    NoteOn noteOn(note.pitch(), note.velocity(), note.channel());
    RawMessage *mesg = new RawMessage(noteOn.size());
    noteOn.exportBytes(mesg->bytes());
    RawEvent *evt = timelines.createEvent(mesg, measure, clock);
    reportEvent(clock, evt, noteOn);
    // note off event
    NoteOff noteOff(note.pitch(), 0, note.channel());
    RawMessage *offmesg = new RawMessage(noteOff.size());
    noteOff.exportBytes(offmesg->bytes());
    RawEvent *offevt = timelines.createEvent(offmesg, measure + note.duration(), clock);
    reportEvent(clock, offevt, noteOff);
    // create return event
    NoteEvent *ret = new NoteEvent(note, measure);
    ret->ptr(evt, offevt);
    return ret;
}

void Connector::schedule(Pattern &pattern, float measure, Clock &clock) {
    for (unsigned int i = 0; i < pattern.len(); i++) {
        const NoteEvent &pn = *pattern[i];
        float nMeasure = measure + pn.measure() - 1;
        const Note &note = *pn.note();
        // note on event
        NoteOn noteOn(note.pitch(), note.velocity(), note.channel());
        RawMessage *mesg = new RawMessage(noteOn.size());
        noteOn.exportBytes(mesg->bytes());
        RawEvent *evt = timelines.createEvent(mesg, nMeasure, clock);
        reportEvent(clock, evt, noteOn);
        // note off event
        NoteOff noteOff(note.pitch(), 64, note.channel());
        RawMessage *offmesg = new RawMessage(noteOff.size());
        noteOff.exportBytes(offmesg->bytes());
        RawEvent *offevt = timelines.createEvent(offmesg, nMeasure + note.duration(), clock);
        reportEvent(clock, offevt, noteOff);
    }
}

Sequence *Connector::schedule(Sequence &sequence, float measure, Clock &clock) {
    Sequence *ret = new Sequence();
    for (unsigned int i = 0; i < sequence.len(); i++) {
        const MessageEvent *evt = sequence[i];
        if (!evt->message()->meta()) {
            float nMeasure = measure + evt->measure() - 1;
            Message *mesg = evt->message();
            RawMessage *buf = new RawMessage(mesg->size());
            mesg->exportBytes(buf->bytes());
            RawEvent *evt = timelines.createEvent(buf, nMeasure, clock);
            reportEvent(clock, evt, *mesg);
            // create return event
            MessageEvent *mevt = new MessageEvent(*mesg, nMeasure);
            mevt->ptr(evt);
            ret->add(mevt);
        }
    }
    return ret;
}

void Connector::send(Message &mesg) {
    RawMessage *raw = new RawMessage(mesg.size());
    mesg.exportBytes(raw->bytes());
    sendq.add(new RawEvent(raw, Command::Type::TRANSIENT, 1.0));
}

void Connector::send(Message &mesg, float measure, Clock &clock) {
    RawMessage *raw = new RawMessage(mesg.size());
    mesg.exportBytes(raw->bytes());
    RawEvent *evt = timelines.sendEvent(raw, measure, clock);
    reportEvent(clock, evt, mesg);
}

void Connector::send(Note &note, float measure, time::Clock &clock) {
    // note on event
    NoteOn noteOn(note.pitch(), note.velocity(), note.channel());
    RawMessage *mesg = new RawMessage(noteOn.size());
    noteOn.exportBytes(mesg->bytes());
    RawEvent *evt = timelines.sendEvent(mesg, measure, clock);
    reportEvent(clock, evt, noteOn);
    // note off event
    NoteOff noteOff(note.pitch(), 0, note.channel());
    RawMessage *offmesg = new RawMessage(noteOff.size());
    noteOff.exportBytes(offmesg->bytes());
    RawEvent *offevt = timelines.sendEvent(offmesg, measure + note.duration(), clock);
    reportEvent(clock, offevt, noteOff);
}

void Connector::send(Pattern &pattern, float measure, Clock &clock) {
    for (unsigned int i = 0; i < pattern.len(); i++) {
        const NoteEvent &pn = *pattern[i];
        send(*pn.note(), measure + pn.measure() - 1, clock);
    }
}

void Connector::send(Sequence &sequence, float measure, Clock &clock) {
    for (unsigned int i = 0; i < sequence.len(); i++) {
        const MessageEvent *evt = sequence[i];
        if (!evt->message()->meta()) {
            send(*(evt->message()), measure + evt->measure() - 1, clock);
        }
    }
}

void Connector::clear(Clock &clock) {
    timelines.clear(clock);
    StatusServer &statusserver = StatusServer::instance();
    if (statusserver.active()) {
        statusserver.clearMidiEvents(this, &clock);
    }
}

void Connector::remove(MessageEvent &evt) {
    Event<RawMessage> *ptr = evt.ptr();
    if (ptr) {
        timelines.removeEvent(ptr);
        reportRemove(ptr);
    }
}

void Connector::remove(NoteEvent &evt) {
    Event<RawMessage> *onptr = evt.onptr();
    if (onptr) {
        Event<RawMessage> *offptr = evt.offptr();
        timelines.removeEvent(onptr);
        timelines.removeEvent(offptr);
        reportRemove(onptr);
        reportRemove(offptr);
    }
}

void Sink::process(samples_t time) {
    for (uint32_t i = 0; i < mchannels; i++) {
        Connection *connection = bus[i].connection();
        if (connection) {
            connection->source()->process(time);
        }
    }
}

void Sink::processRestart() {
    for (uint32_t i = 0; i < mchannels; i++) {
        connector(i).restart();
    }
}

void Sink::scriptRestart() {
    for (uint32_t i = 0; i < mchannels; i++) {
        connector(i).restartScript();
    }
}

RawEvent *Connector::firstEvent() {
    sendq.update();
    sendq.pop(msendEvt);
    mscheduleEvt = timelines.firstEvent();
    pconnection = this->mconnection.load();
    mconnectEvt = pconnection ? pconnection->firstEvent() : nullptr;
    lastconnect = false;
    pstart = 0;
    return nextEvent();
}

RawEvent *Connector::nextEvent() {
    RawEvent *ret = nullptr;
    if (lastconnect) {
        mconnectEvt = pconnection->nextEvent();
        lastconnect = false;
    }
    if (msendEvt) {
        ret = msendEvt;
        ret->start(0); // pstart
        if (!sendq.pop(msendEvt)) {
            msendEvt = nullptr;
        }
    } else if (mscheduleEvt && (!mconnectEvt || mscheduleEvt->start() < mconnectEvt->start())) {
        ret = mscheduleEvt;
        pstart = ret->start();
        mscheduleEvt = timelines.nextEvent();
    } else if (mconnectEvt) {
        ret = mconnectEvt;
        pstart = ret->start();
        lastconnect = true;
    }
    return ret;
}

void Connector::reportEvent(Clock &clock, RawEvent *evt, Message &mesg) {
    StatusServer &statusserver = StatusServer::instance();
    if (statusserver.active()) {
        // TODO: need all this? may just send raw message
        switch (mesg.command()) {
        case NoteOn::COMMAND: {
            NoteOn *note = static_cast<NoteOn *>(&mesg);
            statusserver.addMidiEvent(this, &clock, evt, evt->measure(), "NOTEON", note->pitch(),
                                      note->velocity());
            break;
        }
        case NoteOff::COMMAND: {
            NoteOff *note = static_cast<NoteOff *>(&mesg);
            statusserver.addMidiEvent(this, &clock, evt, evt->measure(), "NOTEOFF", note->pitch(),
                                      note->velocity());
            break;
        }
        }
    }
}

void bipscript::midi::Connector::reportRemove(RawEvent *ptr) {
    StatusServer &statusserver = StatusServer::instance();
    if (statusserver.active()) {
        statusserver.removeMidiEvent(this, &(ptr->clock()), ptr);
    }
}

} // namespace midi
} // namespace bipscript
