/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef MIDISINK_H
#define MIDISINK_H

#include "midiconnection.h"
#include "midipattern.h"
#include "midisequence.h"
#include "sendqueue.h"
#include "statusserver.h"
#include "timeline.h"

namespace bipscript {
namespace midi {

class Sink;

class Connector {
    std::atomic<Connection *> mconnection;
    AbstractSink *mparent;
    TimelineSet<RawMessage> timelines;
    SendQueue<RawEvent> sendq;
    // period scope
    RawEvent *msendEvt;
    RawEvent *mscheduleEvt;
    RawEvent *mconnectEvt;
    Connection *pconnection;
    samples_t pstart; // start of last event in this period
    bool lastconnect;

  public:
    Connector(AbstractSink *parent)
        : mconnection(nullptr), mparent(parent), sendq(256), msendEvt(nullptr),
          mscheduleEvt(nullptr), mconnectEvt(nullptr) {}
    void connect(Connection &connection);
    Connection *connection() { return mconnection.load(); }
    MessageEvent *schedule(Message &mesg, float measure, Clock &clock = Clock::defaultClock());
    NoteEvent *schedule(Note &note, float measure, Clock &clock = Clock::defaultClock());
    void schedule(Pattern &pattern, float measure, Clock &clock = Clock::defaultClock());
    Sequence *schedule(Sequence &seq, float measure, Clock &clock = Clock::defaultClock());
    void send(Message &mesg);
    void send(Message &mesg, float measure, Clock &clock);
    void send(Note &note, float measure, Clock &clock);
    void send(Pattern &pattern, float measure, Clock &clock);
    void send(Sequence &sequence, float measure, Clock &clock);
    void clear(Clock &clock = Clock::defaultClock());
    void remove(MessageEvent &evt);
    void remove(NoteEvent &evt);
    RawEvent *firstEvent();
    RawEvent *nextEvent();
    bool active() { return timelines.active(); }
    void restart() { timelines.restart(); }
    void restartScript() { timelines.restartScript(); }

  private:
    void reportEvent(Clock &clock, RawEvent *evt, Message &mesg);
    void reportRemove(RawEvent *ptr);
};

class Sink : virtual public AbstractSink {
    BoundArray<Connector> bus;
    uint32_t mchannels;

  protected:
    void process(samples_t time);
    Connector &connector(int i) { return bus[i]; }
    bool connectsTo(AbstractSink *sink) {
        for (uint32_t i = 0; i < mchannels; i++) {
            Connection *conn = bus[i].connection();
            if (conn && conn->source()->connectsTo(sink)) {
                return true;
            }
        }
        return false;
    }
    void reportConnectors(void *node) {
        void *array[mchannels];
        for (uint32_t i = 0; i < mchannels; i++) {
            array[i] = &bus[i];
        }
        StatusServer::instance().midiconnectors(node, array, mchannels);
    }

  public:
    Sink(uint32_t inputs) : bus(inputs), mchannels(inputs) {
        for (uint32_t i = 0; i < mchannels; i++) {
            bus.set(i, new Connector(this), binding::ConnectorObject);
        }
    }
    ~Sink() {
        for (uint32_t i = 0; i < mchannels; i++) {
            delete &bus[i];
        }
    }
    BoundArray<Connector> &midiinbus() { return bus; }
    HSQOBJECT midiin() { return bus.hsqobject(); }
    uint32_t channels() { return mchannels; }
    MessageEvent *schedule(Message &mesg, float measure, Clock &clock = Clock::defaultClock()) {
        checkMidiInputs();
        return bus[0].schedule(mesg, measure, clock);
    }
    NoteEvent *schedule(Note &note, float measure, Clock &clock = Clock::defaultClock()) {
        checkMidiInputs();
        return bus[0].schedule(note, measure, clock);
    }
    void schedule(Pattern &pattern, float measure, Clock &clock = Clock::defaultClock()) {
        checkMidiInputs();
        bus[0].schedule(pattern, measure, clock);
    }
    Sequence *schedule(Sequence &sequence, float measure, Clock &clock = Clock::defaultClock()) {
        checkMidiInputs();
        return bus[0].schedule(sequence, measure, clock);
    }
    void send(Message &mesg) {
        checkMidiInputs();
        bus[0].send(mesg);
    }
    void send(Message &mesg, float measure, Clock &clock = Clock::defaultClock()) {
        checkMidiInputs();
        bus[0].send(mesg, measure, clock);
    }
    void send(Note &note, float measure, Clock &clock = Clock::defaultClock()) {
        checkMidiInputs();
        bus[0].send(note, measure, clock);
    }
    void send(Pattern &pattern, float measure, Clock &clock = Clock::defaultClock()) {
        checkMidiInputs();
        bus[0].send(pattern, measure, clock);
    }
    void send(Sequence &seq, float measure, Clock &clock = Clock::defaultClock()) {
        checkMidiInputs();
        bus[0].send(seq, measure, clock);
    }
    void clear(Clock &clock = Clock::defaultClock()) { bus[0].clear(clock); }
    void remove(MessageEvent &evt) { bus[0].remove(evt); }
    void remove(NoteEvent &evt) { bus[0].remove(evt); }
    void processRestart(); // process thread
    void scriptRestart();  // script thread
  private:
    void checkMidiInputs() {
        if (mchannels == 0) {
            throw std::logic_error("Cannot schedule/send MIDI, object has no MIDI inputs");
        }
    }
};

} // namespace midi
} // namespace bipscript

#endif // MIDISINK_H
