/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "tab.h"
#include "midipattern.h"

#include <cmath>
#include <cstring>
#include <regex>

namespace bipscript {
namespace midi {

int TabBar::division = 8;

void TabBar::add(uint8_t vel, float measure) {
    int slot = quantize(measure);
    if (vel > notes[slot]) {
        notes[slot] = vel;
    }
    mopen = slot + 1;
}

void TabBar::end(float measure) {
    if (mopen == -1) {
        throw std::logic_error("tab write error: NOTEOFF without NOTEON");
    }
    int slot = quantize(measure);
    for (int i = mopen; i < slot; i++) {
        notes[i] = -1;
    }
    mopen = -1;
}

bool TabBar::finalize() {
    if (mopen > -1) {
        for (int i = mopen; i < division; i++) {
            notes[i] = -1;
        }
        return false; // still open
    }
    return true; // all notes closed
}

std::string TabBar::print(velocitymap &vmap) {
    std::string ret;
    for (int i = 0; i < division; i++) {
        if (notes[i] > 0) {
            ret += closest(notes[i], vmap);
        } else {
            ret.append(notes[i] < 0 ? "~" : "-");
        }
    }
    return ret;
}

int TabBar::quantize(float measure) {
    float value = measure - mstart + 0.000001; // small bump for rounding
    float delta = 1.0 / division;
    for (int i = 1; i <= division; i++) {
        if (value < delta * i) {
            return i - 1;
        }
    }
    throw std::logic_error("tab error: could not quantize");
}

char TabBar::closest(uint8_t value, velocitymap &vmap) {
    char ret = 'x';
    int min = 128;
    for (auto entry : vmap) {
        uint8_t delta = abs(value - entry.second);
        if (delta < min) {
            ret = entry.first;
            min = delta;
        }
    }
    return ret;
}

void bipscript::midi::TabLine::forward(float measure) {
    while (measure >= current->start() + 1.0) {
        bool final = current->finalize();
        current = new TabBar(current->start() + 1.0, !final);
        mbars.push_back(current);
    }
}

TabLine::~TabLine() {
    for (TabBar *bar : mbars) {
        delete bar;
    }
}

void TabLine::add(uint8_t vel, float measure) {
    forward(measure);
    current->add(vel, measure);
}

void TabLine::end(float measure) {
    forward(measure);
    current->end(measure);
}

std::string TabLine::print(uint index, velocitymap &vmap) {
    forward(index + 1);
    return mbars[index]->print(vmap);
}

bool Tab::linecmp(TabLine *l1, TabLine *l2) { return l1->note() > l2->note(); }

NoteOn *Tab::readbar(std::string prefix, int pitch, NoteOn *lastnote, float start, std::string &bar,
                     Sequence &seq) {
    int len = bar.length();
    float division = 1.0 / len;
    NoteOn *note = lastnote;
    for (int i = 0; i < len; i++) {
        float measure = start + i * division;
        if (bar[i] != '~') {
            if (note) {
                NoteOff *off = new NoteOff(note->pitch(), 64);
                seq.add(*off, measure);
                note = nullptr;
            }
            if (bar[i] != '-') {
                if (!pitch) {
                    // lookup and complain if not found
                    Strikedef *strike = prefixmap[prefix][bar[i]];
                    if (!strike) {
                        std::string mesg("tab error: prefix ");
                        mesg.append(prefix);
                        mesg.append(" has no definition for strike ");
                        mesg.append(bar.substr(i, 1));
                        throw std::logic_error(mesg);
                    }
                    note = new NoteOn(strike->pitch, strike->velocity);
                } else {
                    // lookup velocity
                    uint8_t velocity = 64;
                    if (vmap.find(bar[i]) != vmap.end()) {
                        velocity = vmap[bar[i]];
                    }
                    note = new NoteOn(pitch, velocity);
                }
                seq.add(*note, measure);
            }
        }
    }
    return note;
}

Sequence *Tab::sequence(const char *tab) {
    Sequence *ret = new Sequence();
    std::map<std::string, float> measure;
    std::map<std::string, NoteOn *> lastnote;
    int linenum = 1;
    int i = 0;

    while (tab[i]) {

        // skip any whitespace
        while (tab[i] && std::isspace(tab[i])) {
            if (tab[i++] == '\n') {
                linenum++;
            }
        }

        // parse prefix
        int prestart = i;
        while (tab[i] && tab[i] != '|' && tab[i] != '\n') {
            i++;
        }
        if ((i > prestart && !tab[i]) || tab[i] == '\n' || i - prestart > 4) {
            std::string mesg("tab parse error: no prefix on line ");
            mesg.append(std::to_string(linenum));
            throw std::logic_error(mesg);
        }

        // line has a prefix
        if (i > prestart) {

            // parse prefix to pitch if possible
            std::string prefix(&tab[prestart], i - prestart);
            std::regex numberex("[0-9]+");
            int pitch = 0;
            if (regex_match(prefix, numberex)) {
                pitch = stoi(prefix);
            }

            // get measures start
            float start = measure[prefix];
            if (start < 1.0) {
                measure[prefix] = 1.0;
                start = 1.0;
            }

            // grab line
            int barstart = ++i;
            while (tab[i] && tab[i] != '\n') {
                if (tab[i] == '|') {
                    std::string bar(&tab[barstart], i - barstart);
                    NoteOn *last = lastnote[prefix];
                    lastnote[prefix] = readbar(prefix, pitch, last, start, bar, *ret);
                    barstart = i + 1;
                    start = ++measure[prefix];
                }
                i++;
            }

            // reject markup after last bar
            std::string afterbar(&tab[barstart], i - barstart);
            for (size_t index = 0; index < afterbar.length(); index++) {
                if (!isspace(afterbar[index])) {
                    std::string mesg("tab parse error: markup after last bar on line ");
                    mesg.append(std::to_string(linenum));
                    throw std::logic_error(mesg);
                }
            }
        }
    }

    // close any remaining open notes
    for (auto const &entry : lastnote) {
        NoteOn *note = entry.second;
        if (note) {
            const int pitch = note->pitch();
            NoteOff *off = new NoteOff(pitch, 64);
            ret->add(*off, measure[entry.first]);
        }
    }
    return ret;
}

void Tab::define(std::string prefix, std::string strike, uint8_t pitch, uint8_t velocity) {
    if (strike.length() != 1) {
        throw std::logic_error("tab define method: strike should be a single character string");
    }
    prefixmap[prefix][strike[0]] = new Strikedef(pitch, velocity);
}

void Tab::velocity(std::string strike, uint8_t velocity) {
    if (strike.length() != 1) {
        throw std::logic_error("tab velocity method: strike should be a single character string");
    }
    vmap[strike[0]] = velocity;
}

void Tab::linewrap(int value) {
    if (value <= 0) {
        throw std::logic_error("tab object linewrap property must be greater than zero");
    }
    mlinewrap = value;
}

void Tab::division(int value) {
    if (value <= 0) {
        throw std::logic_error("tab object division property must be greater than zero");
    }
    TabBar::division = value;
}

Tab::~Tab() {
    for (auto entry : prefixmap) {
        for (auto strike : entry.second) {
            delete strike.second;
        }
    }
}

std::string Tab::string(Pattern &pattern) {
    Sequence *seq = pattern.sequence();
    std::string ret = string(*seq);
    delete seq;
    return ret;
}

std::string Tab::string(Sequence &seq) {

    std::map<int, TabLine *> linemap;
    std::vector<TabLine *> linelist;

    uint8_t buf[3]; // for exporting midi events

    for (size_t i = 0; i < seq.len(); i++) {
        float measure = seq[i]->measure();
        int com = seq[i]->message()->command();
        if (com == NoteOn::COMMAND || com == NoteOff::COMMAND) {
            seq[i]->message()->exportBytes(buf);
            uint8_t note = buf[1];
            TabLine *line = linemap[note];
            if (!line) {
                line = new TabLine(note);
                linelist.push_back(line);
                linemap[note] = line;
            }
            if (com == NoteOn::COMMAND) {
                line->add(buf[2], measure);
            } else {
                line->end(measure);
            }
        }
    }

    // finalize lines
    uint seqsize = ceil(seq.length());
    for (TabLine *line : linelist) {
        line->forward(seqsize);
    }

    // sort lines by pitch
    std::sort(linelist.begin(), linelist.end(), linecmp);

    if (!linelist.size()) {
        return ""; // no lines = empty tab
    }
    // grab bars from first line
    TabLine *firstline = linelist[0];
    auto topbars = firstline->bars();

    // find max prefix size
    uint prefixwidth = 1;
    for (TabLine *line : linelist) {
        uint width = std::to_string(line->note()).size();
        prefixwidth = std::max(width, prefixwidth);
    }

    // outer loop over line-wrapped chunks
    std::string output;
    uint bar = 0;
    while (bar < seqsize) {

        // find chunksize from top line
        uint endbar = bar;
        uint chunksize = prefixwidth + 1;
        for (; endbar < topbars.size(); endbar++) {
            uint barsize = topbars[endbar]->size() + 1;
            if (chunksize + barsize > mlinewrap) {
                break;
            }
            chunksize += barsize;
        }
        if (endbar < bar + 1) { // must be at least one bar
            endbar = bar + 1;
        }
        // printf("chunk %d - %d of %d\n", bar, endbar, seqsize);

        // write lines
        for (TabLine *line : linelist) {
            // pad and write prefix
            std::string prefix = std::to_string(line->note());
            prefix.insert(0, prefixwidth - prefix.size(), ' ');
            output.append(prefix);
            output.append("|");
            // write bars
            for (uint i = bar; i < endbar; i++) {
                output.append(line->print(i, vmap));
                output.append("|");
            }
            output.append("\n");
        }
        output.append("\n");
        bar = endbar;
    }

    // clean up
    for (auto entry : linemap) {
        delete entry.second;
    }

    return output;
}

} // namespace midi
} // namespace bipscript
