/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef OBJECTCACHE_H
#define OBJECTCACHE_H

#include <iostream>
#include <map>
#include <set>

namespace bipscript {

class ObjectCache {
  public:
    /**
     * Called after the main script has completed so factories can reset for the next run.
     *
     * Runs in the script thread.
     */
    virtual void scriptComplete() = 0; // TODO: do we really need this
    /**
     * Called after a script run is complete so objects can clean up
     *
     * If restart == false then the entire process is about to exit
     *
     * Runs in the script thread.
     */
    virtual void reset(bool restart) = 0;
    virtual ~ObjectCache() {}
};

/**
 * Cache template for resuable objects
 *
 * T must be a Listable
 */
template <class T> class ActiveCache : public ObjectCache {
    std::map<int, T *> instanceMap;
    std::set<T *> activeScriptObjects;
    virtual void newObject(T *) = 0;

  protected:
    virtual void restartCache() {}
    virtual void removeObject(T *) = 0;
    /**
     * find an existing object with this key
     */
    T *findObject(int key) {
        T *ret = instanceMap[key];
        if (ret) {
            activeScriptObjects.insert(ret);
        }
        return ret;
    }
    /**
     * add a new object with key
     */
    void registerObject(int key, T *obj) {
        instanceMap[key] = obj;
        activeScriptObjects.insert(obj);
        newObject(obj);
    }
    /**
     * remove all existing objects
     */
    void removeAll() {
        for (auto iterator = instanceMap.begin(); iterator != instanceMap.end();) {
            T *obj = iterator->second;
            auto it = activeScriptObjects.find(obj);
            if (it != activeScriptObjects.end()) {
                activeScriptObjects.erase(it);
            }
            iterator = instanceMap.erase(iterator);
            removeObject(obj);
        }
    }
    /**
     * delete all existing objects
     */
    void deleteInstances() {
        for (auto iterator = instanceMap.begin(); iterator != instanceMap.end(); iterator++) {
            T *obj = iterator->second;
            delete obj;
        }
        instanceMap.clear();
    }
    ~ActiveCache() { deleteInstances(); }

  public:
    /**
     * Cleans cache after script has completed.
     *
     * Runs in script thread.
     *
     * Loops over all objects in the instance map and removes those not used in the last run.
     */
    void scriptComplete() override {
        // loop over cached ports
        for (auto iterator = instanceMap.begin(); iterator != instanceMap.end();) {
            T *obj = iterator->second;
            // object was used in the last script run?
            auto it = activeScriptObjects.find(obj);
            if (it != activeScriptObjects.end()) {
                // std::cout << typeid(T).name() << " was used in the last run: " << iterator->first
                // << std::endl;
                activeScriptObjects.erase(it);
                iterator++;
            } else {
                // remove object from map and audio engine
                // std::cout << typeid(T).name() << " NOT used in the last run: " << iterator->first
                // << std::endl;
                iterator = instanceMap.erase(iterator);
                removeObject(obj);
            }
        }
    }
    /**
     *
     * Runs in script thread.
     *
     * Loops over all objects in the instance map and calls their reset method
     */
    void reset(bool) override {
        for (auto iterator = instanceMap.begin(); iterator != instanceMap.end(); iterator++) {
            iterator->second->restartScript();
        }
        restartCache();
    }
};

} // namespace bipscript

#endif // OBJECTCACHE_H
