/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "oscinput.h"
#include "methodqueue.h"
#include "oscmessage.h"

namespace bipscript {
namespace osc {

void osc_error(int num, const char *msg, const char *path) {
    std::string mesg("Osc.Input error ");
    mesg.append(std::to_string(num));
    if (path) {
        mesg.append(" (path=");
        mesg.append(path);
        mesg.append(")");
    }
    mesg.append(": ");
    mesg.append(msg);
    throw std::logic_error(mesg);
}

int osc_handler(const char *path, const char *types, lo_arg **argv, int argc, lo_message data,
                void *user_data) {
    return ((Input *)user_data)->handle(path, types, argv, argc, data);
}

Input::Input() : onReceiveHandler(nullptr) {
    st = lo_server_thread_new(NULL, osc_error);
    lo_server_thread_add_method(st, NULL, NULL, osc_handler, this);
    lo_server_thread_start(st);
}

Input::Input(int port, int protocol) : onReceiveHandler(nullptr) {
    std::string portstr(std::to_string(port));
    int loproto;
    switch (protocol) {
    case net::UDP:
        loproto = LO_UDP;
        break;
    case net::TCP:
        loproto = LO_TCP;
        break;
    default:
        throw std::logic_error("Osc.Input unknown protocol");
    }
    st = lo_server_thread_new_with_proto(portstr.c_str(), loproto, osc_error);
    lo_server_thread_add_method(st, NULL, NULL, osc_handler, this);
    lo_server_thread_start(st);
}

Input::Input(int port, const char *group) : onReceiveHandler(nullptr) {
    std::string portstr(std::to_string(port));
    st = lo_server_thread_new_multicast(group, portstr.c_str(), osc_error);
    lo_server_thread_add_method(st, NULL, NULL, osc_handler, this);
    lo_server_thread_start(st);
}

Input::Input(const char *path) : onReceiveHandler(nullptr) {
    st = lo_server_thread_new_with_proto(path, LO_UNIX, osc_error);
    lo_server_thread_add_method(st, NULL, NULL, osc_handler, this);
    lo_server_thread_start(st);
}

int Input::handle(const char *path, const char *types, lo_arg **argv, int argc, void *) {
    ScriptFunction *handler = onReceiveHandler.load();
    if (handler) {
        Message *message = new Message(path);
        for (int i = 0; i < argc; i++) {
            switch (types[i]) {
            case 'i':
                message->addInteger(*(int32_t *)argv[i]);
                break;
            case 'f':
                message->addFloat(*(float *)argv[i]);
                break;
            case 's':
                message->addString((const char *)argv[i]);
                break;
            case 'T':
                message->addBoolean(true);
                break;
            case 'F':
                message->addBoolean(false);
                break;
            case 'N':
                message->addNull(0);
                break;
            default:
                break;
            }
        }
        MethodQueue::instance().dispatch(new OnReceiveClosure(*handler, message));
    }
    return 0;
}

void Input::onMessage(ScriptFunction &handler) {
    if (handler.numargs() != 2) {
        throw std::logic_error("onReceive handler should take one argument");
    }
    onReceiveHandler.store(new ScriptFunction(handler));
}

void Input::join() {
    lo_server_thread_stop(st);
    lo_server_thread_free(st);
}

void Input::restartScript() {
    ScriptFunction *handler = onReceiveHandler.exchange(nullptr);
    if (handler) {
        delete handler;
    }
}

/**
 * script thread
 */
Input *InputFactory::newOscInput() {
    std::string hash(std::to_string(minstcount++));
    int key = std::hash<std::string>()(hash);
    Input *obj = findObject(key);
    if (!obj) {
        obj = new Input();
        registerObject(key, obj);
    }
    return obj;
}

Input *InputFactory::newOscInput(int port, int protocol) {
    std::string hash(std::to_string(port));
    hash.append("::");
    hash.append(std::to_string(protocol));
    int key = std::hash<std::string>()(hash);
    Input *obj = findObject(key);
    if (!obj) {
        obj = new Input(port, protocol);
        registerObject(key, obj);
    }
    return obj;
}

Input *InputFactory::newOscInput(int port, const char *group) {
    std::string hash(std::to_string(port));
    hash.append("::");
    hash.append(group);
    int key = std::hash<std::string>()(hash);
    Input *obj = findObject(key);
    if (!obj) {
        obj = new Input(port, group);
        registerObject(key, obj);
    }
    return obj;
}

Input *InputFactory::newOscInput(const char *path) {
    int key = std::hash<std::string>()(path);
    Input *obj = findObject(key);
    if (!obj) {
        obj = new Input(path);
        registerObject(key, obj);
    }
    return obj;
}

} // namespace osc
} // namespace bipscript
