/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef OSCMESSAGE_H
#define OSCMESSAGE_H

#include "event.h"
#include "squirrel.h"

#include <cstring>
#include <string>
#include <vector>

namespace bipscript {
namespace osc {

struct Parameter {
    char type;
    union {
        int intValue;
        float floatValue;
        const char *stringValue;
    } value;
    Parameter(char type) : type(type) {}
};

class Message : public Listable<Message> {
    std::string mpath;
    std::vector<Parameter> parameters;

  public:
    Message(std::string path) : mpath(path) {}
    Message(Message &copy) : mpath(copy.mpath), parameters(copy.parameters) {
        for (auto &param : parameters) {
            if (param.type == 's') {
                size_t len = std::strlen(param.value.stringValue) + 1;
                char *const copy = new char[len];
                std::strncpy(copy, param.value.stringValue, len);
                param.value.stringValue = copy;
            }
        }
    }
    ~Message() {
        for (auto &param : parameters) {
            if (param.type == 's') {
                delete[] param.value.stringValue;
            }
        }
    }
    std::string path() { return mpath; }
    void addInteger(int value) {
        Parameter param('i');
        param.value.intValue = value;
        parameters.push_back(param);
    }
    void addFloat(float value) {
        Parameter param('f');
        param.value.floatValue = value;
        parameters.push_back(param);
    }
    void addString(const char *value) {
        Parameter param('s');
        size_t len = std::strlen(value) + 1;
        char *const copy = new char[len];
        std::strncpy(copy, value, len);
        param.value.stringValue = copy;
        parameters.push_back(param);
    }
    void addBoolean(bool value) {
        Parameter param(value ? 'T' : 'F');
        parameters.push_back(param);
    }
    void addNull(const char *) {
        Parameter param('N');
        parameters.push_back(param);
    }
    int parameterCount() { return parameters.size(); }
    Parameter parameter(int i) { return parameters[i]; }
    void arg(HSQUIRRELVM vm, unsigned long i) {
        if (i >= parameters.size()) {
            throw std::logic_error("OSC message argument out of bounds");
        }
        switch (parameters[i].type) {
        case 'i':
            sq_pushinteger(vm, parameters[i].value.intValue);
            break;
        case 'f':
            sq_pushfloat(vm, parameters[i].value.floatValue);
            break;
        case 's':
            sq_pushstring(vm, parameters[i].value.stringValue, -1);
            break;
        case 'T':
            sq_pushbool(vm, true);
            break;
        case 'F':
            sq_pushbool(vm, false);
            break;
        case 'N':
            sq_pushnull(vm);
            break;
        }
    }
};

} // namespace osc
} // namespace bipscript

#endif // OSCMESSAGE_H
