/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "oscoutput.h"
#include "audioengine.h"

namespace bipscript {
namespace osc {

Output::Output(const char *host, int port) : ProcessorThread(true, 5), sendqueue(128) {
    std::string portString = std::to_string(port);
    loAddress = lo_address_new(host, portString.c_str());
    if (!loAddress) {
        throw std::logic_error("could not create OSC address");
    }
    start(); // start thread
}

void Output::remove(MessageEvent &event) { timelineset.removeEvent(&event); }

MessageEvent *Output::schedule(Message &message, float measure, time::Clock &clock) {
    MessageEvent *event =
        new MessageEvent(new Message(message), Command::Type::SCHEDULE, measure, clock);
    timelineset.addRemote(event->addref(), clock);
    return event;
}

bool Output::process() {
    // immediate send method events
    Message *sendevent = nullptr;
    while (sendqueue.pop(sendevent)) {
        sendMessage(*sendevent);
    }
    sendqueue.update();
    //  scheduled events
    Event<Message> *event = timelineset.firstEvent();
    while (event) {
        sendMessage(event->message());
        event = timelineset.nextEvent();
    }
    return timelineset.active();
}

void bipscript::osc::Output::sendMessage(Message &message) {
    lo_message mesg = lo_message_new();
    for (int i = 0; i < message.parameterCount(); i++) {
        Parameter param = message.parameter(i);
        if (param.type == 'i') {
            lo_message_add_int32(mesg, param.value.intValue);
        } else if (param.type == 'f') {
            lo_message_add_float(mesg, param.value.floatValue);
        } else if (param.type == 's') {
            lo_message_add_string(mesg, param.value.stringValue);
        } else if (param.type == 'T') {
            lo_message_add_true(mesg);
        } else if (param.type == 'F') {
            lo_message_add_false(mesg);
        } else if (param.type == 'N') {
            lo_message_add_nil(mesg);
        }
    }
    lo_send_message(loAddress, message.path().c_str(), mesg);
    lo_message_free(mesg);
}

/**
 * script thread
 */
Output *OutputFactory::newOscOutput(const char *host, int port) {
    // TODO: needs better hash
    int key = std::hash<std::string>()(host) + port;
    Output *obj = findObject(key);
    if (!obj) {
        obj = new Output(host, port);
        registerObject(key, obj);
    }
    return obj;
}

} // namespace osc
} // namespace bipscript
