/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef READER_H
#define READER_H

#include "iterator.h"
#include "refcountable.h"

#include <algorithm>
#include <stddef.h>
#include <stdexcept>

namespace bipscript {

namespace binding {
extern HSQOBJECT readerObject;
extern HSQOBJECT bufferObject;
} // namespace binding

class Buffer;

class Reader : public RefCountable<Reader> {

  public:
    static void bindType(HSQUIRRELVM vm) {
        BindUtil binding(vm);
        binding.startClass("reader", 5, binding::readerObject);
        binding.addMethod("_cloned", 7, NoClone);
        binding.addMethod("_get", 4, Get);
        binding.addMethod("_nexti", 6, Nexti);
        binding.addMethod("_tostring", 9, ToString);
        binding.addMethod("lines", 5, Lines);
        binding.addMethod("tell", 4, Tell);
        binding.addMethod("boolean", 7, ReadBool);
        binding.addMethod("byte", 4, ReadByte);
        binding.addMethod("float", 5, ReadFloat);
        binding.addMethod("integer", 7, ReadInt);
        binding.addMethod("string", 6, ReadString);
        binding.addMethod("buffer", 6, ReadBuffer);
        binding.endClass();
    }
    //    static Reader *getReader(HSQUIRRELVM vm, SQInteger index) {
    //        return BindUtil::getObject<Reader>(vm, index, binding::readerObject);
    //    }
    virtual ~Reader() {}
    virtual bool eos() = 0;
    virtual void line(std::string &str) = 0;

  protected:
    virtual bool readBool(bool &value) = 0;
    virtual bool readByte(unsigned char &value) = 0;
    virtual bool readFloat(float &value) = 0;
    virtual bool readInt(long &value, size_t size = sizeof(long)) = 0;
    virtual bool readString(std::string &value) = 0;
    virtual bool readBuffer(Buffer **, size_t) = 0;
    virtual size_t tell() = 0;
    Iterator *lines();

  private:
    // binding methods
#define BIND_CLASS_NAME "reader"
    static SQInteger NoClone(HSQUIRRELVM vm) {
        return sq_throwerror(vm, "cannot clone objects of type " BIND_CLASS_NAME);
    }
#define BIND_METHOD_NAME "_get"
    static SQInteger Get(HSQUIRRELVM vm) {
        SQObjectType argtype = sq_gettype(vm, 2);
        if (argtype == OT_INTEGER) {
            SQInteger index;
            sq_getinteger(vm, 2, &index);
            BIND_GET_THIS_PROPERTY(Reader)
            if (index == 0) {
                unsigned char buf[7];

                // read single utf-8 char into buffer
                obj->readByte(buf[0]);
                uint size = 1;
                // is multibyte?
                if (buf[0] & 0x80) {
                    uint head = buf[0];
                    head >>= 4;
                    head &= 7;
                    size = std::min((head == 4) ? 2 : head - 3, (uint)6);
                    for (uint i = 1; i < size; i++) {
                        obj->readByte(buf[i]);
                    }
                }
                buf[size] = '\0';

                // push codepoint as string
                sq_pushstring(vm, (char *)buf, size);
                return 1;
            }
        }
        sq_pushnull(vm);
        return sq_throwobject(vm);
    }
#undef BIND_METHOD_NAME
#define BIND_METHOD_NAME "(iterator)"
    static SQInteger Nexti(HSQUIRRELVM vm) {
        BIND_GET_THIS_PROPERTY(Reader)
        if (!obj->eos()) {
            sq_pushinteger(vm, 0);
            return 1;
        }
        sq_pushnull(vm);
        return 1;
    }
#undef BIND_METHOD_NAME
#define BIND_METHOD_NAME "_tostring"
    static SQInteger ToString(HSQUIRRELVM vm) {
        sq_pushstring(vm, "(reader)", -1);
        return 1;
    }
#undef BIND_METHOD_NAME
#define BIND_METHOD_NAME "lines"
    static SQInteger Lines(HSQUIRRELVM vm) {
        BIND_GET_NARGS()
        BIND_GET_THIS_INSTANCE(Reader)
        BIND_NARGS_CHECK_EXACT(0) // check exact parameter count
        // return value
        bipscript::Iterator *ret;
        // call the implementation
        BIND_WRAP_TRY(ret = obj->lines())
        // push return value
        sq_pushobject(vm, binding::iteratorObject);
        sq_createinstance(vm, -1);
        sq_remove(vm, -2);
        sq_setinstanceup(vm, -1, ret);
        sq_setreleasehook(vm, -1, BindUtil::Delete<bipscript::Iterator>);
        return 1;
    }
#undef BIND_METHOD_NAME
#define BIND_METHOD_NAME "tell"
    static SQInteger Tell(HSQUIRRELVM vm) {
        BIND_GET_NARGS()
        BIND_NARGS_CHECK_EXACT(0)
        BIND_GET_THIS(Reader, binding::readerObject)
        sq_pushinteger(vm, obj->tell());
        return 1;
    }
#undef BIND_METHOD_NAME
#define BIND_METHOD_NAME "bool"
    static SQInteger ReadBool(HSQUIRRELVM vm) {
        BIND_GET_NARGS()
        BIND_NARGS_CHECK_EXACT(0)
        BIND_GET_THIS(Reader, binding::readerObject)
        bool value;
        bool result;
        BIND_WRAP_TRY(result = obj->readBool(value));
        if (result) {
            sq_pushbool(vm, value);
        } else {
            sq_pushnull(vm);
        }
        return 1;
    }
#undef BIND_METHOD_NAME
#define BIND_METHOD_NAME "byte"
    static SQInteger ReadByte(HSQUIRRELVM vm) {
        BIND_GET_NARGS()
        BIND_NARGS_CHECK_EXACT(0)
        BIND_GET_THIS(Reader, binding::readerObject)
        unsigned char value;
        bool success;
        BIND_WRAP_TRY(success = obj->readByte(value));
        if (success) {
            sq_pushinteger(vm, value);
        } else {
            sq_pushnull(vm);
        }
        return 1;
    }
#undef BIND_METHOD_NAME
#define BIND_METHOD_NAME "float"
    static SQInteger ReadFloat(HSQUIRRELVM vm) {
        BIND_GET_NARGS()
        BIND_NARGS_CHECK_EXACT(0)
        BIND_GET_THIS(Reader, binding::readerObject)
        float value;
        bool success;
        BIND_WRAP_TRY(success = obj->readFloat(value));
        if (success) {
            sq_pushfloat(vm, value);
        } else {
            sq_pushnull(vm);
        }
        return 1;
    }
#undef BIND_METHOD_NAME
#define BIND_METHOD_NAME "int"
    static SQInteger ReadInt(HSQUIRRELVM vm) {
        BIND_GET_NARGS()
        BIND_NARGS_CHECK_MAX(1)
        BIND_GET_THIS(Reader, binding::readerObject)
        long value;
        bool success;
        if (numargs == 2) {
            BIND_GET_CHECK_INTEGER(size, 1)
            BIND_WRAP_TRY(success = obj->readInt(value, size));
        } else {
            BIND_WRAP_TRY(success = obj->readInt(value));
        }
        if (success) {
            sq_pushinteger(vm, value);
        } else {
            sq_pushnull(vm);
        }
        return 1;
    }
#undef BIND_METHOD_NAME
#define BIND_METHOD_NAME "string"
    static SQInteger ReadString(HSQUIRRELVM vm) {
        BIND_GET_NARGS()
        BIND_NARGS_CHECK_EXACT(0)
        BIND_GET_THIS(Reader, binding::readerObject);
        std::string ret;
        bool success;
        BIND_WRAP_TRY(success = obj->readString(ret));
        if (success) {
            sq_pushstring(vm, ret.c_str(), ret.length());
        } else {
            sq_pushnull(vm);
        }
        return 1;
    }
#undef BIND_METHOD_NAME
#define BIND_METHOD_NAME "buffer"
    static SQInteger ReadBuffer(HSQUIRRELVM vm) {
        BIND_GET_NARGS()
        BIND_NARGS_CHECK_EXACT(1)
        BIND_GET_THIS(Reader, binding::readerObject);
        BIND_GET_CHECK_INTEGER(size, 1)
        Buffer *ret;
        bool success;
        BIND_WRAP_TRY(success = obj->readBuffer(&ret, size));
        if (success) {
            sq_pushobject(vm, binding::bufferObject);
            sq_createinstance(vm, -1);
            sq_remove(vm, -2);
            sq_setinstanceup(vm, -1, ret);
            sq_setreleasehook(vm, -1, BindUtil::Decrement<Buffer>);
        } else {
            sq_pushnull(vm);
        }
        return 1;
    }
#undef BIND_METHOD_NAME
#undef BIND_CLASS_NAME
};

class LineIterator : public Iterator {
    Reader *mreader;

  public:
    LineIterator(Reader *reader) : mreader(reader->addref()) {}
    ~LineIterator() {
        if (!mreader->delref()) {
            delete mreader;
        }
    }
    bool hasNext() override { return !mreader->eos(); }
    void next(HSQUIRRELVM vm) override {
        std::string str;
        mreader->line(str);
        sq_pushstring(vm, str.c_str(), str.length());
    }
};

inline Iterator *Reader::lines() { return new LineIterator(this); }

} // namespace bipscript

#endif // READER_H
