/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef SCRIPTTYPES_H
#define SCRIPTTYPES_H

#include "squirrel.h"
#include <stdint.h>

#include "collectorthread.h"

namespace bipscript {

enum ScriptValueType { NULLVALUE, BOOL, INTEGER, FLOAT, STRING, ARRAY };

struct ScriptValue {
    ScriptValueType type;
    union {
        SQInteger intValue;
        SQFloat floatValue;
        const SQChar *stringValue;
        HSQOBJECT arrayObj;
    };
    ScriptValue() {}
    ScriptValue(HSQUIRRELVM &vm, SQInteger idx);
    void pushValue(HSQUIRRELVM &vm);
};

class ScriptArrayIterator {
    HSQUIRRELVM vm;
    HSQOBJECT arrayobj;
    uint32_t mlength;
    uint32_t index;
    ScriptValue element;

  public:
    ScriptArrayIterator(HSQOBJECT &array);
    ~ScriptArrayIterator() { sq_poptop(vm); } // pop array object
    uint32_t length() const { return mlength; }
    bool hasNext() { return index < mlength; }
    ScriptValue &next();
};

struct ScriptHashPair {
    ScriptValue key;
    ScriptValue value;
};

class ScriptHashIterator {
    HSQUIRRELVM &vm;
    ScriptHashPair pair;

  public:
    ScriptHashIterator(HSQUIRRELVM &vm, HSQOBJECT &hash);
    bool hasNext();
    ScriptHashPair &next();
};

/**
 * wrapper around the script function type
 */
class ScriptFunction {
    HSQOBJECT mfunction;
    SQInteger mnumargs;
    const SQChar *mname;

  protected:
    HSQUIRRELVM vm;

  public:
    ScriptFunction(HSQUIRRELVM &vm, HSQOBJECT &func, int pos) : mfunction(func), vm(vm) {
        SQInteger nfreevars;
        sq_getclosureinfo(vm, pos, &mnumargs, &nfreevars);
        sq_getclosurename(vm, 2);
        sq_getstring(vm, -1, &mname);
        sq_poptop(vm);
        sq_addref(vm, &func);
    }
    HSQOBJECT &function() { return mfunction; }
    uint32_t numargs() { return mnumargs; }
    const char *name() { return mname; }
    void release() { sq_release(vm, &mfunction); }
};

class ScriptFunctionCollector {
    Collector<ScriptFunction> functionCollector;
    ScriptFunctionCollector() { CollectorThread::instance().addCollector(&functionCollector); }

  public:
    static ScriptFunctionCollector &instance() {
        static ScriptFunctionCollector instance;
        return instance;
    }
    void recycle(ScriptFunction *handler) { functionCollector.recycle(handler); }
};

} // namespace bipscript

#endif // SCRIPTTYPES_H
