/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "statusserver.h"

#include <chrono>
#include <map>
#include <thread>

namespace bipscript {

StatusServer::~StatusServer() {
    if (mactive) {
        lo_address_free(loAddress);
    }
}

void *run_server(void *arg) {
    ((StatusServer *)arg)->run();
    return 0;
}

void StatusServer::activate() {
    loAddress = lo_address_new("localhost", mport.c_str());
    int result = pthread_create(&thread, 0, run_server, this);
    mactive = !result;
}

void StatusServer::newClock(void *node, const char *name, float tempo) {
    lo_message mesg = lo_message_new();
    lo_message_add_int64(mesg, hash(node));
    lo_message_add_string(mesg, name);
    lo_message_add_float(mesg, tempo);
    lo_send_message(loAddress, "/clock/new", mesg);
    lo_message_free(mesg);
}

void StatusServer::newClock(void *node, const char *name, float tempo, int signum, int sigden) {
    lo_message mesg = lo_message_new();
    lo_message_add_int64(mesg, hash(node));
    lo_message_add_string(mesg, name);
    lo_message_add_float(mesg, tempo);
    lo_message_add_int32(mesg, signum);
    lo_message_add_int32(mesg, sigden);
    lo_send_message(loAddress, "/clock/new", mesg);
    lo_message_free(mesg);
}

void StatusServer::finishClock(void *node) {
    lo_message mesg = lo_message_new();
    lo_message_add_int64(mesg, hash(node));
    lo_send_message(loAddress, "/clock/finish", mesg);
    lo_message_free(mesg);
}

void StatusServer::mixerGains(void *node, int input, int outputs, std::atomic<float> *gains) {
    lo_message mesg = lo_message_new();
    lo_message_add_int64(mesg, hash(node));
    lo_message_add_int32(mesg, input);
    for (int i = 0; i < outputs; i++) {
        lo_message_add_float(mesg, gains[i]);
    }
    lo_send_message(loAddress, "/mixer/gains", mesg);
    lo_message_free(mesg);
}

void StatusServer::removeProcessor(void *node) {
    lo_message mesg = lo_message_new();
    lo_message_add_int64(mesg, hash(node));
    lo_send_message(loAddress, "/node/remove", mesg);
    lo_message_free(mesg);
}

void StatusServer::addProcessorMidi(void *node, int midiins, int midiouts) {
    lo_message mesg = lo_message_new();
    lo_message_add_int64(mesg, hash(node));
    lo_message_add_int32(mesg, midiins);
    lo_message_add_int32(mesg, midiouts);
    lo_send_message(loAddress, "/processor/addmidi", mesg);
    lo_message_free(mesg);
}

void StatusServer::connect(void *connection, void *connector) {
    lo_message mesg = lo_message_new();
    lo_message_add_int64(mesg, hash(connection));
    lo_message_add_int64(mesg, hash(connector));
    lo_send_message(loAddress, "/connect", mesg);
    lo_message_free(mesg);
}

void StatusServer::addMidiEvent(void *port, void *clock, void *event, float measure,
                                const char *type, int arg1, int arg2) {
    lo_message mesg = lo_message_new();
    lo_message_add_int64(mesg, hash(port));
    lo_message_add_int64(mesg, hash(clock));
    lo_message_add_int64(mesg, hash(event));
    lo_message_add_float(mesg, measure);
    lo_message_add_string(mesg, type);
    lo_message_add_int32(mesg, arg1);
    lo_message_add_int32(mesg, arg2);
    lo_send_message(loAddress, "/event/new", mesg);
    lo_message_free(mesg);
}

void StatusServer::removeMidiEvent(void *port, void *clock, void *event) {
    lo_message mesg = lo_message_new();
    lo_message_add_int64(mesg, hash(port));
    lo_message_add_int64(mesg, hash(clock));
    lo_message_add_int64(mesg, hash(event));
    lo_send_message(loAddress, "/event/remove", mesg);
    lo_message_free(mesg);
}

void StatusServer::scriptReset(bool restart) {
    lo_message mesg = lo_message_new();
    lo_message_add_int32(mesg, restart);
    lo_send_message(loAddress, "/script/reset", mesg);
    lo_message_free(mesg);
}

void StatusServer::clearMidiEvents(void *port, void *clock) {
    lo_message mesg = lo_message_new();
    lo_message_add_int64(mesg, hash(port));
    lo_message_add_int64(mesg, hash(clock));
    lo_send_message(loAddress, "/event/clear", mesg);
    lo_message_free(mesg);
}

void StatusServer::clearScriptEvents(void *clock) {
    lo_message mesg = lo_message_new();
    lo_message_add_int64(mesg, hash(clock));
    lo_send_message(loAddress, "/event/script/clear", mesg);
    lo_message_free(mesg);
}

void StatusServer::addScriptEvent(void *clock, void *event, float measure, const char *name) {
    lo_message mesg = lo_message_new();
    lo_message_add_int64(mesg, hash(clock));
    lo_message_add_int64(mesg, hash(event));
    lo_message_add_float(mesg, measure);
    lo_message_add_string(mesg, name);
    lo_send_message(loAddress, "/event/script/new", mesg);
    lo_message_free(mesg);
}

void StatusServer::removeScriptEvent(void *clock, void *event) {
    lo_message mesg = lo_message_new();
    lo_message_add_int64(mesg, hash(clock));
    lo_message_add_int64(mesg, hash(event));
    lo_send_message(loAddress, "/event/script/remove", mesg);
    lo_message_free(mesg);
}

void StatusServer::io(const char *path, void *node, bool midi, void **ports, uint32_t count) {
    lo_message mesg = lo_message_new();
    lo_message_add_int64(mesg, hash(node));
    if (midi) {
        lo_message_add_true(mesg);
    } else {
        lo_message_add_false(mesg);
    }
    for (uint32_t i = 0; i < count; i++) {
        lo_message_add_int64(mesg, hash(ports[i]));
    }
    lo_send_message(loAddress, path, mesg);
    lo_message_free(mesg);
}

void StatusServer::newNode(const char *path, void *node, const char *name, int inputs,
                           int outputs) {
    lo_message mesg = lo_message_new();
    lo_message_add_int64(mesg, hash(node));
    if (name) {
        lo_message_add_string(mesg, name);
    }
    lo_message_add_int32(mesg, inputs);
    lo_message_add_int32(mesg, outputs);
    lo_send_message(loAddress, path, mesg);
    lo_message_free(mesg);
}

void StatusServer::newNode(const char *path, void *node, const char *name, int inputs, int outputs,
                           int midiins, int midiouts) {
    lo_message mesg = lo_message_new();
    lo_message_add_int64(mesg, hash(node));
    if (name) {
        lo_message_add_string(mesg, name);
    }
    lo_message_add_int32(mesg, inputs);
    lo_message_add_int32(mesg, outputs);
    lo_message_add_int32(mesg, midiins);
    lo_message_add_int32(mesg, midiouts);
    lo_send_message(loAddress, path, mesg);
    lo_message_free(mesg);
}

void StatusServer::audioLevel(void *source, float level, uint16_t index, bool final) {
    collector.free();
    AudioLevelMessage *mesg = new AudioLevelMessage(source, level, index, final);
    if (!processq.push(mesg)) {
        delete mesg;
    }
}

void StatusServer::midiLevel(void *node, uint16_t index, bool activity) {
    MidiLevelMessage *mesg = new MidiLevelMessage(node, index, activity);
    if (!processq.push(mesg)) {
        delete mesg;
    }
}

void StatusServer::mixerGain(void *source, uint16_t input, uint16_t output, float level) {
    collector.free();
    MixerGainMessage *mesg = new MixerGainMessage(source, input, output, level);
    if (!processq.push(mesg)) {
        delete mesg;
    }
}

void StatusServer::clockTime(void *node, float measure) {
    collector.free();
    ClockTimeMessage *mesg = new ClockTimeMessage(node, measure);
    if (!processq.push(mesg)) {
        delete mesg;
    }
}

void StatusServer::clockUpdate(void *node, bool running, float measure, float tempo,
                               uint16_t signum, uint16_t sigden) {
    collector.free();
    ClockUpdateMessage *mesg =
        new ClockUpdateMessage(node, running, measure, tempo, signum, sigden);
    if (!processq.push(mesg)) {
        delete mesg;
    }
}

void StatusServer::run() {
    std::map<void *, lo_message> messageMap;
    while (!mfinished.load()) {
        StatusMessage *smesg;
        while (processq.pop(smesg)) {
            switch (smesg->type()) {
            case StatusMessage::AudioLevel: {
                AudioLevelMessage *lmesg = static_cast<AudioLevelMessage *>(smesg);
                void *source = lmesg->source();
                lo_message mesg = messageMap[source];
                if (!mesg) {
                    mesg = lo_message_new();
                    lo_message_add_int64(mesg, hash(source));
                    messageMap[source] = mesg;
                }
                int missing = lmesg->index() - lo_message_get_argc(mesg) + 1;
                for (int i = 0; i < missing; i++) {
                    lo_message_add_float(mesg, 0.0);
                }
                lo_message_add_float(mesg, lmesg->level());
                if (lmesg->final()) {
                    lo_send_message(loAddress, "/node/level", mesg);
                    messageMap[source] = nullptr;
                    lo_message_free(mesg);
                }
                break;
            }
            case StatusMessage::ClockTime: {
                ClockTimeMessage *cmesg = static_cast<ClockTimeMessage *>(smesg);
                lo_message mesg = lo_message_new();
                lo_message_add_int64(mesg, hash(cmesg->source()));
                lo_message_add_float(mesg, cmesg->measure());
                lo_send_message(loAddress, "/clock/time", mesg);
                lo_message_free(mesg);
                break;
            }
            case StatusMessage::MidiLevel: {
                MidiLevelMessage *mmesg = static_cast<MidiLevelMessage *>(smesg);
                lo_message mesg = lo_message_new();
                lo_message_add_int64(mesg, hash(mmesg->source()));
                lo_message_add_int32(mesg, mmesg->index());
                if (mmesg->level()) {
                    lo_message_add_true(mesg);
                } else {
                    lo_message_add_false(mesg);
                }
                lo_send_message(loAddress, "/node/midilevel", mesg);
                lo_message_free(mesg);
                break;
            }
            case StatusMessage::MixerGain: {
                MixerGainMessage *mmesg = static_cast<MixerGainMessage *>(smesg);
                lo_message mesg = lo_message_new();
                lo_message_add_int64(mesg, hash(mmesg->source()));
                lo_message_add_int32(mesg, mmesg->input());
                lo_message_add_int32(mesg, mmesg->output());
                lo_message_add_float(mesg, mmesg->level());
                lo_send_message(loAddress, "/mixer/gain", mesg);
                lo_message_free(mesg);
                break;
            }
            case StatusMessage::ClockUpdate: {
                ClockUpdateMessage *cmesg = static_cast<ClockUpdateMessage *>(smesg);
                lo_message mesg = lo_message_new();
                lo_message_add_int64(mesg, hash(cmesg->source()));
                if (cmesg->running()) {
                    lo_message_add_true(mesg);
                } else {
                    lo_message_add_false(mesg);
                }
                lo_message_add_float(mesg, cmesg->measure());
                lo_message_add_float(mesg, cmesg->tempo());
                lo_message_add_int32(mesg, cmesg->signum());
                lo_message_add_int32(mesg, cmesg->sigden());
                lo_send_message(loAddress, "/clock/update", mesg);
                lo_message_free(mesg);
                break;
            }
            }
            collector.recycle(smesg);
        }
        std::this_thread::sleep_for(std::chrono::milliseconds(75));
    }
    auto itr = messageMap.begin();
    while (itr != messageMap.end()) {
        lo_message mesg = itr->second;
        lo_message_free(mesg);
        itr = messageMap.erase(itr);
    }
}

} // namespace bipscript
