/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef STATUSSERVER_H
#define STATUSSERVER_H

#include <atomic>
#include <boost/lockfree/spsc_queue.hpp>
#include <lo/lo.h>
#include <string>

#include "collector.h"
#include "memorypool.h"

namespace bipscript {

class StatusMessage {

  public:
    enum Type { AudioLevel, MidiLevel, MixerGain, ClockUpdate, ClockTime };
    StatusMessage(Type type) : mtype(type) {}
    virtual ~StatusMessage() {}
    // memory pooled
    void *operator new(size_t size) { return MemoryPool::processPool().malloc(size); }
    void operator delete(void *p) { MemoryPool::processPool().free(p); }
    uint32_t type() { return mtype; }

  private:
    Type mtype;
};

class AudioLevelMessage : public StatusMessage {
    void *msource;
    float mlevel;
    uint16_t mindex;
    bool mfinal;

  public:
    AudioLevelMessage(void *source, float level, uint16_t index, bool final)
        : StatusMessage(AudioLevel), msource(source), mlevel(level), mindex(index), mfinal(final) {}
    void *source() { return msource; }
    float level() { return mlevel; }
    uint16_t index() { return mindex; }
    bool final() { return mfinal; }
};

class MidiLevelMessage : public StatusMessage {
    void *msource;
    uint16_t mindex;
    bool mactivity;

  public:
    MidiLevelMessage(void *node, uint16_t index, bool activity)
        : StatusMessage(MidiLevel), msource(node), mindex(index), mactivity(activity) {}
    void *source() { return msource; }
    uint16_t index() { return mindex; }
    float level() { return mactivity; }
};

class MixerGainMessage : public StatusMessage {
    void *msource;
    uint16_t minput;
    uint16_t moutput;
    float mlevel;

  public:
    MixerGainMessage(void *source, uint16_t input, uint16_t output, float level)
        : StatusMessage(MixerGain), msource(source), minput(input), moutput(output), mlevel(level) {
    }
    void *source() { return msource; }
    uint16_t input() { return minput; }
    uint16_t output() { return moutput; }
    float level() { return mlevel; }
};

class ClockUpdateMessage : public StatusMessage {
    void *msource;
    bool mrunning;
    float mmeasure;
    float mtempo;
    uint16_t msignum;
    uint16_t msigden;

  public:
    ClockUpdateMessage(void *source, bool running, float measure, float tempo, uint16_t signum,
                       uint16_t sigden)
        : StatusMessage(ClockUpdate), msource(source), mrunning(running), mmeasure(measure),
          mtempo(tempo), msignum(signum), msigden(sigden) {}
    void *source() { return msource; }
    bool running() { return mrunning; }
    float measure() { return mmeasure; }
    float tempo() { return mtempo; }
    uint16_t signum() { return msignum; }
    uint16_t sigden() { return msigden; }
};

class ClockTimeMessage : public StatusMessage {
    void *msource;
    float mmeasure;

  public:
    ClockTimeMessage(void *source, float measure)
        : StatusMessage(ClockTime), msource(source), mmeasure(measure) {}
    void *source() { return msource; }
    float measure() { return mmeasure; }
};

class StatusServer {
    std::atomic<bool> mactive;
    std::atomic<bool> mfinished;
    std::string mport;
    lo_address loAddress;
    std::hash<void *> hash;
    // for process thread
    boost::lockfree::spsc_queue<StatusMessage *> processq;
    pthread_t thread;
    Collector<StatusMessage> collector;
    // singleton
    StatusServer() : processq(512) {}

  public:
    ~StatusServer();
    static StatusServer &instance() {
        static StatusServer instance;
        return instance;
    }
    void activate();
    bool active() { return mactive.load(); }
    void port(std::string port) { mport = port; }
    // status updates
    void newClock(void *node, const char *name, float tempo);
    void newClock(void *node, const char *name, float tempo, int signum, int sigden);
    void finishClock(void *node);
    void newMixer(void *node, int inputs, int outputs) {
        newNode("/mixer/new", node, "", inputs, outputs);
    }
    void mixerGains(void *node, int input, int connections, std::atomic<float> *gains);
    void newLv2Plugin(void *node, const char *name, int inputs, int outputs, int midiins,
                      int midiouts) {
        newNode("/lv2/new", node, name, inputs, outputs, midiins, midiouts);
    }
    void newProcessor(void *node, const char *name, int inputs, int outputs) {
        newNode("/processor/new", node, name, inputs, outputs, 0, 0);
    }
    void removeProcessor(void *node);
    void addProcessorMidi(void *node, int midiins, int midiouts);
    void newInput(void *node, const char *name, int channels) {
        newNode("/input/new", node, name, 0, channels);
    }
    void newOutput(void *node, const char *name, int channels) {
        newNode("/output/new", node, name, channels, 0);
    }
    void newMidiInput(void *node, const char *name, int channels) {
        newNode("/midiin/new", node, name, 0, channels);
    }
    void newMidiOutput(void *node, const char *name, int channels) {
        newNode("/midiout/new", node, name, channels, 0);
    }
    void connect(void *connect, void *connector);
    void connectors(void *node, void **ports, uint32_t count) {
        io("/node/connectors", node, false, ports, count);
    }
    void connections(void *node, void **ports, uint32_t count) {
        io("/node/connections", node, false, ports, count);
    }
    void midiconnectors(void *node, void **ports, uint32_t count) {
        io("/node/connectors", node, true, ports, count);
    }
    void midiconnections(void *node, void **ports, uint32_t count) {
        io("/node/connections", node, true, ports, count);
    }
    void addMidiEvent(void *port, void *clock, void *event, float measure, const char *type,
                      int arg1, int arg2);
    void removeMidiEvent(void *port, void *clock, void *event);
    void scriptReset(bool restart);
    void clearMidiEvents(void *port, void *clock);
    void clearScriptEvents(void *clock);
    void addScriptEvent(void *clock, void *event, float measure, const char *name);
    void removeScriptEvent(void *clock, void *event);
    // process thread
    void audioLevel(void *node, float level, uint16_t index, bool final);
    void midiLevel(void *node, uint16_t index, bool activity);
    void mixerGain(void *node, uint16_t input, uint16_t output, float level);
    void clockTime(void *node, float measure);
    void clockUpdate(void *node, bool running, float measure, float tempo, uint16_t signum,
                     uint16_t sigden);
    void run();
    void shutdown() {
        mfinished.store(true);
        pthread_join(thread, nullptr);
    }

  private:
    void newNode(const char *path, void *node, const char *name, int inputs, int outputs);
    void newNode(const char *path, void *node, const char *name, int inputs, int outputs,
                 int midiins, int midiouts);
    void io(const char *path, void *node, bool midi, void **ports, uint32_t count);
};

} // namespace bipscript
#endif // STATUSSERVER_H
