/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef SYSTEMTIME_H
#define SYSTEMTIME_H

#include <chrono>
#include <ctime>
#include <stdexcept>
#include <stdio.h>
#include <time.h>

#include "systemtype.h"

namespace bipscript {
namespace system {

// TODO: give Time object sub-second resolution instead of this class?
class Timer {
    long stime;

  public:
    Timer() : stime(now()) {}
    long elapsed() { return now() - stime; }
    void reset() { stime = now(); }

  private:
    long now() {
        auto time = std::chrono::system_clock::now().time_since_epoch();
        return std::chrono::duration_cast<std::chrono::nanoseconds>(time).count();
    }
};

class Time {

    tm value;

  public:
    // ctors
    Time() {
        time_t t = ::time(nullptr);
        value = *localtime(&t);
    }
    Time(const char *iso) { set(iso); }
    Time(int year, int mon, int day, int hour = 0, int min = 0, int sec = 0) {
        value.tm_sec = sec;
        value.tm_min = min;
        value.tm_hour = hour;
        value.tm_mday = day;
        value.tm_mon = mon - 1;
        value.tm_year = year - 1900;
        value.tm_wday = 0;
        value.tm_yday = 0;
        value.tm_isdst = 0;
        std::mktime(&value);
    }
    // properties
    int second() { return value.tm_sec; }
    void second(int second) {
        value.tm_sec = second;
        std::mktime(&value);
    }
    int minute() { return value.tm_min; }
    void minute(int minute) {
        value.tm_min = minute;
        std::mktime(&value);
    }
    int hour() { return value.tm_hour; }
    void hour(int hour) {
        value.tm_hour = hour;
        std::mktime(&value);
    }
    int day() { return value.tm_mday; }
    void day(int day) {
        value.tm_mday = day;
        std::mktime(&value);
    }
    int weekday() { return value.tm_wday; }
    int month() { return value.tm_mon + 1; }
    void month(int month) {
        value.tm_mon = month - 1;
        std::mktime(&value);
    }
    int year() { return value.tm_year + 1900; }
    void year(int year) { value.tm_year = year - 1900; }
    // methods
    std::string dayname() {
        char buffer[256];
        strftime(buffer, 256, "%A", &value);
        return buffer;
    }
    std::string monthname() {
        char buffer[256];
        strftime(buffer, 256, "%B", &value);
        return buffer;
    }
    std::string date() {
        char buffer[256];
        strftime(buffer, 256, "%F", &value);
        return buffer;
    }
    std::string time() {
        char buffer[256];
        strftime(buffer, 256, "%T", &value);
        return buffer;
    }
    std::string format(const char *spec) {
        char buffer[1024];
        int result = strftime(buffer, 1024, spec, &value);
        if (!result) {
            throw std::logic_error("Error formatting string, please check format spec\n");
        }
        return buffer;
    }
    void set(const char *iso) {
        char *result = strptime(iso, "%Y-%m-%d %H:%M:%S", &value);
        if (!result) {
            throw std::logic_error("Error parsing date format, please use ISO-8601");
        }
        std::mktime(&value);
    }
};

} // namespace system
} // namespace bipscript
#endif // SYSTEMTIME_H
