/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "variableclock.h"

namespace bipscript {
namespace time {

AdvancedClock::~AdvancedClock() {
    ClockChangeEvent *evt = eventBuffer.firstEvent();
    while (evt) {
        ClockChangeEvent *done = evt;
        evt = eventBuffer.pop();
        delete done;
    }
}

/**
 * Move the clock to the following period.
 */
void AdvancedClock::tick() {
    int newEvents = eventBuffer.update();
    if (newEvents) {
        // relocate in case new events are in the past
        relocate(following.frame());
    } else {
        // status reporting on bar change or relocation
        StatusServer &statusserver = StatusServer::instance();
        if (statusserver.active()) {
            if (following.measure() < current.measure() ||
                static_cast<uint32_t>(following.measure()) !=
                    static_cast<uint32_t>(current.measure())) {
                statusserver.clockTime(this, following.measure());
            }
            if (following.bpm() != current.bpm()) {
                statusUpdate(following.bpm());
            }
        }
        // update current to following position
        current = following;
        calculatePeriod();
    }
}

void AdvancedClock::statusUpdate(float bpm) {
    StatusServer &statusserver = StatusServer::instance();
    if (statusserver.active()) {
        float measure = following.measure();
        uint16_t num = following.signature().numerator();
        uint16_t den = following.signature().denominator();
        statusserver.clockUpdate(this, mtimer.running(), measure, bpm, num, den);
    }
}

/**
 * Reposition the current window start to the given sample location. Takes into account all
 * previous events except repositions. Sets nextEvent to the first event not earlier than the
 * new current start position. This method is used only by the external timebase master.
 */
void AdvancedClock::relocate(samples_t location) {
    eventBuffer.update();
    nextEvent = eventBuffer.firstEvent();
    Window *last = &initial;
    samples_t range = location;
    while (nextEvent && nextEvent->offset(last) < range) {
        // set size on last pos
        samples_t subrange = nextEvent->start();
        last->size(subrange);
        // update position attached to this event
        Window *window = nextEvent->window();
        // set frame via offset from last
        window->frame(last->frame() + subrange);
        // event-specific update
        if (nextEvent->type() == ClockChangeEvent::TEMPO) {
            window->signature(last->signature());
        } else if (nextEvent->type() == ClockChangeEvent::SIGNATURE) {
            window->bpm(last->bpm());
        }
        // next event
        nextEvent = eventBuffer.nextEvent(nextEvent);
        range -= subrange;
        last = window;
    }
    // wrap up last position
    last->size(range);
    // define current (minus size)
    current.frame(last->frame() + range);
    current.measure(last->measure() + last->moffset(range));
    current.signature(last->signature());
    current.bpm(last->bpm());
    current.reposition(true);
    calculatePeriod(); // current size defined here
}

void AdvancedClock::relocate(float measure) {
    eventBuffer.update();
    current.measure(measure);
    current.reposition(true);
    nextEvent = calculateWindow(&current);
    calculatePeriod();
}

/**
 * maps a measure to a frame value, taking time change events into account
 */
samples_t AdvancedClock::toFrame(float measure) {
    Window window;
    window.measure(measure);
    calculateWindow(&window);
    return window.frame();
}

/**
 * Calculates the next clock period, taking into account any time change events occuring during this
 * period. This method assumes the "current" position is correct except for size, calculates the
 * "following" period as well as creating the null-terminated list of time windows for this period.
 */
void AdvancedClock::calculatePeriod() {
    // update current with any immediate tempo changes
    if (tempoChange > 0.0f) {
        current.bpm(tempoChange);
    }
    // start with current window
    Window *last = &current;
    samples_t range = period();
    while (nextEvent && nextEvent->offset(last) < range) {
        // set period event
        samples_t subrange = nextEvent->start();
        // update window attached to this event
        Window *window = nextEvent->window();
        window->start(period() - (range - subrange));
        // event-specific update
        if (nextEvent->type() == ClockChangeEvent::TEMPO) {
            // set frame via offset from last
            window->frame(last->frame() + subrange);
            window->signature(last->signature());
            // tempo event cancels any manual tempo change
            tempoChange = -1.0f;
            nextEvent = eventBuffer.nextEvent(nextEvent);
        } else if (nextEvent->type() == ClockChangeEvent::SIGNATURE) {
            window->frame(last->frame() + subrange);
            window->bpm(last->bpm());
            nextEvent = eventBuffer.nextEvent(nextEvent);
        } else if (nextEvent->type() == ClockChangeEvent::RELOCATION) {
            nextEvent = calculateWindow(window); // updates window frame
        }
        // set size and next on last pos
        last->size(subrange);
        last->next = window;
        // shrink range
        range -= subrange;
        last = window;
    }
    // wrap up last position
    last->size(range);
    last->next = nullptr;
    // define following (minus size)
    following.frame(last->frame() + range);
    following.measure(last->measure() + last->moffset(range));
    following.signature(last->signature());
    following.bpm(last->bpm());
    following.reposition(false);
    // update mtempo
    mtempo.store(current.bpm());
}

/**
 * Calculates the frame, signature and tempo of a window with given measure. Takes into account all
 * previous events except repositions. Returns the first event not earlier than this window's start
 * position.
 */
ClockChangeEvent *AdvancedClock::calculateWindow(Window *in) {
    ClockChangeEvent *event = eventBuffer.firstEvent();
    Window *last = &initial;
    while (event && event->measure() < static_cast<float>(in->measure())) {
        Window *window = event->window();
        // set frame and measure via offset from last
        window->frame(last->frame() + last->offset(window->measure()));
        // event-specific update
        if (event->type() == ClockChangeEvent::TEMPO) {
            window->signature(last->signature());
            last = window;
        } else if (event->type() == ClockChangeEvent::SIGNATURE) {
            window->bpm(last->bpm());
            last = window;
        }
        event = eventBuffer.nextEvent(event);
    }
    in->frame(last->frame() + last->offset(in->measure()));
    in->signature(last->signature());
    in->bpm(last->bpm());
    return event;
}

void VariableClock::update(samples_t) {
    ClockChange *command;
    float dotick = true;
    float changed = false;
    while (commands.pop(command)) {
        if (command->type() == ClockChange::Type::STATE) {
            mtimer.running(static_cast<StateChange *>(command)->start());
        } else if (command->type() == ClockChange::Type::RELOCATION) {
            relocate(static_cast<Relocation *>(command)->location());
            dotick = false;
        } else if (command->type() == ClockChange::Type::TEMPO) {
            tempoChange = static_cast<TempoChange *>(command)->tempo();
        }
        ClockCollector::instance().recycle(command);
        changed = true;
    }
    if (dotick && mtimer.running()) {
        tick();
    }
    if (changed) {
        statusUpdate(mtempo);
    }
    updateRemotes();
}

} // namespace time
} // namespace bipscript
